/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef HSS_SEQ_GENERATESCHUR_HPP
#define HSS_SEQ_GENERATESCHUR_HPP

template<typename T, typename S> void HSS_tree<T,S>::HSS_seq_GenerateSchur() {
  /* Assuming the matrix is already compressed and partially
   * factored, build the generators theta and psi s.t.
   * the Schur complement S is S=A22 - theta Vtilda_k^* psi, with
   * A22 (submatrix of the original matrix A) in compressed
   * form and theta and psi two dense matrices.
   *
   * With k the left child of the root node and w the right
   * child of the root node:
   * theta = U_q^big B_qk
   * psi = Utilda_k^-1 Ltilda_k^-1 Ptilda_k U_k B_kq V_q^big^*
   *
   * The low-rank Schur complement update is represented as
   * theta Vtilda_k^* psi
   */

  int i, j, k;
  int qnode;
  int br, bc;
  int ch1, ch2;
  int *children1, *children2;
  int *rowsUbig, *rowsVbig;
  T *Ue, *Ve, *tmp;
  T **Ubig, **Vbig;
  HSS_node<T,S> *child1, *child2;

  /* The right child of the root node is its predecessor in the postorder */
  qnode=dim_tree-2;

  /* We compute U_q^big and V_q^big by traversing the subtree rooted at q
   * and multiplying out the generators (we "unfold" the nested basis).
   * Here U_bug and V_big are not fields  of a node but arrays indexed following the
   * postorder; the pointers "children" of a node are not enough to find the
   * right in y and z, what we need is numbers in the postorder. "children1"
   * and "children2" are the same information as children but they provide
   * numbers (indices in the tree).
   *
   * children1: children[i]==j: node number j is left child of node number i.
   * child1: pointer to left child (type HSS_node) of node nodes[i].
   */

  children1=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    children1[i]=-1;
    if(nodes[i]->nr_children!=0)
      for(j=0;j<dim_tree;j++)
        if(nodes[j]==nodes[i]->children[0]) {
          children1[i]=j;
          break;
        }
  }
  children2=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    children2[i]=-1;
    if(nodes[i]->nr_children!=0)
      for(j=0;j<dim_tree;j++)
        if(nodes[j]==nodes[i]->children[1]) {
          children2[i]=j;
          break;
        }
  }

  Ubig=new T*[dim_tree];
  rowsUbig=new int[dim_tree];
  Vbig=new T*[dim_tree];
  rowsVbig=new int[dim_tree];

  for(i=knode+1;i<=qnode;i++) {
    if(nodes[i]->nr_children==0) {
      br=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      bc=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
    } else {
      child1=nodes[i]->children[0];
      child2=nodes[i]->children[1];
      ch1=children1[i];
      ch2=children2[i];
      br=child1->u_rank+child2->u_rank;
      bc=child1->v_rank+child2->v_rank;
    }
    /* We expand U into its explicit form U=P*[I;E] */
    Ue=new T[br*nodes[i]->u_rank]();
    for(j=0;j<nodes[i]->u_rank;j++)
      Ue[j*br+j]=ONE;
    lacpy('N',nodes[i]->u_rows,nodes[i]->u_rank,nodes[i]->U_E,nodes[i]->u_rows,&Ue[nodes[i]->u_rank],br);
    for(k=br-1;k>=0;k--)
      if(k!=nodes[i]->U_perm[k]-1)
        swap(nodes[i]->u_rank,&Ue[k],br,&Ue[nodes[i]->U_perm[k]-1],br);

    /* We expand V into its explicit form V=P*[I;E] */
    Ve=new T[bc*nodes[i]->v_rank]();
    for(j=0;j<nodes[i]->v_rank;j++)
      Ve[j*bc+j]=ONE;
    lacpy('N',nodes[i]->v_rows,nodes[i]->v_rank,nodes[i]->V_E,nodes[i]->v_rows,&Ve[nodes[i]->v_rank],bc);
    for(k=bc-1;k>=0;k--)
      if(k!=nodes[i]->V_perm[k]-1)
        swap(nodes[i]->v_rank,&Ve[k],bc,&Ve[nodes[i]->V_perm[k]-1],bc);

    if(nodes[i]->nr_children==0) {
      /* Leaf: U^big=U, V^big=V, stored in explicit form */
      Ubig[i]=Ue;
      rowsUbig[i]=br;
      Vbig[i]=Ve;
      rowsVbig[i]=bc;
    } else {
      /* Interior node: U^big = diag(U^big_child1,U^big_child2) * U, ditto for V */
      rowsUbig[i]=rowsUbig[ch1]+rowsUbig[ch2];
      Ubig[i]=new T[rowsUbig[i]*nodes[i]->u_rank]();
      if(rowsUbig[ch1] && br && rowsUbig[i])
        gemm('N','N',rowsUbig[ch1],nodes[i]->u_rank,nodes[ch1]->u_rank,ONE,Ubig[ch1],rowsUbig[ch1],Ue,br,ZERO,Ubig[i],rowsUbig[i]);
      if(rowsUbig[ch2] && br && rowsUbig[i])
        gemm('N','N',rowsUbig[ch2],nodes[i]->u_rank,nodes[ch2]->u_rank,ONE,Ubig[ch2],rowsUbig[ch2],&Ue[nodes[ch1]->u_rank],br,ZERO,&Ubig[i][rowsUbig[ch1]],rowsUbig[i]);
      delete[] Ue;
      delete[] Ubig[ch1];
      delete[] Ubig[ch2];

      rowsVbig[i]=rowsVbig[ch1]+rowsVbig[ch2];
      Vbig[i]=new T[rowsVbig[i]*nodes[i]->v_rank]();
      if(rowsVbig[ch1] && bc && rowsVbig[i])
        gemm('N','N',rowsVbig[ch1],nodes[i]->v_rank,nodes[ch1]->v_rank,ONE,Vbig[ch1],rowsVbig[ch1],Ve,bc,ZERO,Vbig[i],rowsVbig[i]);
      if(rowsVbig[ch2] && bc && rowsVbig[i])
        gemm('N','N',rowsVbig[ch2],nodes[i]->v_rank,nodes[ch2]->v_rank,ONE,Vbig[ch2],rowsVbig[ch2],&Ve[nodes[ch1]->v_rank],bc,ZERO,&Vbig[i][rowsVbig[ch1]],rowsVbig[i]);
      delete[] Ve;
      delete[] Vbig[ch1];
      delete[] Vbig[ch2];
    }

  }

#if defined(debug)
  printMatrix("U_q^big",Ubig[qnode],rowsUbig[qnode],nodes[qnode]->u_rank,rowsUbig[qnode]);
  printMatrix("V_q^big",Vbig[qnode],rowsVbig[qnode],nodes[qnode]->v_rank,rowsVbig[qnode]);
#endif

  delete[] children1;
  delete[] children2;

  /* Compute theta = U_q^big B_qk */
  br=nodes[qnode]->tree_ie-nodes[qnode]->tree_ib+1;
  theta=new T[br*nodes[knode]->v_rank](); /* Field of the HSS tree */
  rtheta=br; /* (same) */
  ctheta=nodes[knode]->v_rank; /* (same) */
  if(nodes[qnode]->u_rank)
    gemm('N','N',br,nodes[knode]->v_rank,nodes[qnode]->u_rank,ONE,Ubig[qnode],br,nodes[dim_tree-1]->B21,nodes[qnode]->u_rank,ZERO,theta,br);

#if defined(debug)
  printMatrix("theta",theta,rtheta,ctheta,rtheta);
#endif

  /* Compute psi = Utilda_k^-1 Ltilda_k^-1 Ptilda_k U_k B_kq V_q^big^*
   * We compute tmp = U_k B_kq using the structured product,
   * then tmp = Utilda_k^-1 Ltilda_k^-1 Ptilda_k tmp inplace, in one shot,
   * using getrs (the factors are in the L field, and the permutation is in
   * the "piv" field).
   * The multiplication with U_q^big comes last (O(n) rows).
   */
  if(nodes[knode]->nr_children==0) {
    br=nodes[knode]->tree_ie-nodes[knode]->tree_ib+1;
  } else {
    child1=nodes[knode]->children[0];
    child2=nodes[knode]->children[1];
    br=child1->u_rank+child2->u_rank;
  }
  tmp=new T[br*nodes[qnode]->v_rank]();
  nodes[knode]->seq_applyGenerator('N','U',nodes[qnode]->v_rank,ONE,nodes[dim_tree-1]->B12,nodes[knode]->u_rank,ZERO,tmp,br);
  if(br && nodes[qnode]->v_rank)
    getrs('N',br,nodes[qnode]->v_rank,nodes[knode]->L,br,nodes[knode]->piv,tmp,br);
  bc=nodes[qnode]->tree_ie-nodes[qnode]->tree_ib+1;
  psi=new T[br*bc](); /* Field of the HSS tree */
  rpsi=br; /* (same) */
  cpsi=bc; /* (same) */
  if(br)
    gemm('N','C',br,bc,nodes[qnode]->v_rank,ONE,tmp,br,Vbig[qnode],bc,ZERO,psi,br);
  delete[] tmp;

#if defined(debug)
  printMatrix("psi",psi,rpsi,cpsi,rpsi);
#endif

  delete[] Ubig[qnode];
  delete[] Vbig[qnode];

  delete[] rowsUbig;
  delete[] rowsVbig;

}

#endif // HSS_SEQ_GENERATESCHUR_HPP
