/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef HSS_TREE_HPP
#define HSS_TREE_HPP

#include "HSS_node.hpp"

template<typename T, typename S> class HSS_tree {
  /* An HSS_tree is essentialy a list of nodes ordered in postorder.
   *
   * Global information:
   *   dim_tree is the number of nodes.
   *   nr_lvl is the number of levels in the tree.
   *   a_dim is the dimension of the matrix associated with the tree.
   *   each process i takes care of a serial subtree defined by nodes
   *   subtree[2*i]...subtree[2*i+1], and takes part to the computations of
   *   nodes par_lvl=subtree[2*i+1]+1...dim_tree-1, with operations using
   *   PBLAS/ScaLAPACK and BLACS contexts defined in the nodes themselves.
   *   comm is the communicator used for the computations.
   *   theta and psi represent the Schur complement low-rank update;
   *   theta has size rtheta x ctheta and psi has size rpsi*cpsi.
   *   desc_theta and desc_psi are the respective descriptors.
   *   y is the partial solution after a forward substitution. It is
   *   useful to save it at the tree level for Schur complement
   *   reduction/expansion operations (cf. HSS_*_Solve.hpp).
   *   knode is the index of the left child of the root node
   *   (used for Schur complement features).
   *   yr is a copy of y[knode] but is it distributed onto
   *   the grid associated with the root node instead of the grid
   *   associated with the knode.
   *   desc_yr is the descriptor of yr;
   *
   * Local information (i.e. different on each process):
   *   par_nodes is the list of interior nodes that "myid" shares with other processes.
   *   time_Dist time spent in the initial matrix redistribution.
   *   time_ID cumulated time in Interpolative decompositions.
   *   time_Rand cumulated time in random number generation.
   *   time_GEMM cumulated time in global sample generation (S=A*R).
   *   obj pointer to user data (used for matrix-free compression).
   */
public:
  int dim_tree;
  int nr_lvl;
  int a_dim;
  int *subtree;
  int *par_nodes;
  int par_lvl;
  HSS_node<T,S> **nodes;
  MPI_Comm comm;
  double time_Dist;
  double time_ID;
  double time_Rand;
  double time_GEMM;
  double time_Amult;
  double time_Barrier;
  double time_Elements;
  double time_Ptc;
  double time_Recv;
  double time_samples_GEMM;
  double time_MPI_Allgather;
  void *obj;
  T *theta;
  T *psi;
  int rtheta;
  int ctheta;
  int rpsi;
  int cpsi;
  int *desc_theta;
  int *desc_psi;
  T **y;
  T *yr;
  int *desc_yr;
  T *zk12;
  int *desc_zk12;
  int knode;

  HSS_tree();
  void generateBinaryTree(int, int);
  void generateConstrainedBinaryTree(int, int, int);
  void generateCombTree(int, int);
  void generateFromPartitionTree(Partition_tree *);
  void proportionalMapping(int);
  void print();
  double getSizeHSSFactors();
  double getSizeULVFactors(int);
  double getSizeSchurUpdate();
  double getFlopsHSSRCompression(int, bool);
  double getFlopsULVFactorization(int);
  double getFlopsSolve(int);
  double getFlopsProduct(int);
  double getFlopsSchurUpdate();
  void HSSR_seq_Compress(T *, int *, S, int, int, int, int, int *, int *, int, T *, T *, T *, T *,
			 StrumpackRand *, void (*f)(void *, int *, int *, T *, int *),
			 void (*g)(void *, T *, T *, int *, T *, T *, int *),int, bool);
  void HSSR_par_Compress(T *, int *, S, int, int, int, int, int *, int *, gridinfo *, T *, T *, T *, T *, int *,
			 void (*f)(void *, int *, int *, T *, int *),
			 void (*g)(void *, T *, T *, int *, T *, T *, int *),bool);
  void HSS_seq_Product(char, T, T *, int, T, T *, int);
  void HSS_par_Product(char, T, T *, int *, int, T, T *, int *, gridinfo *, int);
  void HSS_seq_ULV(int, int, int);
  void HSS_par_ULV(gridinfo *, int);
  void HSS_seq_Solve(T *, int, T *, int);
  void HSS_par_Solve(T *, int *, int, T *, int *, gridinfo *, int);
  void HSS_seq_Extraction(T *, int *, int, int *, int);
  void HSS_par_Extraction(T *, int *, int *, int, int *, int, gridinfo *);
  void HSS_seq_GenerateSchur();
  void HSS_par_GenerateSchur(gridinfo *);
  void writeStats();
  void cleanup(void);

private:
  void setDefaultState();
  void fillBinaryTree(int, int, int, int, int);
  void distVectorTree(T *, int *, int, T **, int **, gridinfo *, int);
  void undistVectorTree(T *, int *, int, T *, int *, gridinfo *, int, int);
  void extractRowVectorTree(T *, int *, int *, int, int, T *, int *, gridinfo *, int);
  void distMatrixTree(T *, int *, int, T **, int **, T **, int **, T **, int **, gridinfo *, int);
  void extractLocalSubtree(HSS_tree<T,S>* loc_htree);
  void restart(T **, T **, T **, T **, int **, int, int, T *, int *,
	       void (*f)(void *, T *, T *, int *, T *, T *, int *),
	       gridinfo *, int, StrumpackRand *, double *, double *, double *);
  void participate_to_calls(void (*elements)(void *, int *, int *, T *, int *), int, MPI_Comm, double *, double *, double *);

  /* Constants */
public:
  T NONE;
  T ZERO;
  T  ONE;
};

template<typename T, typename S> HSS_tree<T,S>::HSS_tree(void) {
  NONE = -1;
  ZERO =  0;
  ONE  =  1;
}

template<typename T, typename S> void HSS_tree<T,S>::setDefaultState(){
  /* Default state of every field of the tree and every field of
   * every node.
   */
  int i;

  theta=NULL;
  psi=NULL;
  rtheta=-1;
  ctheta=-1;
  rpsi=-1;
  cpsi=-1;
  desc_theta=NULL;
  desc_psi=NULL;
  knode=-1;
  zk12=NULL;
  desc_zk12=NULL;
  y=NULL;
  yr=NULL;
  desc_yr=NULL;
  for(i=0;i<dim_tree;i++) {
    if(nodes[i]==NULL) {
      std::cout << "Problem. Exiting.\n" << std::endl;
      exit(1);
    }
    nodes[i]->U_E=NULL;
    nodes[i]->U_perm=NULL;
    nodes[i]->u_rows=-1;
    nodes[i]->u_rank=-1;
    nodes[i]->descU=NULL;
    nodes[i]->Ir=NULL;
    nodes[i]->Jr=NULL;
    nodes[i]->V_E=NULL;
    nodes[i]->V_perm=NULL;
    nodes[i]->v_rows=-1;
    nodes[i]->v_rank=-1;
    nodes[i]->descV=NULL;
    nodes[i]->Ic=NULL;
    nodes[i]->Jc=NULL;
    nodes[i]->B12=NULL;
    nodes[i]->B21=NULL;
    nodes[i]->L=NULL;
    nodes[i]->Vt=NULL;
    nodes[i]->W=NULL;
    nodes[i]->Q=NULL;
    nodes[i]->piv=NULL;
    nodes[i]->D=NULL;
    nodes[i]->Dt=NULL;
    nodes[i]->procs[0]=-1;
    nodes[i]->procs[1]=-1;
    nodes[i]->context=-1;
    nodes[i]->nprow=-1;
    nodes[i]->npcol=-1;
    nodes[i]->myrow=-1;
    nodes[i]->mycol=-1;
    nodes[i]->allctxt=-1;
  }
  time_Dist=0;
  time_ID=0;
  time_Rand=0;
  time_GEMM=0;
  time_Amult=0;
  time_Barrier=0;
  time_Elements=0;
  time_Ptc=0;
  time_Recv=0;
  time_samples_GEMM=0;
  time_MPI_Allgather=0;

}

template<typename T, typename S> void HSS_tree<T,S>::fillBinaryTree(int start, int end, int ib, int ie, int lvl) {
  /* Map interval [ib ie] on nodes start...end of the tree with lvl levels in the subtree.
   * The root of the subtree is already allocated and has interval values. The routine
   * creates its children.
   */
  int half,ih;

  if(end-start+1!=std::pow(2e0,lvl)-1) {
    std::cout << "Problem size vs lvl." << std::endl;
    return;
  }

  if(lvl==1) {
    nodes[end]->nr_children=0;
    nodes[end]->children=NULL;
  } else {
    nodes[end]->nr_children=2;
    nodes[end]->children=new HSS_node<T,S>*[2];
    ih=(ie-ib)/2;
    half=(end-start)/2;
    nodes[end]->children[0]=new HSS_node<T,S>;
    nodes[end]->children[0]->tree_ib=ib;
    nodes[end]->children[0]->tree_ie=ib+ih;
    nodes[end]->children[0]->parent=nodes[end];
    nodes[start+half-1]=nodes[end]->children[0];
    nodes[end]->children[1]=new HSS_node<T,S>;
    nodes[end]->children[1]->tree_ib=ib+ih+1;
    nodes[end]->children[1]->tree_ie=ie;
    nodes[end]->children[1]->parent=nodes[end];
    nodes[end-1]=nodes[end]->children[1];
    fillBinaryTree(start,start+half-1,ib,ib+ih,lvl-1);
    fillBinaryTree(start+half,end-1,ib+ih+1,ie,lvl-1);
  }

}

template<typename T, typename S> void HSS_tree<T,S>::generateBinaryTree(int m, int n) {
  /* Generate a binary tree with m levels whose leaves cluster [1 n] */

  int i;

  a_dim=n;
  nr_lvl=m;
  dim_tree=std::pow(2e0,m)-1;

  nodes=new HSS_node<T,S>*[dim_tree];
  subtree=NULL;
  par_lvl=-1;
  par_nodes=NULL;
  for(i=0;i<dim_tree;i++)
    nodes[i]=NULL;
  nodes[dim_tree-1]=new HSS_node<T,S>;
  nodes[dim_tree-1]->tree_ib=1;
  nodes[dim_tree-1]->tree_ie=n;
  nodes[dim_tree-1]->parent=NULL;
  fillBinaryTree(0,dim_tree-1,1,n,nr_lvl);

  setDefaultState();

}

template<typename T, typename S> void HSS_tree<T,S>::generateConstrainedBinaryTree(int m, int ns, int n) {
  /* Generate a binary tree with m levels whose leaves cluster [1 n] and
   * such that the two children of the root node (if m>1) cluster [1 ns]
   * and [ns+1 n] respectively, with 1 <= ns < n.
   *
   * The number of levels m is assumed to be compatible with n and ns,
   * i.e., m == max(log2(ns)+1,log2(n-ns)+1)+1.
   */
  int i;
  int dim_left_tree;
  int nr_left_lvl, nr_right_lvl;

  a_dim=n;
  nr_lvl=m;

  dim_tree=0;
  /* Left subtree: m-1 levels or less depending on ns */
  if(m-1>static_cast<int>(floor(log2((double)ns))+1e0)) {
    dim_tree+=2*ns-1;
    nr_left_lvl=static_cast<int>(floor(log2((double)ns))+1e0);
  } else {
    dim_tree+=std::pow(2e0,m-1)-1;
    nr_left_lvl=m-1;
  }
  dim_left_tree=dim_tree;
  /* Right subtree: m-1 levels or less depending on ns */
  if(m-1>static_cast<int>(floor(log2((double)n-ns))+1e0)) {
    dim_tree+=2*(n-ns)+1;
    nr_right_lvl=static_cast<int>(floor(log2((double)n-ns))+1e0);
  } else {
    dim_tree+=std::pow(2e0,m-1)-1;
    nr_right_lvl=m-1;
  }
  /* Root */
  dim_tree++;

  nodes=new HSS_node<T,S>*[dim_tree];
  subtree=NULL;
  par_lvl=-1;
  par_nodes=NULL;
  for(i=0;i<dim_tree;i++)
    nodes[i]=NULL;
  /* Root */
  nodes[dim_tree-1]=new HSS_node<T,S>;
  nodes[dim_tree-1]->tree_ib=1;
  nodes[dim_tree-1]->tree_ie=n;
  nodes[dim_tree-1]->parent=NULL;
  nodes[dim_tree-1]->nr_children=2;
  nodes[dim_tree-1]->children=new HSS_node<T,S>*[2];
  /* Left child */
  nodes[dim_left_tree-1]=new HSS_node<T,S>;
  nodes[dim_left_tree-1]->tree_ib=1;
  nodes[dim_left_tree-1]->tree_ie=ns;
  nodes[dim_left_tree-1]->parent=nodes[dim_tree-1];
  fillBinaryTree(0,dim_left_tree-1,1,ns,nr_left_lvl);
  nodes[dim_tree-1]->children[0]=nodes[dim_left_tree-1];
  /* Right child */
  nodes[dim_tree-2]=new HSS_node<T,S>;
  nodes[dim_tree-2]->tree_ib=ns+1;
  nodes[dim_tree-2]->tree_ie=n;
  nodes[dim_tree-2]->parent=nodes[dim_tree-1];
  nodes[dim_tree-1]->children[1]=nodes[dim_tree-2];
  fillBinaryTree(dim_left_tree,dim_tree-2,ns+1,n,nr_right_lvl);

  setDefaultState();

}

template<typename T, typename S> void HSS_tree<T,S>::generateCombTree(int m, int n) {
  /* Generate a "comb" tree with m levels whose leaves cluster [1 n].
   * The tree looks like:  o
   *                      / \
   *                     o   o
   *                    / \
   *                   o   o
   *                  / \
   *                 o   o
   * I.e., right nodes do not have children. Right nodes have an odd index
   * (in 0-based postorder numbering).
   * This mostly used for special experiments. By default, a complete
   * binary tree is preferred.
   */

  int i;
  int ib,ie,ih;

  if(m<3) {
    std::cout << "m should be at least 3" << std::endl;
    return;
  }

  a_dim=n;
  nr_lvl=m;
  dim_tree=2*m-1;

  nodes=new HSS_node<T,S>*[dim_tree];
  subtree=NULL;
  par_lvl=-1;
  par_nodes=NULL;
  nodes[dim_tree-1]=new HSS_node<T,S>;
  nodes[dim_tree-1]->tree_ib=1;
  nodes[dim_tree-1]->tree_ie=n;
  nodes[dim_tree-1]->parent=NULL;
  nodes[dim_tree-1]->nr_children=2;
  for(i=dim_tree-1;i>=2;i-=2) {
    ib=nodes[i]->tree_ib;
    ie=nodes[i]->tree_ie;
    ih=ib+(ie-ib)/2;
    /* i has two children with indices i-1 (right) and i-2 (left) */
    nodes[i-1]=new HSS_node<T,S>;
    nodes[i-1]->nr_children=0;
    nodes[i-1]->children=NULL;
    nodes[i-1]->parent=nodes[i];
    nodes[i-1]->tree_ib=ih+1;
    nodes[i-1]->tree_ie=ie;
    nodes[i-2]=new HSS_node<T,S>;
    nodes[i-2]->nr_children=i==2?0:2;
    nodes[i-2]->children=NULL;
    nodes[i-2]->parent=nodes[i];
    nodes[i-2]->tree_ib=ib;
    nodes[i-2]->tree_ie=ih;
    nodes[i]->children=new HSS_node<T,S>*[2];
    nodes[i]->children[0]=nodes[i-2];
    nodes[i]->children[1]=nodes[i-1];
  }

  setDefaultState();

}

template<typename T, typename S> void HSS_tree<T,S>::generateFromPartitionTree(Partition_tree *utree) {
  /* Generate a binary tree from a user-given tree */

  int i;
  int lindex;

  if(!utree->check()) {
    std::cout << "Something is wrong with the provided tree" << std::endl;
    // TODO add a state to prevent compression from starting
  }

  a_dim=utree->size; /* Size of the cluster at the root node */
  nr_lvl=utree->treeheight();
  dim_tree=utree->treesize();

  nodes=new HSS_node<T,S>*[dim_tree];
  subtree=NULL;
  par_lvl=-1;
  par_nodes=NULL;

  Partition_tree **porder=utree->postorder();
  int *lchildren=utree->leftchildren();
  int *rchildren=utree->rightchildren();

  /* Create nodes */
  for(i=0;i<dim_tree;i++)
    nodes[i]=new HSS_node<T,S>;

  /* Compute children pointers */
  for(i=0;i<dim_tree;i++)
    if(porder[i]->c[0]==NULL) {
      /* Leaf */
      nodes[i]->nr_children=0;
      nodes[i]->children=NULL;
    } else {
      /* Interior node */
      nodes[i]->nr_children=2;
      nodes[i]->children=new HSS_node<T,S>*[2];
      nodes[i]->children[0]=nodes[lchildren[i]];
      nodes[i]->children[1]=nodes[rchildren[i]];
    }

  /* Compute parent pointers */
  for(i=0;i<dim_tree;i++)
    if(nodes[i]->nr_children) {
      nodes[i]->children[0]->parent=nodes[i];
      nodes[i]->children[1]->parent=nodes[i];
    }

  /* Compute tree_ib, tree_ie at leaves */
  nodes[0]->tree_ib=1;
  nodes[0]->tree_ie=porder[0]->size;
  lindex=nodes[0]->tree_ie;
  for(i=1;i<dim_tree;i++)
    if(!nodes[i]->nr_children) {
      /* Leaf */
      nodes[i]->tree_ib=lindex+1;
      nodes[i]->tree_ie=porder[i]->size+nodes[i]->tree_ib-1;
      lindex=nodes[i]->tree_ie;
    }

  /* Compute tree_ib, tree_ie at interior nodes */
  for(i=0;i<dim_tree;i++)
    if(nodes[i]->nr_children) {
      /* Interior node */
      nodes[i]->tree_ib=nodes[i]->children[0]->tree_ib;
      nodes[i]->tree_ie=nodes[i]->children[1]->tree_ie;
    }

  setDefaultState();

}

template<typename T, typename S> void HSS_tree<T,S>::print() {
  int i, j;

  std::cout << nr_lvl << " level(s), " << dim_tree << " node(s)" << std::endl << std::endl;

  std::cout << "Subset(s) in postorder:" << std::endl;
  for(i=0;i<dim_tree;i++) {
    std::cout << "Node " << i << " = [" << nodes[i]->tree_ib << " " << nodes[i]->tree_ie << "] has " << nodes[i]->nr_children << " children" << std::endl;
    for(j=0;j<nodes[i]->nr_children;j++)
      std::cout << "  [" << nodes[i]->children[j]->tree_ib << " " << nodes[i]->children[j]->tree_ie << "]" << std::endl;
    std::cout << std::endl;
  }
  for(i=0;i<dim_tree;i++) {
    if(nodes[i]->procs[0]>=0) {
      if(i==0) std::cout << "Mapping in postorder:" << std::endl;
    std::cout << "Node " << i << " is mapped on procs " << nodes[i]->procs[0] << "..." << nodes[i]->procs[1];
    if(nodes[i]->procs[0]!=nodes[i]->procs[1])
      std::cout << "; grid: " << nodes[i]->nprow << " x " << nodes[i]->npcol << std::endl;
    else
      std::cout << std::endl;
    std::cout << std::endl;
    }
  }

}

template<typename T, typename S> void HSS_tree<T,S>::proportionalMapping(int np) {
  /* Proportional mapping of an HSS tree on np processes.
   * The mapping is done using a top-down traversal:
   *   All the np processes work at the root node.
   *   For any interior node, the set of processes working at that node is split into
   *   disjoint sets that are assigned to its children. The split is done according
   *   to weights based on the size of the clusters associated to children.
   *   Assume node i has 2 children c1 and c2. We denote w1=c1.tree_ie-c1.tree_ib+1 and
   *   w2=c2.tree_ie-c2.tree_ib+1. Assume i is mapped on p processes; child c1 gets
   *   w1/(w1+w2)*p processes and child c2 gets the other w2/(w1+w2)*p processes.
   *   This is based on the fact, assuming constant/low rank, the amount of work
   *   at leaf nodes is proportional to the size of the clusters. This strategy
   *   should balance work at leaf nodes and hopefully interior nodes.
   * At each parallel node, we setup a BLACS grid that we make as square as
   * possible; some processes might be left idle and this complicates the
   * code a little. Cf. HSSR_par_Compress for more detail.
   */

   int i, j, nprocs, tmp, myid, ierr;
   int *usermap;
   HSS_node<T,S> *child1, *child2;
   S w1, w2;

   ierr=MPI_Comm_rank(comm,&myid);

   /* Map the tree by traversing it in reverse postorder */
   if(dim_tree<=0)
     return;

   nodes[dim_tree-1]->procs[0]=0;
   nodes[dim_tree-1]->procs[1]=np-1;
   for(i=dim_tree-1;i>=0;i--) {
     if(nodes[i]->nr_children!=0) {
       child1=nodes[i]->children[0];
       child2=nodes[i]->children[1];
       nprocs=nodes[i]->procs[1]-nodes[i]->procs[0]+1;
       /* First proc of first child = first proc of parent */
       child1->procs[0]=nodes[i]->procs[0];
       /* Last proc of second child = last proc of parent */
       child2->procs[1]=nodes[i]->procs[1];
       /* The following assign the last process of child 1
        * and the first process of child 0.
        */
       if(nprocs==1) {
         /* The two children are mapped on the same processor.
          * They are part of a serial subtree.
          */
         child1->procs[1]=nodes[i]->procs[0];
         child2->procs[0]=nodes[i]->procs[0];
       } else if(nprocs==2) {
         /* Distribute the two processors to the two children
          * regardless of weights.
          */
         child1->procs[1]=nodes[i]->procs[0];
         child2->procs[0]=nodes[i]->procs[1];
       } else {
         /* General case: distribute the processors according
          * to weights w1 and w2. Denoting nprocs the number
          * of processes working at i (the parent node), each
          * child gets at least 1 process and at most nprocs-1
          * processes.
          */
         w1=static_cast<S>(child1->tree_ie-child1->tree_ib+1);
         w2=static_cast<S>(child2->tree_ie-child2->tree_ib+1);
#if defined(testMapping)
         /* Experimental, for comb trees */
         w1=16;
         if(i==2)
           w1=0.5;
         w2=48;
         if(i==2)
           w2=0.5;
#endif
         tmp=static_cast<int>(floor(w1/(w1+w2)*static_cast<S>(nprocs)));
         tmp=std::max(tmp,1);
         tmp=std::min(tmp,nprocs-1);
         child1->procs[1]=nodes[i]->procs[0]+tmp-1;
         child2->procs[0]=nodes[i]->procs[0]+tmp;
#if defined(debug)
         if(!myid)
           std::cout << "node " << i << "; weights=" << w1 << "," << w2 << "; proc splitting=" << child1->procs[1]-child1->procs[0]+1 << "," << child2->procs[1]-child2->procs[0]+1 << std::endl;
#endif
      }
    }
  }

  /* Find the local subtree of every process */
  subtree=new int[2*np];
  for(j=0;j<np;j++) {
    subtree[2*j]=-1;
    subtree[2*j+1]=-1;
    for(i=0;i<dim_tree;i++) {
      if(nodes[i]->procs[0]==j && nodes[i]->procs[1]==j) {
        if(subtree[2*j]==-1)
          subtree[2*j]=i;
        subtree[2*j+1]=i;
      }
    }
  }

  /* Find the list of parallel interior nodes of myid. */
  par_lvl=0;
  for(i=0;i<dim_tree;i++)
    if(   nodes[i]->procs[0]<=myid
       && nodes[i]->procs[1]>=myid
       && nodes[i]->procs[1]>nodes[i]->procs[0]
       && nodes[i]->nr_children!=0
      )
      par_lvl++;
  par_nodes=new int[par_lvl];

  par_lvl=0;
  for(i=0;i<dim_tree;i++)
    if(   nodes[i]->procs[0]<=myid
       && nodes[i]->procs[1]>=myid
       && nodes[i]->procs[1]>nodes[i]->procs[0]
       && nodes[i]->nr_children!=0
      ) {
      par_nodes[par_lvl]=i;
      par_lvl++;
    }

  /* Each parallel node gets a BLACS context for a grid
   * that we make as square as possible, and a another
   * context that simply contains all the processes working
   * at that node. */
  for(i=0;i<dim_tree;i++) {
    nprocs=nodes[i]->procs[1]-nodes[i]->procs[0]+1;
    if(nprocs>1) {
      //blacs_get_(&IZERO,&IZERO,&nodes[i]->context);
      nodes[i]->context=Csys2blacs_handle(comm);

      nodes[i]->nprow=floor(sqrt((float)nprocs));
      nodes[i]->npcol=nprocs/nodes[i]->nprow;
      usermap=new int[nodes[i]->nprow*nodes[i]->npcol];
      for(j=0;j<nodes[i]->nprow*nodes[i]->npcol;j++)
        usermap[j]=nodes[i]->procs[0]+j;
      //blacs_gridmap_(&nodes[i]->context,usermap,&nodes[i]->nprow,&nodes[i]->nprow,&nodes[i]->npcol);
      Cblacs_gridmap(&nodes[i]->context,usermap,nodes[i]->nprow,nodes[i]->nprow,nodes[i]->npcol);
      delete[] usermap;

      tmp=nodes[i]->nprow;
      ierr=nodes[i]->npcol;
      blacs_gridinfo_(&nodes[i]->context,&tmp,&ierr,&nodes[i]->myrow,&nodes[i]->mycol);

      //blacs_get_(&IZERO,&IZERO,&nodes[i]->allctxt);
      nodes[i]->allctxt=Csys2blacs_handle(comm);
      usermap=new int[nprocs];
      for(j=0;j<nprocs;j++)
        usermap[j]=nodes[i]->procs[0]+j;
      //blacs_gridmap_(&nodes[i]->allctxt,usermap,&nprocs,&nprocs,&IONE);
      Cblacs_gridmap(&nodes[i]->allctxt,usermap,IONE,IONE,nprocs);
      delete[] usermap;
    }
  }

  /* Each root of a serial subtree also gets a context (with one proc)
   * and the nprow and npcol fields are set to 1
   * because it makes things easier in HSSR_par_Compress
   */
  for(j=0;j<np;j++) {
    i=subtree[2*j+1];
    if(i>=0) {
      //blacs_get_(&IZERO,&IZERO,&nodes[i]->context);
      nodes[i]->context=Csys2blacs_handle(comm);

      nodes[i]->nprow=IONE;
      nodes[i]->npcol=IONE;
      //blacs_gridmap_(&nodes[i]->context,&nodes[i]->procs[0],&IONE,&IONE,&IONE);
      Cblacs_gridmap(&nodes[i]->context,&nodes[i]->procs[0],IONE,IONE,IONE);

      tmp=nodes[i]->nprow;
      ierr=nodes[i]->npcol;
      blacs_gridinfo_(&nodes[i]->context,&tmp,&ierr,&nodes[i]->myrow,&nodes[i]->mycol);

      nodes[i]->allctxt=nodes[i]->context;
    }
  }

  /* Set isParLeaf for every node */
  for(i=0;i<dim_tree;i++)
    nodes[i]->isParLeaf = (
         (nodes[i]->procs[1]==nodes[i]->procs[0] && nodes[i]->nprow>0)
      || (nodes[i]->procs[1]> nodes[i]->procs[0] && nodes[i]->nr_children==0)
    );
}

template<typename T, typename S> double HSS_tree<T,S>::getSizeHSSFactors() {
  /* Compute the size of the HSS factors in the HSS_tree (which can be the
   * local part of a distributed tree).
   * For every node,
   *   U   is u_rows * u_rank words + (u_rows+u_rank) integers,
   *   V   is v_rows * v_rank words + (v_rows+v_rank) integers,
   *   D   is b*b words with b=tree_ie-tree_ib+1 if leaf,
   *   B12 is u_rank_child1*v_rank_child2 if non-leaf,
   *   B12 is u_rank_child1*v_rank_child2 if non-leaf,
   *   Ir  is u_rank integers,
   *   Ic  is v_rank integers.
   *
   *   U, V, D, B12, B21 are (reasonably) evenly distributed among
   *   nprow*npcol processes; proceses out of the grid don't
   *   own any piece.
   */

  int i, b, np, nptot, myid;
  HSS_node<T,S> *child1, *child2;
  double mem;

  MPI_Comm_rank(comm,&myid);
  nptot=nodes[dim_tree-1]->procs[1]-nodes[dim_tree-1]->procs[0]+1;

  mem=0;
  for(i=0;i<dim_tree;i++) {
    if(nodes[i]->procs[0]==nodes[i]->procs[1])
      np=1;
    else
      np=nodes[i]->nprow*nodes[i]->npcol;
    if(nptot==1 || (myid>=nodes[i]->procs[0] && myid<nodes[i]->procs[0]+np)) {
      /* The process owns a piece of U, V, D, B12, B21 */
      if(i<dim_tree-1) {
        /* U_E */
        mem+=nodes[i]->u_rows*nodes[i]->u_rank*sizeof(T)/np;
        /* V_E */
        mem+=nodes[i]->v_rows*nodes[i]->v_rank*sizeof(T)/np;
        /* Ir, U_perm */
        mem+=(2*nodes[i]->u_rank+nodes[i]->u_rows)*sizeof(int);
        /* Ic, V_perm */
        mem+=(2*nodes[i]->v_rank+nodes[i]->v_rows)*sizeof(int);
      }
      if(nodes[i]->nr_children==0) {
        b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
        mem+=b*b*sizeof(T)/np;
      } else {
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        /* B12 */
        mem+=child1->u_rank*child2->v_rank*sizeof(T)/np;
        /* B21 */
        mem+=child2->u_rank*child2->v_rank*sizeof(T)/np;
      }
    } else if(myid>=nodes[i]->procs[0]+np && myid<=nodes[i]->procs[1]) {
      if(i<dim_tree-1) {
        /* Ir, U_perm */
        mem+=(2*nodes[i]->u_rank+nodes[i]->u_rows)*sizeof(int);
        /* Ic, V_perm */
        mem+=(2*nodes[i]->v_rank+nodes[i]->v_rows)*sizeof(int);
      }
    }
  }

  return mem;

}

template<typename T, typename S> double HSS_tree<T,S>::getSizeULVFactors(int partial) {
  /* Compute the size of the ULV factors in the HSS_tree (which can be the
   * local part of a distributed tree).
   * For every node,
   *   U   is u_rows * u_rank words + (u_rows+u_rank) integers,
   *   V   is v_rows * v_rank words + (v_rows+v_rank) integers,
   *   B12 is u_rank_child1*v_rank_child2 if non-leaf,
   *   B12 is u_rank_child1*v_rank_child2 if non-leaf,
   *   Vt  is (u_rows+u_rank) * v_rank words,
   *   Q   is (u_rows+u_rank)*(u_rows+u_rank) words,
   *   W   is (u_rows+u_rank)*(u_rows+u_rank) words,
   *   L   is (u_rows*u_rows) except for the root node where it is
   *       (u_rows+u_rank)*(u_rows+u_rank) words + (u_rows+u_rank) integers.
   *
   *   The different generators are (reasonably) evenly distributed among
   *   nprow*npcol processes; proceses out of the grid don't
   *   own any piece.
   *
   *   When "partial" is nonzero, only nodes in the leftmost subtree
   *   have the Vt, Q, W, L generators.
   */

  int i, last, b, np, nptot, myid;
  HSS_node<T,S> *child1, *child2;
  double mem;

  MPI_Comm_rank(comm,&myid);
  nptot=nodes[dim_tree-1]->procs[1]-nodes[dim_tree-1]->procs[0]+1;

  if(partial)
    last=knode;
  else
    last=dim_tree-1;

  mem=0;
  for(i=0;i<dim_tree;i++) {
    if(nodes[i]->procs[0]==nodes[i]->procs[1])
      np=1;
    else
      np=nodes[i]->nprow*nodes[i]->npcol;
    if(nptot==1 || (myid>=nodes[i]->procs[0] && myid<nodes[i]->procs[0]+np)) {
      /* The process owns a piece of U, V, B12, B21, Vt, Q, W, L */
      if(i!=dim_tree-1) {
        /* U_E */
        mem+=nodes[i]->u_rows*nodes[i]->u_rank*sizeof(T)/np;
        /* V_E */
        mem+=nodes[i]->v_rows*nodes[i]->v_rank*sizeof(T)/np;
        /* Ir, U_perm */
        mem+=(2*nodes[i]->u_rank+nodes[i]->u_rows)*sizeof(int);
        /* Ic, V_perm */
        mem+=(2*nodes[i]->v_rank+nodes[i]->v_rows)*sizeof(int);
        if(i<last) {
          /* Q + W */
          b=nodes[i]->u_rows+nodes[i]->u_rank;
          mem+=2*b*b*sizeof(T)/np;
          /* L */
          mem+=nodes[i]->u_rows*nodes[i]->u_rows*sizeof(T)/np;
        }
      } else {
        /* L */
        if(nodes[i]->nr_children==0) {
          /* Tree with 1 node */
          b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
        } else {
          child1=nodes[i]->children[0];
          child2=nodes[i]->children[1];
          b=child1->u_rank+child2->u_rank;
        }
        mem+=b*(b+1)*sizeof(T)/np;
      }
      if(nodes[i]->nr_children!=0) {
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        /* B12 */
        mem+=child1->u_rank*child2->v_rank*sizeof(T)/np;
        /* B21 */
        mem+=child2->u_rank*child2->v_rank*sizeof(T)/np;
      }
    } else if(myid>=nodes[i]->procs[0]+np && myid<=nodes[i]->procs[1]) {
      if(i!=dim_tree-1) {
        /* Ir, U_perm */
        mem+=(2*nodes[i]->u_rank+nodes[i]->u_rows)*sizeof(int);
        /* Ic, V_perm */
        mem+=(2*nodes[i]->v_rank+nodes[i]->v_rows)*sizeof(int);
      }
    }
  }

  return mem;

}

template<typename T, typename S> double HSS_tree<T,S>::getSizeSchurUpdate() {
  /* The low-rank Schur complement update is the product theta Vt_k^T psi,
   * stored as three separate matrices, thus the memory requirements are
   * rtheta * ctheta + rpsi * ctheta + rpsi * cpsi.
   */

  int i, np, nptot, myid;
  double mem;

  MPI_Comm_rank(comm,&myid);
  nptot=nodes[dim_tree-1]->procs[1]-nodes[dim_tree-1]->procs[0]+1;

  mem=0;
  i=dim_tree-1;
  if(nodes[i]->procs[0]==nodes[i]->procs[1])
    np=1;
  else
    np=nodes[i]->nprow*nodes[i]->npcol;
  if(nptot==1 || (myid>=nodes[i]->procs[0] && myid<nodes[i]->procs[0]+np)) {
    /* theta */
    mem+=rtheta*ctheta*sizeof(T)/np;
    /* Vt_k */
    mem+=rpsi*ctheta*sizeof(T)/np;
    /* psi */
    mem+=rpsi*cpsi*sizeof(T)/np;
  }

  return mem;

}

inline double costID(int m, int n, int r) {
  /* Flop count for ID of an m x n matrix with returned rank r */
  double dr,dm,dn;

  dm=static_cast<double>(m);
  dn=static_cast<double>(n);
  dr=static_cast<double>(r);

  return 4*dr*dm*dn-2*dr*dr*(dm+dn)+4*dr*dr*dr/3+dr*dr*(dn-dr);
}

inline double costGEMM(int m, int n, int k, int alpha, int beta) {
  /* Flop count for alpha*A*B+beta*C with C m x n, k col/row dimension of A/B. */
  double dm,dn,dk;

  dm=static_cast<double>(m);
  dn=static_cast<double>(n);
  dk=static_cast<double>(k);
  return (alpha!=0)            *dm*dn*(2*dk-1) /* matrix product */
        +(alpha!=0 && beta!=0) *dm*dn          /* matrix sum */
        +(alpha!=0 && alpha!=1)*dm*dn          /* alpha scaling */
        +(beta!=0 && beta!=1)  *dm*dn;         /* beta scaling */
}

inline double costLQ(int m, int n) {
  /* Flop count for an LQ factorization */
  double dm,dn;

  dm=static_cast<double>(m);
  dn=static_cast<double>(n);

  return (m>n?dn*(2*dn*(0.5-(1./3.)*dn+dm)+dm+2*17./3.):dm*(2*dm*(-0.5-(1./3.)*dm+dn)+4.*dn+17./3));
}

inline double costConstrQ(int m, int n, int k) {
  /* Flop count for construction of the explicit Q factor */
  double dm,dn,dk;

  dm=static_cast<double>(m);
  dn=static_cast<double>(n);
  dk=static_cast<double>(k);

  return (dk*(4.*dm*dn+3.*dn-dm-4./3.+dk*(4./3.*dk-2*(dm+dn)-1.)));
}

template<typename T, typename S> double HSS_tree<T,S>::getFlopsHSSRCompression(int d, bool dosamples) {
  /* Compute the flop count for the compression with randomized sampling
   * of the HSS_tree (which can be the local part of a distributed tree).
   * For every node, flops=
       2/4 GEMMs for sampling (if dosamples true)
       2 IDs
       2/4 GEMMs for update
       Flops are assumed to be evenly distributed among processes
       doing the same operation.
     Note that d is the final number of random vectors, but a specific
     node has been compressed with nodes[i]->numrand <=d.
    */
  int i, b, np, nptot, npA, myid;
  HSS_node<T,S> *child1, *child2;
  double flops;

  MPI_Comm_rank(comm,&myid);
  nptot=nodes[dim_tree-1]->procs[1]-nodes[dim_tree-1]->procs[0]+1;
  npA=nptot==1?1:nodes[dim_tree-1]->nprow*nodes[dim_tree-1]->npcol;
  flops=0;

  /* Computing samples; processes performing the product should be
   * the same as the one in the grid of the root node */
  if(myid<npA && dosamples)
    flops+=2*costGEMM(a_dim,d,a_dim,1,0)/npA;

  /* Simulating the HSS construction */
  for(i=0;i<dim_tree-1;i++) {
    /* No flops at root node */
    if(nodes[i]->procs[0]==nodes[i]->procs[1])
      np=1;
    else
      np=nodes[i]->nprow*nodes[i]->npcol;
    if(nptot==1 || (myid>=nodes[i]->procs[0] && myid<nodes[i]->procs[0]+np)) {
      /* The process works at i */

      /* Sample */
      if(nodes[i]->nr_children==0) {
        b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
        flops+=2*costGEMM(b,d,b,-1,1)/np;
      } else {
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        flops+=costGEMM(child1->u_rank,d,child2->v_rank,-1,1)/np;
        flops+=costGEMM(child2->u_rank,d,child1->v_rank,-1,1)/np;
        flops+=costGEMM(child1->v_rank,d,child2->u_rank,-1,1)/np;
        flops+=costGEMM(child2->v_rank,d,child1->u_rank,-1,1)/np;
      }

      /* ID */
      if(nodes[i]->nr_children==0) {
        b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
        flops+=costID(nodes[i]->numrand,b,nodes[i]->u_rank)/np;
        flops+=costID(nodes[i]->numrand,b,nodes[i]->v_rank)/np;
      } else {
        flops+=costID(nodes[i]->numrand,child1->u_rank+child2->u_rank,nodes[i]->u_rank)/np;
        flops+=costID(nodes[i]->numrand,child1->v_rank+child2->v_rank,nodes[i]->v_rank)/np;
      }

      /* Update */
      if(nodes[i]->nr_children==0) {
        b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
        flops+=costGEMM(nodes[i]->v_rank,d,b-nodes[i]->v_rank,1,1)/np;
        flops+=costGEMM(nodes[i]->u_rank,d,b-nodes[i]->u_rank,1,1)/np;
      } else {
        flops+=costGEMM(nodes[i]->v_rank,d,child1->v_rank+child2->v_rank-nodes[i]->v_rank,1,1)/np;
        flops+=costGEMM(nodes[i]->u_rank,d,child1->u_rank+child2->u_rank-nodes[i]->u_rank,1,1)/np;
      }
    }
  }

  return flops;
}

template<typename T, typename S> double HSS_tree<T,S>::getFlopsULVFactorization(int partial) {
  /* Compute the flop count the ULV factorization of the HSS_tree
   * (which can be the local part of a distributed tree).
   * For every node, flops=
       2 GEMMs to get D
       2 GEMMS to get Vt
       1 LU at root node
       1 GEMM to get W
       1 LQ + reconstruction of the factor to get Q and L
       1 GEMM to get Vt
       1 GEMM to get Dt
       2/4 GEMMs for update
       Flops are assumed to be evenly distributed among processes
       doing the same operation.

      If "partial" is non zero, traverse only the leftmost subtree.
    */
  int i, last, b, np, nptot, myid;
  HSS_node<T,S> *child1, *child2;
  double flops;

  MPI_Comm_rank(comm,&myid);
  nptot=nodes[dim_tree-1]->procs[1]-nodes[dim_tree-1]->procs[0]+1;
  flops=0;

  if(partial)
    last=knode;
  else
    last=dim_tree-1;

  /* Simulating the ULV factorization */
  for(i=0;i<=last;i++) {
    if(nodes[i]->procs[0]==nodes[i]->procs[1])
      np=1;
    else
      np=nodes[i]->nprow*nodes[i]->npcol;
    if(nptot==1 || (myid>=nodes[i]->procs[0] && myid<nodes[i]->procs[0]+np)) {
      /* The process works at i */
      if(nodes[i]->nr_children!=0) {
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        flops+=costGEMM(child1->u_rank,child2->u_rank,child2->v_rank,1,0)/np;
        flops+=costGEMM(child2->u_rank,child1->u_rank,child1->v_rank,1,0)/np;
        if(i!=dim_tree-1) {
          flops+=costGEMM(child1->u_rank,nodes[i]->v_rank,child1->v_rank,1,0)/np;
          flops+=costGEMM(child2->u_rank,nodes[i]->v_rank,child2->v_rank,1,0)/np;
        }
      }
      if(i==last) {
        if(nodes[i]->nr_children==0) {
          /* Tree with 1 node */
          b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
        } else {
          child1=nodes[i]->children[0];
          child2=nodes[i]->children[1];
          b=child1->u_rank+child2->u_rank;
        }
        flops+=2.0/3.0*std::pow(static_cast<double>(b),3.0)/np;
      } else {
        b=nodes[i]->u_rows+nodes[i]->u_rank;
        /* W */
        flops+=costGEMM(nodes[i]->u_rows,b,nodes[i]->u_rank,-1,1)/np;
        /* L,Q */
        flops+=costLQ(nodes[i]->u_rows,b)/np;
        flops+=costConstrQ(b,b,nodes[i]->u_rows)/np;
        /* Vt */
        flops+=costGEMM(b,nodes[i]->v_rank,b,1,0)/np;
        /* Dt */
        flops+=costGEMM(nodes[i]->u_rank,nodes[i]->u_rank,b,1,0)/np;
      }
    }
  }

  return flops;
}

template<typename T, typename S> double HSS_tree<T,S>::getFlopsSolve(int d) {
  /* Compute the flop count for the triangular solution with d RHS
   * Bottom-up traversal: for every node, flops=
       6 GEMMs to get f
       1 TRSM at the root node
       1 GEMM to get ft
       1 TRSM to get y
       1 GEMM to get z, plus an extra for non leaves
     Top-down traversal: for every node, flops=
       1 GEMM to get x1, 1 GEMM to get x2
     Flops are assumed to be evenly distributed among processes
     doing the same operation.
    */
  int i, b, br1, br2, np, nptot, myid;
  HSS_node<T,S> *child1, *child2;
  double flops;

  MPI_Comm_rank(comm,&myid);
  nptot=nodes[dim_tree-1]->procs[1]-nodes[dim_tree-1]->procs[0]+1;
  flops=0;

  /* Simulating the solution: bottom-up traversal */
  for(i=0;i<dim_tree;i++) {
    if(nodes[i]->procs[0]==nodes[i]->procs[1])
      np=1;
    else
      np=nodes[i]->nprow*nodes[i]->npcol;
    if(nptot==1 || (myid>=nodes[i]->procs[0] && myid<nodes[i]->procs[0]+np)) {
      /* The process works at i */
      if(nodes[i]->nr_children!=0) {
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        br1=child1->u_rows+child1->u_rank;
        br2=child2->u_rows+child2->u_rank;
        /* f */
        flops+=costGEMM(br1,d,child1->u_rows,1,0)/np;
        flops+=costGEMM(child1->u_rank,d,br1,-1,1)/np;
        flops+=costGEMM(child1->u_rank,d,child2->v_rank,-1,1)/np;
        flops+=costGEMM(br2,d,child2->u_rows,1,0)/np;
        flops+=costGEMM(child2->u_rank,d,br2,-1,1)/np;
        flops+=costGEMM(child2->u_rank,d,child1->v_rank,-1,1)/np;
      }
      if(i==dim_tree-1) {
        /* x */
        if(nodes[i]->nr_children==0) {
          /* Tree with 1 node */
          b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
        } else {
          child1=nodes[i]->children[0];
          child2=nodes[i]->children[1];
          b=child1->u_rank+child2->u_rank;
        }
        /* Two triangular solves */
        flops+=2.0*d*b*b/np;
      } else {
        /* ft */
        flops+=costGEMM(nodes[i]->u_rows,d,nodes[i]->u_rank,-1,1)/np;
        /* y */
        flops+=nodes[i]->u_rows*nodes[i]->u_rows*d/np;
        /* z */
        flops+=costGEMM(nodes[i]->v_rank,d,nodes[i]->u_rows,1,0)/np;
        if(nodes[i]->nr_children!=0) {
          flops+=costGEMM(nodes[i]->v_rank,d,nodes[i]->v_rows,1,1)/np;
        }
      }
    }
  }

  /* Simulating the solution: bottom-up traversal */
  for(i=dim_tree-1;i>=0;i--) {
    if(nodes[i]->procs[0]==nodes[i]->procs[1])
      np=1;
    else
      np=nodes[i]->nprow*nodes[i]->npcol;
    if(nptot==1 || (myid>=nodes[i]->procs[0] && myid<nodes[i]->procs[0]+np)) {
      /* The process works at i */
      if(nodes[i]->nr_children!=0) {
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        br1=child1->u_rows+child1->u_rank;
        br2=child2->u_rows+child2->u_rank;
        flops+=costGEMM(br1,d,br1,1,0)/np;
        flops+=costGEMM(br2,d,br2,1,0)/np;
      }
    }
  }

  return flops;
}

template<typename T, typename S> double HSS_tree<T,S>::getFlopsProduct(int d) {
  /* Compute the flop count for a matrix-vector product with d RHS
   * Bottom-up traversal: for every node, flops=
       1 GEMM to get y
     Top-down traversal: for every node, flops=
       2 GEMMs at leaves to get b
       3 GEMMS at non-leaves to get z1 z2
     Flops are assumed to be evenly distributed among processes
     doing the same operation.
   */
  int i, b, np, nptot, myid;
  HSS_node<T,S> *child1, *child2;
  double flops;

  MPI_Comm_rank(comm,&myid);
  nptot=nodes[dim_tree-1]->procs[1]-nodes[dim_tree-1]->procs[0]+1;
  flops=0;

  /* Simulating a product: bottom-up traversal */
  for(i=0;i<dim_tree;i++) {
    if(nodes[i]->procs[0]==nodes[i]->procs[1])
      np=1;
    else
      np=nodes[i]->nprow*nodes[i]->npcol;
    if(nptot==1 || (myid>=nodes[i]->procs[0] && myid<nodes[i]->procs[0]+np))
      /* The process works at i */
      flops+=costGEMM(nodes[i]->v_rank,d,nodes[i]->v_rows,1,1)/np;
  }

  /* Simulating a product: bottom-up traversal */
  for(i=dim_tree-1;i>=0;i--) {
    if(nodes[i]->procs[0]==nodes[i]->procs[1])
      np=1;
    else
      np=nodes[i]->nprow*nodes[i]->npcol;
    if(nptot==1 || (myid>=nodes[i]->procs[0] && myid<nodes[i]->procs[0]+np)) {
      /* The process works at i */
      if(nodes[i]->nr_children==0) {
        b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
        flops+=costGEMM(nodes[i]->u_rows,d,nodes[i]->u_rank,1,1)/np;
        flops+=costGEMM(b,d,b,1,1)/np;
      } else {
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        flops+=costGEMM(child1->u_rank,d,child2->v_rank,1,0)/np;
        flops+=costGEMM(child2->u_rank,d,child1->v_rank,1,0)/np;
        /* GEMM with special structure + addition into z1 z2 */
        flops+=costGEMM(nodes[i]->u_rows,d,nodes[i]->u_rank,1,1)/np;
        flops+=(nodes[i]->u_rows+nodes[i]->u_rank)*d/np;
      }
    }
  }

  return flops;

}

template<typename T, typename S> double HSS_tree<T,S>::getFlopsSchurUpdate() {
  /* Compute the flop count for computing the low-rank update defining
   * the Schur complement:
   * theta = U_q^big B_qk
   * psi = Utilda_k^-1 Ltilda_k^-1 Ptilda_k U_k B_kq V_q^big^*
   */
  int i, qnode, np, nptot, myid;
  double flops;

  qnode=dim_tree-2;

  MPI_Comm_rank(comm,&myid);
  nptot=nodes[dim_tree-1]->procs[1]-nodes[dim_tree-1]->procs[0]+1;
  flops=0;

  /* Simulating the Schur complement generation */
  for(i=knode+1;i<=qnode;i++) {
    if(nodes[i]->procs[0]==nodes[i]->procs[1])
      np=1;
    else
      np=nodes[i]->nprow*nodes[i]->npcol;
    if(nptot==1 || (myid>=nodes[i]->procs[0] && myid<nodes[i]->procs[0]+np)) {
      /* The process works at i */
      if(nodes[i]->nr_children!=0) {
        /* Computing Ubig from Ubig_ch1 and Ubig_ch2 */
        flops+=costGEMM(nodes[i]->tree_ie-nodes[i]->tree_ib+1,nodes[i]->u_rank,nodes[i]->u_rows+nodes[i]->u_rank,1,0)/np;
        /* Computing Vbig from Vbig_ch1 and Vbig_ch2 */
        flops+=costGEMM(nodes[i]->tree_ie-nodes[i]->tree_ib+1,nodes[i]->v_rank,nodes[i]->v_rows+nodes[i]->v_rank,1,0)/np;
      }
    }
  }

  /* Root node */
  i=dim_tree-1;
  if(nodes[i]->procs[0]==nodes[i]->procs[1])
    np=1;
  else
    np=nodes[i]->nprow*nodes[i]->npcol;
  if(nptot==1 || (myid>=nodes[i]->procs[0] && myid<nodes[i]->procs[0]+np)) {
    /* theta = U_q^big B_qk */
    flops+=costGEMM(rtheta,ctheta,nodes[qnode]->u_rank,1,0)/np;
    /* psi = [ Dtilda_k^-1 \ (U_k B_kq) ] V_q^big^* */
    flops+=costGEMM(nodes[knode]->u_rows,nodes[qnode]->v_rank,nodes[knode]->u_rank,1,1)/np;
    flops+=2*std::pow(static_cast<double>(nodes[knode]->u_rows+nodes[knode]->u_rank),2)*nodes[qnode]->v_rank;
    flops+=costGEMM(rpsi,cpsi,nodes[qnode]->v_rank,1,1)/np;
  }

  return flops;
}

template<typename T, typename S> void HSS_tree<T,S>::writeStats(void) {
  int myid, np;
  int i, j;
  int nfields;
  int ninfo, ninfos;
  int *parent;
  int *counts, *offsets;
  int *info, *infos;
  FILE *fp;

  MPI_Comm_rank(comm,&myid);
  MPI_Comm_size(comm,&np);

  /* Get parent ids from parent pointers */
  parent=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    parent[i]=-1;
    for(j=0;j<dim_tree;j++) {
      if(nodes[j]==nodes[i]->parent) {
        parent[i]=j;
        break;
      }
    }
  }

  /* Pack in a vector, for all the nodes for which I am the master:
   *   - node id
   *   - parent
   *   - tree_ib, tree_ie
   *   - u_rows, u_rank
   *   - v_rows, v_rank
   *   - numrand
   * => 9 integers */
  nfields=9;
  ninfo=0;
  for(i=0;i<dim_tree;i++) {
    if(np==1 || myid==nodes[i]->procs[0])
      ninfo++;
  }
  ninfo*=nfields;
  info=new int[ninfo];
  j=0;
  for(i=0;i<dim_tree;i++) {
    if(np==1 || myid==nodes[i]->procs[0]) {
      info[j++]=i+1;
      info[j++]=parent[i]+1;
      info[j++]=nodes[i]->tree_ib;
      info[j++]=nodes[i]->tree_ie;
      info[j++]=nodes[i]->u_rows;
      info[j++]=nodes[i]->u_rank;
      info[j++]=nodes[i]->v_rows;
      info[j++]=nodes[i]->v_rank;
      info[j++]=nodes[i]->numrand;
    }
  }
  delete[] parent;

  /* Gather counts */
  if(!myid)
    counts=new int[np];
  MPI_Gather((void *)&ninfo,IONE,MPI_INTEGER,(void *)counts,IONE,MPI_INTEGER,IZERO,comm);

  if(!myid) {
    ninfos=0;
    for(i=0;i<np;i++)
      ninfos+=counts[i];
    infos=new int[ninfos];
    offsets=new int[np]();
    for(i=1;i<np;i++)
      offsets[i]=counts[i-1]+offsets[i-1];
  }

  MPI_Gatherv((void *)info,ninfo,MPI_INTEGER,(void *)infos,counts,offsets,MPI_INTEGER,0,comm);
  if(!myid) {
    fp=fopen("stats.m","w");
    fprintf(fp,"%% Every node of the tree is associated with a subdomain, a segment\n");
    fprintf(fp,"%% [ib:ie] of [1:n]. At every node i, there's:\n");
    fprintf(fp,"%%  - A D_i matrix, size (ie-ib+1) x (ie-ib+1) if i is a leaf.\n");
    fprintf(fp,"%%  - A B_{c1,c2} and a B_{c2,c1} if i is a non leaf (parent of c1, c2).\n");
    fprintf(fp,"%%    B_{c1,c2} has size c1.urank x c2.vrank and B_{c2,c1} has size\n");
    fprintf(fp,"%%    c2.urank x c2.vrank.\n");
    fprintf(fp,"%%  - A U_i and a V_i. They have the special structure\n");
    fprintf(fp,"%%    U=P*[I;E] with P a permutation matrix, I = identity, and E dense\n");
    fprintf(fp,"%%    matrix. So all we store is a vector for P and a u_rows x u_rank\n");
    fprintf(fp,"%%    matrix for E, but remember that the size of U is actually\n");
    fprintf(fp,"%%    (u_rows+u_rank) x u_rank.  Same thing for V.\n");
    fprintf(fp,"%% Note that at the root there's no U/V, only B_{c1,c2} and B_{c2,c1}.\n");
    fprintf(fp,"%% This is for the HSS compression/representation. For ULV, there are\n");
    fprintf(fp,"%% additional fields:\n");
    fprintf(fp,"%%  - Vt of size (u_rank+u_rows) x v_rank.\n");
    fprintf(fp,"%%  - W of size (u_rank+u_rows) x (u_rank+u_rows).\n");
    fprintf(fp,"%%  - Q of size (u_rank+u_rows) x (u_rank+u_rows).\n");
    fprintf(fp,"%%  - L of size u_rows x u_rows.\n");
    fprintf(fp,"%%    (expect at root: (u_rank+u_rows) x (u_rank+u_rows)).\n");
    fprintf(fp,"%%\n");
    fprintf(fp,"data = [ ...\n");
    for(i=0;i<ninfos/nfields;i++) {
      for(j=0;j<nfields;j++)
         fprintf(fp," %6d",infos[i*nfields+j]);
      fprintf(fp,"\n");
    }
    fprintf(fp,"] ;\n");
    fprintf(fp,"[s,idx]=sort(data(:,1));\n");
    fprintf(fp,"data=data(idx,:);\n");
    fclose(fp);
    delete[] infos;
  }
  delete[] info;
}

template<typename T, typename S> void HSS_tree<T,S>::cleanup(void) {
  int i;

  delete[] subtree;
  delete[] par_nodes;
  for(i=0;i<dim_tree;i++) {
    delete nodes[i];
  }
  delete[] nodes;
  delete[] theta;
  delete[] psi;
}

#endif // HSS_TREE_HPP
