/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef STRUMPACKDENSEPACKAGE_CPP_HPP
#define STRUMPACKDENSEPACKAGE_CPP_HPP

#if defined(WITH_PAPI)
#include <papi.h>
#endif

template<typename T, typename S> class StrumpackDensePackage {
public:

  /* User options */
  bool use_HSS;          /* Use HSS compression/factorization/solve or classical LU? */
  S    tol_HSS;          /* Tolerance for HSS compression */
  int  min_rand_HSS;     /* Minimum number of random vectors for HSS compression */
  int  max_rand_HSS;     /* Maximum number of random vectors for HSS compression */
  int  lim_rand_HSS;     /* Criterion of increasing the number of random vectors */
  int  inc_rand_HSS;     /* Number of extra random vectors at each increase */
  int  block_HSS;        /* Block size for the leaves of the HSS tree */
  int  levels_HSS;       /* Number of levels in the HSS tree */
  int  split_HSS;        /* Constrained partitioning of the root node */
  S    tol_IR;           /* Tolerance for iterative refinement */
  int  steps_IR;         /* Number of steps of iterative refinement */
  bool fast_IR;          /* Matvecs with HSS(A) or A in iterative refinement? */
  bool verbose;          /* Verbose or silent? */
  bool access_is_global; /* User-given access routine: requires global synchronization? */
  int  sched_strat;      /* Scheduling strategy: 1 or 2 */
  void *obj;             /* Handle to user data/object, for strategy 2 */
  Partition_tree *utree; /* User-given clustering */

  /* User routines */
  void compress(T *A, int *descA);
  void compress(T* A, int *descA, T *Rr, T *Rc, T* Sr, T* Sc, int *descRS);
  void compress(T *Rr, T*Rc, T*Sr, T* Sc, int *descRS, void (*elements)(void *, int *, int *, T *, int *));
  void compress(void (*elements)(void *, int *, int *, T *, int *),
		void (*Amult)(void *, T *, T *, int *, T *, T *, int *),
		int* descA);
  S check_compression(T *A, int *descA);
  void factor(T *A, int *descA);
  void partially_factor(T* A, int *descA, int nfact);
  void compute_schur();
  void reduce_RHS(T *X, int *descX, T* B, int *descB);
  void schur_product(char Trans, T alpha, T *X, int *descX, T beta, T *B, int *descB);
  void expand_solution(T *X, int *descX, T *B, int *descB);
  void solve(T *X, int *descX, T *B, int *descB);
  S check_solution(T *A, int *descA, T *X, int *descX, T *B, int *descB);
  S refine(T *A, int *descA, T *X, int *descX, T *B, int *descB);
  void product(char Trans, T alpha, T *A, int *descA, T *B, int *descB, T beta, T *C, int *descC);
  void extract(T *A, int *descA, T *B, int *descB, int *I, int nI, int *J, int nJ);
  void extract_schur(T *B, int *descB, int *I, int nI, int *J, int nJ);
  void destroy_A22();
  void print_statistics();

private:
  void print_min_ave_max(double, double, double);
  void print_min_ave_max_total(double, double, double, double);
  void print_call(double);

public:
  StrumpackDensePackage(MPI_Comm);
  ~StrumpackDensePackage();

  /* Version number */
  std::string version;

  /* States */
  bool compressed;             /* True if HSS compressed */
  bool factored_LU;            /* True if successfully LU-factored */
  bool factored_HSS;           /* True if successfully ULV-factored */
  bool partially_factored_LU;  /* True if succesfully partially factored (LU) */
  bool partially_factored_HSS; /* True if succesfully partially factored (HSS) */
  bool schur_LU;               /* True if exact Schur was computed */
  bool schur_HSS;              /* True is HSS Schur was computed */

  /* Statistics */
  int max_rank;                /* Maximum rank found in compression */
  int eff_rand_HSS;            /* Effective number of random vectors used in compression */
  double mem_compr;            /* Total memory used in compression */
  double mem_HSS;              /* Memory usage for the HSS representation only */
  double mem_fact;             /* Total memory used in factorization */
  double mem_ULV;              /* Memory usage for the LU/ULV factors only */
  double flops_compr;          /* Operation count of the compression */
  double flops_fact;           /* Operation count of the factorization */
  double flops_solve;          /* Cumulated flops in solution steps */
  double flops_prod;           /* Cumulated flops in matrix-vector products */
  double time_compr;           /* Time spent in compression */
  double time_fact;            /* Time spent in factorization */
  double time_solve;           /* Cumulated time in solution steps */
  double time_prod;            /* Cumulated time in matrix-vector products */

  /* Internal variables */
  gridinfo ginfo;              /* Grid info */
  int np;                      /* Communicator size */
  int n;                       /* Problem size */
  HSS_tree<T,S> htree;         /* Holds all the HSS stuff */
  T *LU;                       /* Traditional LU factors (if used) */
  int *descLU;                 /* (same) */
  int *piv;                    /* (same) */
  int var_schur;               /* First variable in Schur complement */

  /* Constants */
  T NONE;
  T ZERO;
  T  ONE;

  bool check_descriptor(std::string, int *, bool);

  /* Public for strategy 2 */
  T *A_in;                     /* Pointer to input matrix */
  int *descA_in;               /* (same) */
  MPI_Comm comm;               /* Global communicator */
  int ctxtglob;                /* Global grid */
};

template<typename T, typename S> StrumpackDensePackage<T,S>::StrumpackDensePackage(MPI_Comm user_comm) {
  //  int myid;
  /* Version */
  version = "1.1.1";

  /* Default parameters */
  use_HSS          = false;
  tol_HSS          = 1e-4;
  min_rand_HSS     = 16;
  max_rand_HSS     = 2048;
  lim_rand_HSS     = 10;
  inc_rand_HSS     = 32;
  block_HSS        = 16;
  levels_HSS       = 0;
  split_HSS        = 0;
  tol_IR           = 1e-14;
  steps_IR         = 10;
  fast_IR          = false;
  verbose          = true;
  access_is_global = true;
  sched_strat      = 1;
  obj              = NULL;
  utree            = NULL;

  /* Default state */
  NONE                   = -1;
  ZERO                   = 0;
  ONE                    = 1;
  compressed             = false;
  factored_LU            = false;
  factored_HSS           = false;
  partially_factored_LU  = false;
  partially_factored_HSS = false;
  schur_LU               = false;
  schur_HSS              = false;
  max_rank               = 0;
  eff_rand_HSS           = 0;
  mem_compr              = 0.0;
  mem_HSS                = 0.0;
  mem_fact               = 0.0;
  mem_ULV                = 0.0;
  flops_compr            = 0.0;
  flops_fact             = 0.0;
  flops_solve            = 0.0;
  flops_prod             = 0.0;
  time_compr             = 0.0;
  time_fact              = 0.0;
  time_solve             = 0.0;
  time_prod              = 0.0;
  comm                   = user_comm;
  LU                     = NULL;
  descLU                 = NULL;
  piv                    = NULL;

  MPI_Comm_size(comm,&np);
#if defined(WITH_PAPI)
  float rtime, ptime, mflops;
  long long flpops;
  PAPI_library_init(PAPI_VER_CURRENT);
  PAPI_flops(&rtime,&ptime,&flpops,&mflops);
#endif

  // MPI_Comm_rank(comm,&myid);
  // if(!myid) std::cout << "STRUMPACK Dense Package version " << version << ", ";

  // if(!myid && typeid(ONE)==typeid(0.0e0))
  //   std::cout << "real double precision arithmetic." << std::endl << std::endl;
  // if(!myid && typeid(ONE)==typeid(0.0f))
  //   std::cout << "real single precision arithmetic." << std::endl << std::endl;
  // if(!myid && typeid(ONE)==typeid((dcomplex){0,0}))
  //   std::cout << "complex double precision arithmetic." << std::endl << std::endl;
  // if(!myid && typeid(ONE)==typeid((scomplex){0,0}))
  //   std::cout << "complex single precision arithmetic." << std::endl << std::endl;

}

template<typename T, typename S> void access(void *, int *, int *, T *, int *);

template<typename T, typename S> void StrumpackDensePackage<T,S>::compress(T *A, int *descA) {
  /* HSS compression with explicit matrix and classical matrix-vector product */
  int ctxt;
  int nb;
  int myid;
  int myrow, mycol, nprow, npcol;
  int m;
  bool msg;
  double tstart, tend;
#if defined(WITH_PAPI)
  float rtime, ptime, mflops;
  long long flpops1, flpops2;
  double flpops, mi, ma, av;
#endif

  tstart=MPI_Wtime();

  MPI_Comm_rank(comm,&myid);
  msg=verbose && myid==0;

  if(!check_descriptor("A",descA,true)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }

  n=descA[BLACSn];
  ctxt=descA[BLACSctxt];
  nb=descA[BLACSnb];
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);

#if defined(MEMTRACK)
  MemoryTracker();
#endif

#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops1,&mflops);
#endif

  if(use_HSS) {
    if(compressed) {
      if(!myid) std::cout << "Matrix was previously compressed" << std::endl;
      if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }

    if(utree==NULL) {
      /* No user-given clustering */
      if(levels_HSS>0) {
        if(block_HSS>0)
          if(!myid) std::cout << "Warning: both block_HSS and levels_HSS are defined; block_HSS is ignored" << std::endl;
        if(split_HSS)
          m=std::min(levels_HSS,static_cast<int>(floor(log2((double)std::min(split_HSS,n-split_HSS)))+2e0));
        else
          m=std::min(levels_HSS,static_cast<int>(floor(log2((double)n))+1e0));
      } else {
        if(block_HSS<0) {
          if(!myid) std::cout << "levels_HSS=0 so block_size must be >0." << std::endl;
          if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
          return;
        }
        if(split_HSS)
          m=std::min(std::max(1,static_cast<int>(floor(log2((double)std::min(split_HSS,n-split_HSS)/double(block_HSS))+2e0))),static_cast<int>(floor(log2((double)n))+1e0));
        else
          m=std::min(std::max(1,static_cast<int>(floor(log2((double)n/double(block_HSS))+1e0))),static_cast<int>(floor(log2((double)n))+1e0));
      }
      if(split_HSS && m<=1)
        m=2;

      if(split_HSS) {
        /* Constrained tree */
        if(split_HSS<1 || split_HSS>n) {
          if(!myid) std::cout << "1 <= split_HSS < n must hold." << std::endl;
          if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
          return;
        }
        htree.generateConstrainedBinaryTree(m,split_HSS,n);
      } else {
        /* Unconstrained binary tree */
        htree.generateBinaryTree(m,n);
      }
    } else {
      /* Generate a tree from the user-given clustering */
      htree.generateFromPartitionTree(utree);
      m=htree.nr_lvl;
    }
    htree.comm=comm;
    htree.obj=obj;
    if(np==1) {
      /* Sequential compression */
      if(msg) std::cout << "Sequential HSS compression with " << m << " levels in the tree..." << std::endl;
#if defined(MEMTRACK)
      MemoryTracker::reset();
#endif
      htree.HSSR_seq_Compress(A,descA,tol_HSS,min_rand_HSS,max_rand_HSS,lim_rand_HSS,inc_rand_HSS,&eff_rand_HSS,&max_rank,0,NULL,NULL,NULL,NULL,NULL,NULL,NULL,0,false);

      /* Flops */
      flops_compr+=htree.getFlopsHSSRCompression(eff_rand_HSS,true);

      /* Memory: factors */
      mem_HSS+=htree.getSizeHSSFactors();

#if defined(MEMTRACK)
      /* Memory: highmark */
      mem_compr+=MemoryTracker::get();
#endif
    } else {
      /* Parallel compression */
      if(msg) std::cout << "Parallel HSS compression with " << m << " levels in the tree..." << std::endl;
      htree.proportionalMapping(np);
      ginfo.ctxt=ctxt;
      ginfo.nprow=nprow;
      ginfo.npcol=npcol;
      ginfo.myrow=myrow;
      ginfo.mycol=mycol;
      ginfo.mb=nb;
      ginfo.nb=nb;
#if defined(MEMTRACK)
      MemoryTracker::reset();
#endif
      if(sched_strat<1 || sched_strat>2) {
        if(!myid)
          std::cout << "Warning: sched_strat should be 1 or 2. Defaulted to 1." << std::endl;
        sched_strat=1;
      }
      switch(sched_strat) {
        case 1:
          /* Explicit matrix directly passed to compression routine. */
          htree.HSSR_par_Compress(A,descA,tol_HSS,min_rand_HSS,max_rand_HSS,lim_rand_HSS,inc_rand_HSS,&eff_rand_HSS,&max_rank,&ginfo,NULL,NULL,NULL,NULL,NULL,NULL,NULL,false);
          break;
        case 2:
          /* Explicit matrix passed to compression via an element access routine.
           * Every time the element access routine is called, its first
           * parameter is htree.obj = this = the StrumpackDensePackage object
           * calling the compression routine. This way, the compression routine
           * can access the input matrix, the global communicator and the global
           * grid.
           */
          htree.obj=this;
          A_in=A;
          descA_in=descA;
          ctxtglob=htree.nodes[htree.dim_tree-1]->allctxt;
          htree.HSSR_par_Compress(A,descA,tol_HSS,min_rand_HSS,max_rand_HSS,lim_rand_HSS,inc_rand_HSS,&eff_rand_HSS,&max_rank,&ginfo,NULL,NULL,NULL,NULL,NULL,access<T,S>,NULL,true);
          break;
      }
      /* Flops; remove the cost of computing S_r and S_c */
      flops_compr+=htree.getFlopsHSSRCompression(eff_rand_HSS,true);

      /* Memory: factors */
      mem_HSS+=htree.getSizeHSSFactors();

#if defined(MEMTRACK)
      /* Memory: highmark */
      mem_compr+=MemoryTracker::get();
#endif
    }
    compressed=true;
  } else {
    if(!myid) std::cout << "Warning; compression routine was called but HSS is off (use_HSS=false). Nothing to do." << std::endl << std::endl;
  }

  tend=MPI_Wtime();
#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops2,&mflops);
  flpops=static_cast<double>(flpops2-flpops1);
  MPI_Allreduce((void *) &flpops,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
  MPI_Allreduce((void *) &flpops,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
  MPI_Allreduce((void *) &flpops,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
  av/=static_cast<double>(np);
  if(msg) std::cout << "PAPI GFlop count: min= " << mi/1.0e9 << "; avg= " << av/1.0e9 << "; max= " << ma/1.0e9 << std::endl;
#endif
  if(msg) std::cout << "Done in " << tend-tstart << "s; max rank = " << max_rank << "; effective number of random vectors = " << eff_rand_HSS << std::endl << std::endl;
  time_compr+=tend-tstart;
#if defined(MEMTRACK)
  MemoryTracker::reset();
#endif
  MPI_Bcast((void *) &compressed,IONE,MPI_INTEGER,IZERO,comm);
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::compress(T* A, int *descA, T *Rr, T *Rc, T* Sr, T* Sc, int *descRS) {
  /* HSS compression with explicit matrix and user-given samples and random vectors*/
  int ctxt;
  int nb;
  int myid;
  int myrow, mycol, nprow, npcol;
  int m;
  bool msg;
  double tstart, tend;
  T *Rrcopy, *Rccopy, *Srcopy, *Sccopy;
#if defined(WITH_PAPI)
  float rtime, ptime, mflops;
  long long flpops1, flpops2;
  double flpops, mi, ma, av;
#endif

  tstart=MPI_Wtime();

  MPI_Comm_rank(comm,&myid);
  msg=verbose && myid==0;

  if(!check_descriptor("A",descA,true)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }
  if(!check_descriptor("R and S",descRS,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }
  if(descA[BLACSm]!=descRS[BLACSm]) {
      std::cout << "A, R and S have different row counts! A:" << descA[BLACSm] << "; R and S:" << descRS[BLACSm] << std::endl;
      std::cout << "Aborting..." << std::endl << std::endl;
      return;
  }
  if(descA[BLACSmb]!=descRS[BLACSmb]) {
    std::cout << "A, R and S have different block sizes! A:" << descA[BLACSmb] << "; R and S:" << descRS[BLACSmb] << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }

  n=descA[BLACSn];
  ctxt=descA[BLACSctxt];
  nb=descA[BLACSnb];
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);

#if defined(MEMTRACK)
  MemoryTracker();
#endif

#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops1,&mflops);
#endif

  if(use_HSS) {
    if(compressed) {
      if(!myid) std::cout << "Matrix was previously compressed" << std::endl;
      if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }

    if(utree==NULL) {
      /* No user-given clustering */
      if(levels_HSS>0) {
        if(block_HSS>0)
          if(!myid) std::cout << "Warning: both block_HSS and levels_HSS are defined; block_HSS is ignored" << std::endl;
        if(split_HSS)
          m=std::min(levels_HSS,static_cast<int>(floor(log2((double)std::min(split_HSS,n-split_HSS)))+2e0));
        else
          m=std::min(levels_HSS,static_cast<int>(floor(log2((double)n))+1e0));
      } else {
        if(block_HSS<0) {
          if(!myid) std::cout << "levels_HSS=0 so block_size must be >0." << std::endl;
          if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
          return;
        }
        if(split_HSS)
          m=std::min(std::max(1,static_cast<int>(floor(log2((double)std::min(split_HSS,n-split_HSS)/double(block_HSS))+2e0))),static_cast<int>(floor(log2((double)n))+1e0));
        else
          m=std::min(std::max(1,static_cast<int>(floor(log2((double)n/double(block_HSS))+1e0))),static_cast<int>(floor(log2((double)n))+1e0));
      }
      if(split_HSS && m<=1)
        m=2;
  
      if(split_HSS) {
        /* Constrained tree */
        if(split_HSS<1 || split_HSS>n) {
          if(!myid) std::cout << "1 <= split_HSS < n must hold." << std::endl;
          if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
          return;
        }
        htree.generateConstrainedBinaryTree(m,split_HSS,n);
      } else {
        /* Unconstrained binary tree */
        htree.generateBinaryTree(m,n);
      }
    } else {
      /* Generate a tree from the user-given clustering */
      htree.generateFromPartitionTree(utree);
      m=htree.nr_lvl;
    }
    htree.comm=comm;
    htree.obj=obj;
    if(min_rand_HSS>descRS[BLACSn]) {
      min_rand_HSS=descRS[BLACSn];
      if(!myid) std::cout << "Warning: min_rand_HSS was larger than the number of columns in R and R. min_rand_HSS reset to " << min_rand_HSS << std::endl;
    }
    //if(!myid) std::cout << "Warning: in this version, adaptive sampling is disabled when samples are provided. Only min_rand_HSS is used." << std::endl;
    if(np==1) {
      /* Sequential compression */
      if(msg) std::cout << "Sequential HSS compression with " << m << " levels in the tree..." << std::endl;
#if defined(MEMTRACK)
      MemoryTracker::reset();
#endif
      StrumpackRand rgen(time(NULL));
      /* The sequential compression routine has side effects on the random vectors and samples.
       * We make copies.
       */
      Rrcopy=new T[n*descRS[BLACSn]];
      lacpy('N',n,descRS[BLACSn],Rr,n,Rrcopy,n);
      Rccopy=new T[n*descRS[BLACSn]];
      lacpy('N',n,descRS[BLACSn],Rc,n,Rccopy,n);
      Srcopy=new T[n*descRS[BLACSn]];
      lacpy('N',n,descRS[BLACSn],Sr,n,Srcopy,n);
      Sccopy=new T[n*descRS[BLACSn]];
      lacpy('N',n,descRS[BLACSn],Sc,n,Sccopy,n);
      htree.HSSR_seq_Compress(A,descA,tol_HSS,min_rand_HSS,0,0,0,&eff_rand_HSS,&max_rank,0,Rrcopy,Rccopy,Srcopy,Sccopy,&rgen,NULL,NULL,0,false);
      delete[] Rrcopy;
      delete[] Rccopy;
      delete[] Srcopy;
      delete[] Sccopy;

      /* Flops */
      flops_compr+=htree.getFlopsHSSRCompression(eff_rand_HSS,false);

      /* Memory: factors */
      mem_HSS+=htree.getSizeHSSFactors();

#if defined(MEMTRACK)
      /* Memory: highmark */
      mem_compr+=MemoryTracker::get();
#endif
    } else {
      /* Parallel compression */
      if(msg) std::cout << "Parallel HSS compression with " << m << " levels in the tree..." << std::endl;
      htree.proportionalMapping(np);
      ginfo.ctxt=ctxt;
      ginfo.nprow=nprow;
      ginfo.npcol=npcol;
      ginfo.myrow=myrow;
      ginfo.mycol=mycol;
      ginfo.mb=nb;
      ginfo.nb=nb;
#if defined(MEMTRACK)
      MemoryTracker::reset();
#endif
      htree.HSSR_par_Compress(A,descA,tol_HSS,min_rand_HSS,0,0,0,&eff_rand_HSS,&max_rank,&ginfo,Rr,Rc,Sr,Sc,descRS,NULL,NULL,false);

      /* Flops */
      flops_compr+=htree.getFlopsHSSRCompression(eff_rand_HSS,false);

      /* Memory: factors */
      mem_HSS+=htree.getSizeHSSFactors();

#if defined(MEMTRACK)
      /* Memory: highmark */
      mem_compr+=MemoryTracker::get();
#endif
    }
    compressed=true;
  } else {
    if(!myid) std::cout << "Warning; compression routine was called but HSS is off (use_HSS=false). Nothing to do." << std::endl << std::endl;
  }

  tend=MPI_Wtime();
#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops2,&mflops);
  flpops=static_cast<double>(flpops2-flpops1);
  MPI_Allreduce((void *) &flpops,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
  MPI_Allreduce((void *) &flpops,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
  MPI_Allreduce((void *) &flpops,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
  av/=static_cast<double>(np);
  if(msg) std::cout << "PAPI GFlop count: min= " << mi/1.0e9 << "; avg= " << av/1.0e9 << "; max= " << ma/1.0e9 << std::endl;
#endif
  if(msg) std::cout << "Done in " << tend-tstart << "s; max rank = " << max_rank << "; effective number of random vectors = " << eff_rand_HSS << std::endl << std::endl;
  time_compr+=tend-tstart;
#if defined(MEMTRACK)
  MemoryTracker::reset();
#endif
  MPI_Bcast((void *) &compressed,IONE,MPI_INTEGER,IZERO,comm);
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::compress(T *Rr, T*Rc, T*Sr, T* Sc, int *descRS, void (*elements)(void *, int *, int *, T *, int *)) {
  /* HSS compression with user-given samples, random vectors, and selected access */
  int ctxt;
  int nb;
  int myid;
  int myrow, mycol, nprow, npcol;
  int m;
  bool msg;
  double tstart, tend;
  T *Rrcopy, *Rccopy, *Srcopy, *Sccopy;
#if defined(WITH_PAPI)
  float rtime, ptime, mflops;
  long long flpops1, flpops2;
  double flpops, mi, ma, av;
#endif

  tstart=MPI_Wtime();

  MPI_Comm_rank(comm,&myid);
  msg=verbose && myid==0;

  if(!check_descriptor("R and S",descRS,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }

  n=descRS[BLACSm];
  ctxt=descRS[BLACSctxt];
  nb=descRS[BLACSnb];
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);

#if defined(MEMTRACK)
  MemoryTracker();
#endif

#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops1,&mflops);
#endif

  if(use_HSS) {
    if(utree==NULL) {
      /* No user-given clustering */
      if(compressed) {
        if(!myid) std::cout << "Matrix was previously compressed" << std::endl;
        if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
        return;
      }

      if(levels_HSS>0) {
        if(block_HSS>0)
          if(!myid) std::cout << "Warning: both block_HSS and levels_HSS are defined; block_HSS is ignored" << std::endl;
        if(split_HSS)
          m=std::min(levels_HSS,static_cast<int>(floor(log2((double)std::min(split_HSS,n-split_HSS)))+2e0));
        else
          m=std::min(levels_HSS,static_cast<int>(floor(log2((double)n))+1e0));
      } else {
        if(block_HSS<0) {
          if(!myid) std::cout << "levels_HSS=0 so block_size must be >0." << std::endl;
          if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
          return;
        }
        if(split_HSS)
          m=std::min(std::max(1,static_cast<int>(floor(log2((double)std::min(split_HSS,n-split_HSS)/double(block_HSS))+2e0))),static_cast<int>(floor(log2((double)n))+1e0));
        else
          m=std::min(std::max(1,static_cast<int>(floor(log2((double)n/double(block_HSS))+1e0))),static_cast<int>(floor(log2((double)n))+1e0));
      }
      if(split_HSS && m<=1)
        m=2;

      if(split_HSS) {
        /* Constrained tree */
        if(split_HSS<1 || split_HSS>n) {
          if(!myid) std::cout << "1 <= split_HSS < n must hold." << std::endl;
          if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
          return;
        }
        htree.generateConstrainedBinaryTree(m,split_HSS,n);
      } else {
        /* Unconstrained binary tree */
        htree.generateBinaryTree(m,n);
      }
    } else {
      /* Generate a tree from the user-given clustering */
      htree.generateFromPartitionTree(utree);
      m=htree.nr_lvl;
    }
    htree.comm=comm;
    htree.obj=obj;
    if(min_rand_HSS>descRS[BLACSn]) {
      min_rand_HSS=descRS[BLACSn];
      if(!myid) std::cout << "Warning: min_rand_HSS was larger than the number of columns in R and R. min_rand_HSS reset to " << min_rand_HSS << std::endl;
    }
    //if(!myid) std::cout << "Warning: in this version, adaptive sampling is disabled when samples are provided. Only min_rand_HSS is used." << std::endl;
    if(np==1) {
      /* Sequential compression */
      if(msg) std::cout << "Sequential HSS compression with " << m << " levels in the tree..." << std::endl;
#if defined(MEMTRACK)
      MemoryTracker::reset();
#endif
      StrumpackRand rgen(time(NULL));
      /* The sequential compression routine has side effects on the random vectors and samples.
       * We make copies.
       */
      Rrcopy=new T[n*descRS[BLACSn]];
      lacpy('N',n,descRS[BLACSn],Rr,n,Rrcopy,n);
      Rccopy=new T[n*descRS[BLACSn]];
      lacpy('N',n,descRS[BLACSn],Rc,n,Rccopy,n);
      Srcopy=new T[n*descRS[BLACSn]];
      lacpy('N',n,descRS[BLACSn],Sr,n,Srcopy,n);
      Sccopy=new T[n*descRS[BLACSn]];
      lacpy('N',n,descRS[BLACSn],Sc,n,Sccopy,n);
      htree.HSSR_seq_Compress(NULL,NULL,tol_HSS,min_rand_HSS,0,0,0,&eff_rand_HSS,&max_rank,0,Rrcopy,Rccopy,Srcopy,Sccopy,&rgen,elements,NULL,0,false);
      delete[] Rrcopy;
      delete[] Rccopy;
      delete[] Srcopy;
      delete[] Sccopy;

      /* Flops */
      flops_compr+=htree.getFlopsHSSRCompression(eff_rand_HSS,false);

      /* Memory: factors */
      mem_HSS+=htree.getSizeHSSFactors();

#if defined(MEMTRACK)
      /* Memory: highmark */
      mem_compr+=MemoryTracker::get();
#endif
    } else {
      /* Parallel compression */
      if(msg) std::cout << "Parallel HSS compression with " << m << " levels in the tree..." << std::endl;
      htree.proportionalMapping(np);
      ginfo.ctxt=ctxt;
      ginfo.nprow=nprow;
      ginfo.npcol=npcol;
      ginfo.myrow=myrow;
      ginfo.mycol=mycol;
      ginfo.mb=nb;
      ginfo.nb=nb;
#if defined(MEMTRACK)
      MemoryTracker::reset();
#endif
      htree.HSSR_par_Compress(NULL,NULL,tol_HSS,min_rand_HSS,0,0,0,&eff_rand_HSS,&max_rank,&ginfo,Rr,Rc,Sr,Sc,descRS,elements,NULL,access_is_global);

      /* Flops */
      flops_compr+=htree.getFlopsHSSRCompression(eff_rand_HSS,false);

      /* Memory: factors */
      mem_HSS+=htree.getSizeHSSFactors();

#if defined(MEMTRACK)
      /* Memory: highmark */
      mem_compr+=MemoryTracker::get();
#endif
    }
    compressed=true;
  } else {
    if(!myid) std::cout << "Warning; compression routine was called but HSS is off (use_HSS=false). Nothing to do." << std::endl << std::endl;
  }

  tend=MPI_Wtime();
#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops2,&mflops);
  flpops=static_cast<double>(flpops2-flpops1);
  MPI_Allreduce((void *) &flpops,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
  MPI_Allreduce((void *) &flpops,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
  MPI_Allreduce((void *) &flpops,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
  av/=static_cast<double>(np);
  if(msg) std::cout << "PAPI GFlop count: min= " << mi/1.0e9 << "; avg= " << av/1.0e9 << "; max= " << ma/1.0e9 << std::endl;
#endif
  if(msg) std::cout << "Done in " << tend-tstart << "s; max rank = " << max_rank << "; effective number of random vectors = " << eff_rand_HSS << std::endl << std::endl;
  time_compr+=tend-tstart;
#if defined(MEMTRACK)
  MemoryTracker::reset();
#endif
  MPI_Bcast((void *) &compressed,IONE,MPI_INTEGER,IZERO,comm);
}

template<typename T, typename S> void
StrumpackDensePackage<T,S>::compress(void (*elements)(void *, int *, int *, T *, int *),
				     void (*Amult)(void *, T *, T *, int *, T *, T *, int *),
				     int* descA) {
  /* HSS adaptive compression with selected access and matrix multiplication */
  int ctxt;
  int nb;
  int myid;
  int myrow, mycol, nprow, npcol;
  int m;
  bool msg;
  double tstart, tend;
  //  T *Rrcopy, *Rccopy, *Srcopy, *Sccopy;
#if defined(WITH_PAPI)
  float rtime, ptime, mflops;
  long long flpops1, flpops2;
  double flpops, mi, ma, av;
#endif

  tstart=MPI_Wtime();

  MPI_Comm_rank(comm,&myid);
  msg=verbose && myid==0;

  if(!check_descriptor("A",descA,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }

  nb=descA[BLACSnb];
  n=descA[BLACSn];
  ctxt=descA[BLACSctxt];
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);

#if defined(MEMTRACK)
  MemoryTracker();
#endif

#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops1,&mflops);
#endif

  if(use_HSS) {
    if(utree==NULL) {
      /* No user-given clustering */
      if(compressed) {
        if(!myid) std::cout << "Matrix was previously compressed" << std::endl;
        if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
        return;
      }

      if(levels_HSS>0) {
        if(block_HSS>0)
          if(!myid) std::cout << "Warning: both block_HSS and levels_HSS are defined; block_HSS is ignored" << std::endl;
        if(split_HSS)
          m=std::min(levels_HSS,static_cast<int>(floor(log2((double)std::min(split_HSS,n-split_HSS)))+2e0));
        else
          m=std::min(levels_HSS,static_cast<int>(floor(log2((double)n))+1e0));
      } else {
        if(block_HSS<0) {
          if(!myid) std::cout << "levels_HSS=0 so block_size must be >0." << std::endl;
          if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
          return;
        }
        if(split_HSS)
          m=std::min(std::max(1,static_cast<int>(floor(log2((double)std::min(split_HSS,n-split_HSS)/double(block_HSS))+2e0))),static_cast<int>(floor(log2((double)n))+1e0));
        else
          m=std::min(std::max(1,static_cast<int>(floor(log2((double)n/double(block_HSS))+1e0))),static_cast<int>(floor(log2((double)n))+1e0));
      }
      if(split_HSS && m<=1)
        m=2;

      if(split_HSS) {
        /* Constrained tree */
        if(split_HSS<1 || split_HSS>n) {
          if(!myid) std::cout << "1 <= split_HSS < n must hold." << std::endl;
          if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
          return;
        }
        htree.generateConstrainedBinaryTree(m,split_HSS,n);
      } else {
        /* Unconstrained binary tree */
        htree.generateBinaryTree(m,n);
      }
    } else {
      /* Generate a tree from the user-given clustering */
      htree.generateFromPartitionTree(utree);
      m=htree.nr_lvl;
    }
    htree.comm=comm;
    htree.obj=obj;

    if(np==1) {
      /* Sequential compression */
      if(msg) std::cout << "Sequential HSS compression with " << m << " levels in the tree..." << std::endl;
#if defined(MEMTRACK)
      MemoryTracker::reset();
#endif
      StrumpackRand rgen(time(NULL));

      htree.HSSR_seq_Compress(NULL,descA,tol_HSS,min_rand_HSS,max_rand_HSS,lim_rand_HSS,inc_rand_HSS,&eff_rand_HSS,&max_rank,0,NULL,NULL,NULL,NULL,&rgen,elements,Amult,0,false);

      /* Flops */
      flops_compr+=htree.getFlopsHSSRCompression(eff_rand_HSS,false);

      /* Memory: factors */
      mem_HSS+=htree.getSizeHSSFactors();

#if defined(MEMTRACK)
      /* Memory: highmark */
      mem_compr+=MemoryTracker::get();
#endif
    } else {
      /* Parallel compression */
      if(msg) std::cout << "Parallel HSS compression with " << m << " levels in the tree..." << std::endl;
      htree.proportionalMapping(np);
      ginfo.ctxt=ctxt;
      ginfo.nprow=nprow;
      ginfo.npcol=npcol;
      ginfo.myrow=myrow;
      ginfo.mycol=mycol;
      ginfo.mb=nb;
      ginfo.nb=nb;
#if defined(MEMTRACK)
      MemoryTracker::reset();
#endif
      htree.HSSR_par_Compress(NULL,descA,tol_HSS,min_rand_HSS,max_rand_HSS,lim_rand_HSS,inc_rand_HSS,&eff_rand_HSS,&max_rank,&ginfo,NULL,NULL,NULL,NULL,NULL,elements,Amult,access_is_global);

      /* Flops */
      flops_compr+=htree.getFlopsHSSRCompression(eff_rand_HSS,false);

      /* Memory: factors */
      mem_HSS+=htree.getSizeHSSFactors();

#if defined(MEMTRACK)
      /* Memory: highmark */
      mem_compr+=MemoryTracker::get();
#endif
    }
    compressed=true;
  } else {
    if(!myid) std::cout << "Warning; compression routine was called but HSS is off (use_HSS=false). Nothing to do." << std::endl << std::endl;
  }

  tend=MPI_Wtime();
#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops2,&mflops);
  flpops=static_cast<double>(flpops2-flpops1);
  MPI_Allreduce((void *) &flpops,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
  MPI_Allreduce((void *) &flpops,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
  MPI_Allreduce((void *) &flpops,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
  av/=static_cast<double>(np);
  if(msg) std::cout << "PAPI GFlop count: min= " << mi/1.0e9 << "; avg= " << av/1.0e9 << "; max= " << ma/1.0e9 << std::endl;
#endif
  if(msg) std::cout << "Done in " << tend-tstart << "s; max rank = " << max_rank << "; effective number of random vectors = " << eff_rand_HSS << std::endl << std::endl;
  time_compr+=tend-tstart;
#if defined(MEMTRACK)
  MemoryTracker::reset();
#endif
  MPI_Bcast((void *) &compressed,IONE,MPI_INTEGER,IZERO,comm);
}

template<typename T, typename S> S StrumpackDensePackage<T,S>::check_compression(T *A, int *descA) {
  /* Check the quality of the compression by computing ||HSS(A)*I-A||_fro/||A||_fro.
   * NB: this is VERY expensive (usually longer than HSS compression or LU factorization). */

  int ctxt;
  int locr, locc;
  int n;
  int nb;
  int myid;
  int myrow, mycol, nprow, npcol;
  bool msg;
  T *X, *R;
  S *work;
  S err;
  int i,j, ii, jj;
  double tstart, tend;

  tstart=MPI_Wtime();

  MPI_Comm_rank(comm,&myid);
  msg=verbose && myid==0;

  if(!check_descriptor("A",descA,true)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return -1;
  }

  n=descA[BLACSn];
  ctxt=descA[BLACSctxt];
  nb=descA[BLACSnb];
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);

  if(use_HSS) {
    if(!compressed) {
      if(!myid) std::cout << "Matrix was not previously compressed." << std::endl;
      if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
      return -1;
    }
    if(msg) std::cout << "Checking compression quality by computing HSS(A)*I-A..." << std::endl;
    if(np==1) {
      X=new T[n*n];
      R=new T[n*n];
      for(j=0;j<n;j++)
        for(i=0;i<n;i++) {
          X[j*n+i]=(i==j)?ONE:ZERO;
          R[j*n+i]=A[j*n+i];
        }
      htree.HSS_seq_Product('N',ONE,X,n,NONE,R,0);
      work=new S[n];
      err=lange('F',n,n,R,n,work)/lange('F',n,n,A,n,work);
      delete[] work;
      delete[] X;
      delete[] R;

      /* Flops */
      flops_prod+=htree.getFlopsProduct(n);
    } else {
      if(nprow>0) {
        locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
        locc=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
        X=new T[locr*locc];
        R=new T[locr*locc];
        for(i=1;i<=locr;i++)
          for(j=1;j<=locc;j++) {
            ii=indxl2g_(&i,&nb,&myrow,&IZERO,&nprow);
            jj=indxl2g_(&j,&nb,&mycol,&IZERO,&npcol);
            X[locr*(j-1)+(i-1)]=ii==jj?ONE:ZERO;
            R[locr*(j-1)+(i-1)]=A[locr*(j-1)+(i-1)];
          }
      }
      htree.HSS_par_Product('N',ONE,X,descA,n,NONE,R,descA,&ginfo,0);
      if(nprow>0) {
        work=new S[locr];
        err=plange('F',n,n,R,IONE,IONE,descA,work)/plange('F',n,n,A,IONE,IONE,descA,work);
        delete[] work;
        delete[] X;
        delete[] R;
      }

      /* Flops */
      flops_prod+=htree.getFlopsProduct(n);
    }
    if(msg) std::cout << "Compression quality: ||HSS(A)*I-A||_fro/||A||_fro=" << err << std::endl;
  } else {
    if(!myid) std::cout << "Warning; compression checking routine was called but HSS is off (use_HSS=false). Nothing to do." << std::endl;
  }
  tend=MPI_Wtime();
  if(msg) std::cout << "Done in " << tend-tstart << "s" << std::endl << std::endl;
  time_prod+=tend-tstart;
  return err;
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::factor(T *A, int *descA) {
  /* Factorization: either using HSS compression (if use_HSS true and the matrix was
   * previously compressed), or ScaLAPACK LU factorization */
  int ctxt;
  int locr, locc;
  int nb;
  int myid;
  int myrow, mycol, nprow, npcol;
  int ierr;
  bool msg;
  double tstart, tend;
#if defined(WITH_PAPI)
  float rtime, ptime, mflops;
  long long flpops1, flpops2;
  double flpops, mi, ma, av;
#endif

  tstart=MPI_Wtime();

  MPI_Comm_rank(comm,&myid);
  msg=verbose && myid==0;

#if defined(MEMTRACK)
  MemoryTracker();
#endif

#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops1,&mflops);
#endif

  if(use_HSS) {
    if(!compressed) {
      if(myid==0) std::cout << "Matrix was not previously compressed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(factored_HSS) {
      if(myid==0) std::cout << "Matrix is already ULV-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(partially_factored_HSS) {
      if(myid==0) std::cout << "Matrix is already partially ULV-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(schur_HSS) {
      if(myid==0) std::cout << "An HSS Schur complement was already computed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    /* Get info from the tree */
    ctxt=htree.nodes[htree.dim_tree-1]->allctxt;
    if(ctxt!=-1)
      blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
  } else {
    if(factored_LU) {
      if(myid==0) std::cout << "Matrix is already LU-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(partially_factored_LU) {
      if(myid==0) std::cout << "Matrix is already partially LU-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(schur_LU) {
      if(myid==0) std::cout << "An exact Schur complement was already computed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!check_descriptor("A",descA,true)) {
      std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    descLU=new int[BLACSCTXTSIZE];
    copyDesc(descA,descLU);
    /* Get info from descA */
    n=descA[BLACSn];
    ctxt=descA[BLACSctxt];
    nb=descA[BLACSnb];
    blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
    if(nprow==-1) nprow=0;
  }

  if(use_HSS) {
    if(np==1) {
      if(msg) std::cout << "Sequential ULV factorization..." << std::endl;
#if defined(MEMTRACK)
      MemoryTracker::reset();
#endif
      htree.HSS_seq_ULV(0,0,0);

      /* Flops */
      flops_fact+=htree.getFlopsULVFactorization(0);

      /* Memory: factors */
      mem_ULV+=htree.getSizeULVFactors(0);

#if defined(MEMTRACK)
      /* Memory: highmark */
      mem_fact=MemoryTracker::get();
#endif
    } else {
      if(msg) std::cout << "Parallel ULV factorization..." << std::endl;
#if defined(MEMTRACK)
      MemoryTracker::reset();
#endif
      htree.HSS_par_ULV(&ginfo,0);

      /* Flops */
      flops_fact+=htree.getFlopsULVFactorization(0);

      /* Memory: factors */
      mem_ULV+=htree.getSizeULVFactors(0);

#if defined(MEMTRACK)
      /* Memory: highmark */
      mem_fact=MemoryTracker::get();
#endif
    }
    factored_HSS=true;
  } else {
    if(np==1) {
      /* Factorization with LAPACK */
      if(msg) std::cout << "Factorization with LAPACK..." << std::endl;
      piv=new int[n];
      LU=new T[n*n];
      lacpy('N',n,n,A,n,LU,n);
      ierr=getrf(n,n,LU,n,piv);
      if(ierr==0)
        factored_LU=true;
      else
        std::cout << "Failed!" << std::endl;

      /* Memory: factors */
      mem_ULV+=static_cast<double>(n)*static_cast<double>(n)*sizeof(T);

      /* Flops */
      flops_fact+=2.0/3.0*static_cast<double>(n)*static_cast<double>(n)*static_cast<double>(n);
    } else if(nprow>0){
      /* Factorization with ScaLAPACK */
      if(msg) std::cout << "Factorization with ScaLAPACK..." << std::endl;
      locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
      locc=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
      LU=new T[locr*locc];
      placpy('N',n,n,A,IONE,IONE,descA,LU,IONE,IONE,descA);
      piv=new int[locr+nb];
      ierr=pgetrf(n,n,LU,IONE,IONE,descA,piv);
      if(ierr==0)
        factored_LU=true;
      else
        if(myid==0) std::cout << "Failed!" << std::endl;

      /* Memory: factors */
      mem_ULV+=static_cast<double>(n)*static_cast<double>(n)*sizeof(T)/static_cast<double>(nprow)/static_cast<double>(npcol);

      /* Flops */
      flops_fact+=2.0/3.0*static_cast<double>(n)*static_cast<double>(n)*static_cast<double>(n)/static_cast<double>(nprow)/static_cast<double>(npcol);
    }
  }

  tend=MPI_Wtime();
#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops2,&mflops);
  flpops=static_cast<double>(flpops2-flpops1);
  MPI_Allreduce((void *) &flpops,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
  MPI_Allreduce((void *) &flpops,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
  MPI_Allreduce((void *) &flpops,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
  av/=static_cast<double>(np);
  if(msg) std::cout << "PAPI GFlop count: min= " << mi/1.0e9 << "; avg= " << av/1.0e9 << "; max= " << ma/1.0e9 << std::endl;
#endif
  time_fact+=tend-tstart;
  MPI_Bcast((void *) &factored_LU,IONE,MPI_INTEGER,IZERO,comm);
  MPI_Bcast((void *) &factored_HSS,IONE,MPI_INTEGER,IZERO,comm);
  if(msg) std::cout << "Done in " << tend-tstart << "s" << std::endl << std::endl;
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::partially_factor(T *A, int *descA, int nfact) {
  /* Partial factorization: either using HSS compression (if use_HSS true and the matrix was
   * previously compressed), or ScaLAPACK LU factorization. "nfact" is the number of
   * variables to be eliminated.
   */
  int ctxt;
  int locr, locc;
  int nb;
  int myid;
  int myrow, mycol, nprow, npcol;
  int ierr;
  int ns;
  int npiv, idummy;
  int descPiv[BLACSCTXTSIZE];
  bool msg;
  double tstart, tend;
#if defined(WITH_PAPI)
  float rtime, ptime, mflops;
  long long flpops1, flpops2;
  double flpops, mi, ma, av;
#endif

  tstart=MPI_Wtime();

  MPI_Comm_rank(comm,&myid);
  msg=verbose && myid==0;

#if defined(MEMTRACK)
  MemoryTracker();
#endif

#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops1,&mflops);
#endif

  if(use_HSS) {
    if(!compressed) {
      if(myid==0) std::cout << "Matrix was not previously compressed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(factored_HSS) {
      if(myid==0) std::cout << "Matrix is already ULV-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(partially_factored_HSS) {
      if(myid==0) std::cout << "Matrix is already partially ULV-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(schur_HSS) {
      if(myid==0) std::cout << "An HSS Schur complement was already computed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    /* Check the indices of the left child of the root node against nfact */
    if(nfact!=htree.nodes[htree.dim_tree-1]->children[0]->tree_ie) {
      if(myid==0) std::cout << "Parameter nfact does not conform with HSS tree. Check parameter split_HSS." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    /* Get info from the tree */
    ctxt=htree.nodes[htree.dim_tree-1]->allctxt;
    if(ctxt!=-1)
      blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
  } else {
    if(factored_LU) {
      if(myid==0) std::cout << "Matrix is already LU-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(partially_factored_LU) {
      if(myid==0) std::cout << "Matrix is already partially LU-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(schur_LU) {
      if(myid==0) std::cout << "An exact Schur complement was already computed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!check_descriptor("A",descA,true)) {
      std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    descLU=new int[BLACSCTXTSIZE];
    copyDesc(descA,descLU);
    /* Get info from descA */
    n=descA[BLACSn];
    ctxt=descA[BLACSctxt];
    nb=descA[BLACSnb];
    blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
    if(nprow==-1) nprow=0;
    if(nfact<0 || nfact>=n) {
      if(myid==0) std::cout << "0<=nfact<n must hold." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
  }

  if(use_HSS) {
    if(np==1) {
      if(msg) std::cout << "Sequential partial ULV factorization..." << std::endl;
#if defined(MEMTRACK)
      MemoryTracker::reset();
#endif
      htree.HSS_seq_ULV(0,1,0);

      /* Flops */
      flops_fact+=htree.getFlopsULVFactorization(1);

      /* Memory: factors */
      mem_ULV+=htree.getSizeULVFactors(1);

#if defined(MEMTRACK)
      /* Memory: highmark */
      mem_fact=MemoryTracker::get();
#endif
    } else {
      if(msg) std::cout << "Parallel partial ULV factorization..." << std::endl;
#if defined(MEMTRACK)
      MemoryTracker::reset();
#endif
      htree.HSS_par_ULV(&ginfo,1);

      /* Flops */
      flops_fact+=htree.getFlopsULVFactorization(1);

      /* Memory: factors */
      mem_ULV+=htree.getSizeULVFactors(1);

#if defined(MEMTRACK)
      /* Memory: highmark */
      mem_fact=MemoryTracker::get();
#endif
    }
    partially_factored_HSS=true;
  } else {
    ns=n-nfact;
    if(np==1) {
      /* Partial factorization with LAPACK */
      if(msg) std::cout << "Partial factorization with LAPACK..." << std::endl;
      LU=new T[n*n];
      lacpy('N',n,n,A,n,LU,n);

      if(nfact) {
        /* A11 <- P11' L11 U11 (P11 A11 = L11 U11) */
        piv=new int[nfact];
        ierr=getrf(nfact,nfact,LU,n,piv);
        if(ierr==0)
          partially_factored_LU=true;
        else
          std::cout << "Failed!" << std::endl;

        /* A12 <- L11 \ (P11 A12) */
        laswp(ns,&LU[n*nfact],n,IONE,nfact,piv,IONE);
        trsm('L','L','N','U',nfact,ns,ONE,LU,n,&LU[n*nfact],n);

        /* A21 <- A21 U11^-1 */
        trsm('R','U','N','N',ns,nfact,ONE,LU,n,&LU[nfact],n);
      } else {
        partially_factored_LU=true;
      }

      /* Memory: factors */
      mem_ULV+=static_cast<double>(n)*static_cast<double>(n)*sizeof(T);

      /* Flops */
      flops_fact+=2.0/3.0*static_cast<double>(nfact)*static_cast<double>(nfact)*static_cast<double>(nfact);
      flops_fact+=2.0*static_cast<double>(nfact)*static_cast<double>(nfact)*static_cast<double>(ns);
    } else if(nprow>0){
      /* Partial factorization with ScaLAPACK */
      if(msg) std::cout << "Partial factorization with ScaLAPACK..." << std::endl;
      locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
      locc=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
      LU=new T[locr*locc];

      if(nfact) {
        /* A11 <- P11' L11 U11 (P11 A11 = L11 U11) */
        placpy('N',n,n,A,IONE,IONE,descA,LU,IONE,IONE,descA);
        piv=new int[locr+nb];
        ierr=pgetrf(nfact,nfact,LU,IONE,IONE,descA,piv);
        if(ierr==0)
          partially_factored_LU=true;
        else
          if(myid==0) std::cout << "Failed!" << std::endl;

        /* A12 <- L11 \ (P11 A12) */
        npiv=n+nb*nprow;
        idummy=nb+numroc_(&n,&nb,&myrow,&IZERO,&nprow);
        descinit_(descPiv,&npiv,&IONE,&nb,&IONE,&IZERO,&mycol,&ctxt,&idummy,&nprow);
        plapiv('F','R','C',nfact,ns,LU,IONE,nfact+1,descLU,piv,IONE,IONE,descPiv,NULL);
        ptrsm('L','L','N','U',nfact,ns,ONE,LU,IONE,IONE,descLU,LU,IONE,nfact+1,descLU);

        /* A21 <- A21 U11^-1 */
        ptrsm('R','U','N','N',ns,nfact,ONE,LU,IONE,IONE,descLU,LU,nfact+1,IONE,descLU);
      }

      /* Memory: factors */
      mem_ULV+=static_cast<double>(n)*static_cast<double>(n)*sizeof(T)/static_cast<double>(nprow)/static_cast<double>(npcol);

      /* Flops */
      flops_fact+=2.0/3.0*static_cast<double>(nfact)*static_cast<double>(nfact)*static_cast<double>(nfact)/static_cast<double>(nprow)/static_cast<double>(npcol);
      flops_fact+=2.0*static_cast<double>(nfact)*static_cast<double>(nfact)*static_cast<double>(ns)/static_cast<double>(nprow)/static_cast<double>(npcol);
    }
  }

  var_schur=nfact+1;

  tend=MPI_Wtime();
#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops2,&mflops);
  flpops=static_cast<double>(flpops2-flpops1);
  MPI_Allreduce((void *) &flpops,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
  MPI_Allreduce((void *) &flpops,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
  MPI_Allreduce((void *) &flpops,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
  av/=static_cast<double>(np);
  if(msg) std::cout << "PAPI GFlop count: min= " << mi/1.0e9 << "; avg= " << av/1.0e9 << "; max= " << ma/1.0e9 << std::endl;
#endif
  time_fact+=tend-tstart;
  MPI_Bcast((void *) &partially_factored_LU,IONE,MPI_INTEGER,IZERO,comm);
  MPI_Bcast((void *) &partially_factored_HSS,IONE,MPI_INTEGER,IZERO,comm);
  if(msg) std::cout << "Done in " << tend-tstart << "s" << std::endl << std::endl;
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::compute_schur() {
  /* Schur complement computation: either using LAPACK/ScaLAPACK operations,
   * either using HSS operations.
   */
  int ctxt;
  int myid;
  int myrow, mycol, nprow, npcol;
  int nfact, ns;
  bool msg;
  double tstart, tend;
#if defined(WITH_PAPI)
  float rtime, ptime, mflops;
  long long flpops1, flpops2;
  double flpops, mi, ma, av;
#endif

  tstart=MPI_Wtime();

  MPI_Comm_rank(comm,&myid);
  msg=verbose && myid==0;

#if defined(MEMTRACK)
  MemoryTracker();
#endif

#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops1,&mflops);
#endif

  if(use_HSS) {
    if(!compressed) {
      if(myid==0) std::cout << "Matrix was not previously compressed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(factored_HSS) {
      if(myid==0) std::cout << "Matrix is already ULV-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!partially_factored_HSS) {
      if(myid==0) std::cout << "Matrix is not partially ULV-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(schur_HSS) {
      if(myid==0) std::cout << "An HSS Schur complement was already computed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    /* Get info from the tree */
    ctxt=htree.nodes[htree.dim_tree-1]->allctxt;
    if(ctxt!=-1)
      blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
  } else {
    if(factored_LU) {
      if(myid==0) std::cout << "Matrix is already LU-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!partially_factored_LU) {
      if(myid==0) std::cout << "Matrix is not partially LU-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(schur_LU) {
      if(myid==0) std::cout << "An exact Schur complement was already computed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    /* Get info from descLU */
    n=descLU[BLACSn];
    ctxt=descLU[BLACSctxt];
    blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
    if(nprow==-1) nprow=0;
  }

  if(use_HSS) {
    if(np==1) {
      if(msg) std::cout << "Sequential HSS Schur complement computation..." << std::endl;
#if defined(MEMTRACK)
      MemoryTracker::reset();
#endif
      htree.HSS_seq_GenerateSchur();

      /* Flops */
      flops_fact+=htree.getFlopsSchurUpdate();

      /* Memory: Schur complement update */
      mem_ULV+=htree.getSizeSchurUpdate();

#if defined(MEMTRACK)
      /* Memory: highmark */
      mem_fact=+MemoryTracker::get();
#endif
    } else {
      if(msg) std::cout << "Parallel HSS Schur complement computation..." << std::endl;
#if defined(MEMTRACK)
      MemoryTracker::reset();
#endif
      htree.HSS_par_GenerateSchur(&ginfo);

      /* Flops */
      flops_fact+=htree.getFlopsSchurUpdate();

      /* Memory: Schur complement update */
      mem_ULV+=htree.getSizeSchurUpdate();

#if defined(MEMTRACK)
      /* Memory: highmark */
      mem_fact+=MemoryTracker::get();
#endif
    }
    schur_HSS=true;
  } else {
    nfact=var_schur-1;
    ns=n-var_schur+1;

    if(np==1) {
      /* Schur complement computation with LAPACK */
      if(msg) std::cout << "Schur complement computation with LAPACK..." << std::endl;

      if(nfact) {
        /* A22 <- A22 - A21 A12 */
        gemm('N','N',ns,ns,nfact,NONE,&LU[nfact],n,&LU[n*nfact],n,ONE,&LU[nfact+nfact*n],n);
      }

      schur_LU=true;

      /* Flops */
      flops_fact-=2.0*static_cast<double>(nfact)*static_cast<double>(ns)*static_cast<double>(ns);
    } else if(nprow>0){
      /* Schur complement computation with ScaLAPACK */
      if(msg) std::cout << "Schur complement computation with ScaLAPACK..." << std::endl;

      if(nfact) {
        /* A22 <- A22 - A21 A12 */
        pgemm('N','N',ns,ns,nfact,NONE,LU,var_schur,IONE,descLU,LU,IONE,var_schur,descLU,ONE,LU,var_schur,var_schur,descLU);
      }

      schur_LU=true;

      /* Flops */
      flops_fact-=2.0*static_cast<double>(nfact)*static_cast<double>(ns)*static_cast<double>(ns)/static_cast<double>(nprow)/static_cast<double>(npcol);
    }
  }

  tend=MPI_Wtime();
#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops2,&mflops);
  flpops=static_cast<double>(flpops2-flpops1);
  MPI_Allreduce((void *) &flpops,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
  MPI_Allreduce((void *) &flpops,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
  MPI_Allreduce((void *) &flpops,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
  av/=static_cast<double>(np);
  if(msg) std::cout << "PAPI GFlop count: min= " << mi/1.0e9 << "; avg= " << av/1.0e9 << "; max= " << ma/1.0e9 << std::endl;
#endif
  time_fact+=tend-tstart;
  MPI_Bcast((void *) &schur_LU,IONE,MPI_INTEGER,IZERO,comm);
  MPI_Bcast((void *) &schur_HSS,IONE,MPI_INTEGER,IZERO,comm);
  if(msg) std::cout << "Done in " << tend-tstart << "s" << std::endl << std::endl;
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::solve(T *X, int *descX, T *B, int *descB) {
  /* Solution of AX=B: either using the HSS factors (if use_HSS true and the matrix was
   * previously compressed and factorized), or ScaLAPACK triangular solve */
  int ctxt;
  int n;
  int myid;
  int nrhs;
  int myrow, mycol, nprow, npcol;
  int ierr;
  bool msg;
  double tstart, tend;
#if defined(WITH_PAPI)
  float rtime, ptime, mflops;
  long long flpops1, flpops2;
  double flpops, mi, ma, av;
#endif

  tstart=MPI_Wtime();

  MPI_Comm_rank(comm,&myid);
  msg=verbose && myid==0;

#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops1,&mflops);
#endif

  if(!check_descriptor("X",descX,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }
  if(!check_descriptor("B",descB,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }

  if(use_HSS) {
    if(!compressed) {
      if(myid==0) std::cout << "Matrix was not previously compressed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!factored_HSS) {
      if(myid==0) std::cout << "Matrix was not previously ULV-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(partially_factored_HSS) {
      if(myid==0) std::cout << "Matrix is partially ULV-factored; regular solve cannot be used." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    /* Get info from the tree */
    ctxt=htree.nodes[htree.dim_tree-1]->allctxt;
    if(ctxt!=-1)
      blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
  } else {
    if(!factored_LU) {
      if(myid==0) std::cout << "Matrix was not previously LU-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    /* Get info from descLU */
    n=descLU[BLACSn];
    ctxt=descLU[BLACSctxt];
    blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
    if(nprow==-1) nprow=0;
  }
  nrhs=descB[BLACSn];

  if(!use_HSS) {
    if(descLU[BLACSm]!=descX[BLACSm]||descLU[BLACSm]!=descB[BLACSm]||descB[BLACSm]!=descX[BLACSm]) {
      std::cout << "A, X, B have different row counts! A:" << descLU[BLACSm] << "; B" << descB[BLACSm] << "; X:" << descX[BLACSm] << std::endl;
      std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(descLU[BLACSmb]!=descX[BLACSmb]||descLU[BLACSmb]!=descB[BLACSmb]||descB[BLACSmb]!=descX[BLACSmb]) {
      std::cout << "A, X, B have different block sizes! A:" << descLU[BLACSmb] << "; B" << descB[BLACSmb] << "; X:" << descX[BLACSmb] << std::endl;
      std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
  }
  if(descB[BLACSn]!=descX[BLACSn]) {
    std::cout << "X, B have different column counts! B:" << descB[BLACSn] << "; X:" << descX[BLACSn] << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }

  if(use_HSS) {
    if(np==1) {
      /* Sequential HSS triangular solution */
      if(msg) std::cout << "Sequential HSS triangular solution..." << std::endl;
      htree.HSS_seq_Solve(X,nrhs,B,0);

      /* Flops */
      flops_solve+=htree.getFlopsSolve(nrhs);
    } else {
      /* Parallel HSS triangular solution */
      if(msg) std::cout << "Parallel HSS triangular solution..." << std::endl;
      htree.HSS_par_Solve(X,descX,nrhs,B,descB,&ginfo,0);

      /* Flops */
      flops_solve+=htree.getFlopsSolve(nrhs);
    }
  } else {
    if(np==1) {
      /* Solution with LAPACK */
      if(msg) std::cout << "Triangular solution with LAPACK..." << std::endl;
      lacpy('N',n,nrhs,B,n,X,n);
      ierr=getrs('N',n,nrhs,LU,n,piv,X,n);
      if(ierr)
        std::cout << "Failed!" << std::endl;

      /* Flops */
      flops_solve+=2*static_cast<double>(n)*static_cast<double>(n)*static_cast<double>(nrhs);

    } else if(nprow>0){
      /* Solution with ScaLAPACK */
      if(msg) std::cout << "Triangular solution with ScaLAPACK..." << std::endl;
      pgeadd('N',n,nrhs,ONE,B,IONE,IONE,descB,ZERO,X,IONE,IONE,descX);
      ierr=pgetrs('N',n,nrhs,LU,IONE,IONE,descLU,piv,X,IONE,IONE,descX);
      if(ierr)
        if(myid==0) std::cout << "Failed!" << std::endl;

      /* Flops */
      flops_solve+=2*static_cast<double>(n)*static_cast<double>(n)*static_cast<double>(nrhs)/static_cast<double>(nprow)/static_cast<double>(npcol);
    }
  }

  tend=MPI_Wtime();
#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops2,&mflops);
  flpops=static_cast<double>(flpops2-flpops1);
  MPI_Allreduce((void *) &flpops,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
  MPI_Allreduce((void *) &flpops,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
  MPI_Allreduce((void *) &flpops,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
  av/=static_cast<double>(np);
  if(msg) std::cout << "PAPI GFlop count: min= " << mi/1.0e9 << "; avg= " << av/1.0e9 << "; max= " << ma/1.0e9 << std::endl;
#endif
  if(msg) std::cout << "Done in " << tend-tstart << "s" << std::endl << std::endl;
  time_solve+=tend-tstart;
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::reduce_RHS(T *X, int *descX, T *B, int *descB) {
  /* Reduces a right-hand side to a vector to be fed into a Schur complement solver.  */
  int ctxt;
  int n;
  int nb;
  int myid;
  int nrhs;
  int myrow, mycol, nprow, npcol;
  int npiv, idummy, ierr;
  int locr, locc;
  int descPiv[BLACSCTXTSIZE], desc_tmp[BLACSCTXTSIZE], desc_tmpk[BLACSCTXTSIZE], desc_tmpX[BLACSCTXTSIZE];
  bool msg;
  double tstart, tend;
  T *tmpk, *tmp, *tmpX;
  gridinfo tmpinfo;
#if defined(WITH_PAPI)
  float rtime, ptime, mflops;
  long long flpops1, flpops2;
  double flpops, mi, ma, av;
#endif

  tstart=MPI_Wtime();

  MPI_Comm_rank(comm,&myid);
  msg=verbose && myid==0;

#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops1,&mflops);
#endif

  if(!check_descriptor("X",descX,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }
  if(!check_descriptor("B",descB,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }

  if(use_HSS) {
    if(!compressed) {
      if(myid==0) std::cout << "Matrix was not previously compressed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(factored_HSS) {
      if(myid==0) std::cout << "Matrix fully factored; this cannot be used." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!partially_factored_HSS) {
      if(myid==0) std::cout << "Matrix was not previously partially ULV-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!schur_HSS) {
      if(myid==0) std::cout << "HSS Schur complement was not previously computed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    /* Get info from the tree */
    n=htree.a_dim;
  } else {
    /* Get info from descLU */
    n=descLU[BLACSn];
    ctxt=descLU[BLACSctxt];
    blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
    if(nprow==-1) nprow=0;
    if(!partially_factored_LU) {
      if(myid==0) std::cout << "Matrix was not previously partially LU-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!schur_LU) {
      if(myid==0) std::cout << "An exact Schur complement was not previously computed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
  }
  nrhs=descB[BLACSn];
  nb=descB[BLACSmb];

  if(!use_HSS) {
    if(descLU[BLACSm]!=descX[BLACSm]||descLU[BLACSm]!=descB[BLACSm]||descB[BLACSm]!=descX[BLACSm]) {
      std::cout << "A, X, B have different row counts! A:" << descLU[BLACSm] << "; B" << descB[BLACSm] << "; X:" << descX[BLACSm] << std::endl;
      std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(descLU[BLACSmb]!=descX[BLACSmb]||descLU[BLACSmb]!=descB[BLACSmb]||descB[BLACSmb]!=descX[BLACSmb]) {
      std::cout << "A, X, B have different block sizes! A:" << descLU[BLACSmb] << "; B" << descB[BLACSmb] << "; X:" << descX[BLACSmb] << std::endl;
      std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
  }
  if(descB[BLACSn]!=descX[BLACSn]) {
    std::cout << "X, B have different column counts! B:" << descB[BLACSn] << "; X:" << descX[BLACSn] << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }

  if(use_HSS) {
    if(np==1) {
      /* Sequential HSS RHS reduction */
      if(msg) std::cout << "Sequential HSS RHS reduction..." << std::endl;
      /* Compute y1 = "L11" \ b1 using a partial ULV forward solve */
      htree.HSS_seq_Solve(X,nrhs,B,1);
      /* Compute y2 = b2 - "F21" y1 using the HSS Schur update: theta*(Vtilda_k^T y_k + V_k^T zk12) */
      lacpy('N',n-var_schur+1,nrhs,&B[var_schur-1],n,&X[var_schur-1],n);
      if(htree.ctheta) {
        tmp=new T[htree.ctheta*nrhs]();
        /* tmp <- Vtilda_k^T y_k */
        gemm('C','N',htree.ctheta,nrhs,htree.rpsi,ONE,htree.nodes[htree.knode]->Vt,htree.rpsi,htree.y[htree.knode],htree.rpsi,ZERO,tmp,htree.ctheta);
        /* tmp += V_k zk12 if k has descendants */
        if(htree.nodes[htree.knode]->nr_children)
          htree.nodes[htree.knode]->seq_applyGenerator('C','V',nrhs,ONE,htree.zk12,htree.nodes[htree.knode]->v_rows+htree.nodes[htree.knode]->v_rank,ONE,tmp,htree.ctheta);
        /* theta * tmp */
        gemm('N','N',n-var_schur+1,nrhs,htree.ctheta,NONE,htree.theta,htree.rtheta,tmp,htree.ctheta,ONE,&X[var_schur-1],n);
        delete[] tmp;
      }

      /* Flops */
      // TODO
    } else {
      /* Parallel HSS triangular solution */
      if(msg) std::cout << "Parallel HSS RHS reduction..." << std::endl;
      /* Compute y1 = "L11" \ b1 using a partial ULV forward solve */
      htree.HSS_par_Solve(X,descX,nrhs,B,descB,&ginfo,1);

      /* Compute y2 = b2 - "F21" y1 using the HSS Schur update: theta*(Vtilda_k^T y_k + V_k^T zk12):
       *   1/ Compute Vk^T zk12 on the knode-grid (in "tmpk").
       *   2/ Move the result to the root node grid (in tmp).
       *   3/ Add Vtilda_k^T y_k (root node grid).
       *   4/ Multiply by theta, in tmpX (root node grid).
       *   5/ Redistribute tmpX into X.
       *   6/ Do X <- B -X
       *
       */
      ctxt=htree.nodes[htree.dim_tree-1]->context;
      blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
      if(myrow>=0 && mycol>=0) {
        locr=numroc_(&htree.ctheta,&nb,&myrow,&IZERO,&nprow);
        locc=numroc_(&nrhs,&nb,&mycol,&IZERO,&npcol);
        tmp=new T[locr*locc]();
        idummy=std::max(locr,1);
        descinit_(desc_tmp,&htree.ctheta,&nrhs,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
      } else {
        descset_(desc_tmp,&htree.ctheta,&nrhs,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
      }

      if(htree.nodes[htree.knode]->nr_children) {
        /* Compute Vk^T * zk12 into tmpk and move it to tmp.
         * If knode doesn't have children, tmp is simply
         * 0 at this point, after new[...]().
         */
        ctxt=htree.nodes[htree.knode]->context;
        blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
        if(myrow>=0 && mycol>=0) {
          locr=numroc_(&htree.ctheta,&nb,&myrow,&IZERO,&nprow);
          locc=numroc_(&nrhs,&nb,&mycol,&IZERO,&npcol);
          tmpk=new T[locr*locc]();
          idummy=std::max(locr,1);
          descinit_(desc_tmpk,&htree.ctheta,&nrhs,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
          tmpinfo.ctxt=ctxt;
          tmpinfo.nprow=nprow;
          tmpinfo.npcol=npcol;
          tmpinfo.myrow=myrow;
          tmpinfo.mycol=mycol;
          tmpinfo.mb=nb;
          tmpinfo.nb=nb;
          if(htree.nodes[htree.knode]->procs[0]==htree.nodes[htree.knode]->procs[1])
            htree.nodes[htree.knode]->seq_applyGenerator('C','V',nrhs,ONE,htree.zk12,htree.nodes[htree.knode]->v_rows+htree.nodes[htree.knode]->v_rank,ZERO,tmpk,htree.ctheta);
          else
            htree.nodes[htree.knode]->par_applyGenerator('C','V',nrhs,ONE,htree.zk12,htree.desc_zk12,ZERO,tmpk,desc_tmpk,&tmpinfo);
        } else {
          tmpk=NULL;
          descset_(desc_tmpk,&htree.ctheta,&nrhs,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
        }
        pgemr2d(htree.ctheta,nrhs,tmpk,IONE,IONE,desc_tmpk,tmp,IONE,IONE,desc_tmp,htree.nodes[htree.dim_tree-1]->allctxt);
      }

      ctxt=htree.nodes[htree.dim_tree-1]->context;
      blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
      if(myrow>=0 && mycol>=0) {
        /* tmp <- tmp + Vtilda_k^T y_k */
        pgemm('C','N',htree.ctheta,nrhs,htree.rpsi,ONE,htree.nodes[htree.knode]->Vt,IONE,IONE,htree.nodes[htree.knode]->descVt,htree.yr,IONE,IONE,htree.desc_yr,ONE,tmp,IONE,IONE,desc_tmp);

        /* tmpX <- theta * tmp */
        locr=numroc_(&htree.rtheta,&nb,&myrow,&IZERO,&nprow);
        locc=numroc_(&nrhs,&nb,&mycol,&IZERO,&npcol);
        tmpX=new T[locr*locc]();
        idummy=std::max(locr,1);
        descinit_(desc_tmpX,&htree.rtheta,&nrhs,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
        pgemm('N','N',htree.rtheta,nrhs,htree.ctheta,ONE,htree.theta,IONE,IONE,htree.desc_theta,tmp,IONE,IONE,desc_tmp,ZERO,tmpX,IONE,IONE,desc_tmpX);
        delete[] tmp;
      } else {
        tmpX=NULL;
        descset_(desc_tmpX,&htree.rtheta,&nrhs,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
      }

      pgemr2d(htree.rtheta,nrhs,tmpX,IONE,IONE,desc_tmpX,X,var_schur,IONE,descX,htree.nodes[htree.dim_tree-1]->allctxt);
      delete[] tmpX;

      ctxt=descB[BLACSctxt];
      blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
      if(myrow>=0 && mycol>=0)
        pgeadd('N',n,nrhs,ONE,B,IONE,IONE,descB,NONE,X,IONE,IONE,descX);

      /* Flops */
      // TODO
    }
  } else {
    if(np==1) {
      /* RHS reduction with LAPACK */
      if(msg) std::cout << "RHS reduction with LAPACK..." << std::endl;
      lacpy('N',n,nrhs,B,n,X,n);
      if(var_schur>1) {
        laswp(nrhs,X,n,IONE,var_schur-1,piv,IONE);
        trsm('L','L','N','U',var_schur-1,nrhs,ONE,LU,n,X,n);
        gemm('N','N',n-var_schur+1,nrhs,var_schur-1,NONE,&LU[var_schur-1],n,X,n,ONE,&X[var_schur-1],n);
      }

      /* Flops */
      // TODO

    } else if(nprow>0){
      /* RHS reduction with ScaLAPACK */
      if(msg) std::cout << "RHS reduction with ScaLAPACK..." << std::endl;

      if(var_schur>1) {
        npiv=n+nb*nprow;
        idummy=nb+numroc_(&n,&nb,&myrow,&IZERO,&nprow);
        descinit_(descPiv,&npiv,&IONE,&nb,&IONE,&IZERO,&mycol,&ctxt,&idummy,&nprow);

        placpy('N',n,nrhs,B,IONE,IONE,descB,X,IONE,IONE,descX);
        plapiv('F','R','C',var_schur-1,nrhs,X,IONE,IONE,descX,piv,IONE,IONE,descPiv,NULL);
        ptrsm('L','L','N','U',var_schur-1,nrhs,ONE,LU,IONE,IONE,descLU,X,IONE,IONE,descX);
        pgemm('N','N',n-var_schur+1,nrhs,var_schur-1,NONE,LU,var_schur,IONE,descLU,X,IONE,IONE,descX,ONE,X,var_schur,IONE,descX);
      }

      /* Flops */
      // TODO
    }
  }

  tend=MPI_Wtime();
#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops2,&mflops);
  flpops=static_cast<double>(flpops2-flpops1);
  MPI_Allreduce((void *) &flpops,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
  MPI_Allreduce((void *) &flpops,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
  MPI_Allreduce((void *) &flpops,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
  av/=static_cast<double>(np);
  if(msg) std::cout << "PAPI GFlop count: min= " << mi/1.0e9 << "; avg= " << av/1.0e9 << "; max= " << ma/1.0e9 << std::endl;
#endif
  if(msg) std::cout << "Done in " << tend-tstart << "s" << std::endl << std::endl;
  time_solve+=tend-tstart;
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::expand_solution(T *X, int *descX, T *B, int *descB) {
  /* Expand solution after Schur complement solve.  */
  int ctxt;
  int n;
  int myid;
  int nrhs;
  int nb;
  int myrow, mycol, nprow, npcol;
  int locr, locc;
  int idummy, ierr;
  int desc_tmpX[BLACSCTXTSIZE];
  bool msg;
  double tstart, tend;
  T *tmpX=NULL;
#if defined(WITH_PAPI)
  float rtime, ptime, mflops;
  long long flpops1, flpops2;
  double flpops, mi, ma, av;
#endif

  tstart=MPI_Wtime();

  MPI_Comm_rank(comm,&myid);
  msg=verbose && myid==0;

#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops1,&mflops);
#endif

  if(!check_descriptor("X",descX,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }
  if(!check_descriptor("B",descB,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }

  if(use_HSS) {
    if(!compressed) {
      if(myid==0) std::cout << "Matrix was not previously compressed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(factored_HSS) {
      if(myid==0) std::cout << "Matrix fully factored; this cannot be used." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!partially_factored_HSS) {
      if(myid==0) std::cout << "Matrix was not previously partially ULV-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!schur_HSS) {
      if(myid==0) std::cout << "HSS Schur complement was not previously computed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    /* Get info from the tree */
    n=htree.a_dim;
    ctxt=htree.nodes[htree.dim_tree-1]->allctxt;
    if(ctxt!=-1)
      blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
  } else {
    /* Get info from descLU */
    n=descLU[BLACSn];
    ctxt=descLU[BLACSctxt];
    blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
    if(nprow==-1) nprow=0;
    if(!partially_factored_LU) {
      if(myid==0) std::cout << "Matrix was not previously partially LU-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!schur_LU) {
      if(myid==0) std::cout << "An exact Schur complement was not previously computed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
  }
  nrhs=descB[BLACSn];
  nb=descB[BLACSmb];

  if(!use_HSS) {
    if(descLU[BLACSm]!=descX[BLACSm]||descLU[BLACSm]!=descB[BLACSm]||descB[BLACSm]!=descX[BLACSm]) {
      std::cout << "A, X, B have different row counts! A:" << descLU[BLACSm] << "; B" << descB[BLACSm] << "; X:" << descX[BLACSm] << std::endl;
      std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(descLU[BLACSmb]!=descX[BLACSmb]||descLU[BLACSmb]!=descB[BLACSmb]||descB[BLACSmb]!=descX[BLACSmb]) {
      std::cout << "A, X, B have different block sizes! A:" << descLU[BLACSmb] << "; B" << descB[BLACSmb] << "; X:" << descX[BLACSmb] << std::endl;
      std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
  }
  if(descB[BLACSn]!=descX[BLACSn]) {
    std::cout << "X, B have different column counts! B:" << descB[BLACSn] << "; X:" << descX[BLACSn] << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }

  if(use_HSS) {
    if(np==1) {
      /* Sequential HSS solution expansion */
      if(msg) std::cout << "Sequential HSS solution expansion..." << std::endl;

      /* Compute x1 = "U11" \ (y1 - "F12" x2)  using a partial ULV forward solve
       * and the HSS Schur complement update.
       */
      if(htree.rpsi)
        gemm('N','N',htree.rpsi,nrhs,htree.cpsi,NONE,htree.psi,htree.rpsi,&X[var_schur-1],n,ONE,htree.y[htree.knode],htree.rpsi);
      htree.HSS_seq_Solve(X,nrhs,B,2);

      /* Flops */
      // TODO

    } else {
      /* Parallel HSS solution expansion */
      if(msg) std::cout << "Parallel HSS solution expansion..." << std::endl;

      /* Compute x1 = "U11" \ (y1 - "F12" x2)  using a partial ULV forward solve
       * and the HSS Schur complement update.
       * The bottom part of X needs to be moved to the root node grid before multiplying
       * with theta.
       */
      ctxt=htree.nodes[htree.dim_tree-1]->context;
      blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
      if(myrow>=0 && mycol>=0) {
        locr=numroc_(&htree.rtheta,&nb,&myrow,&IZERO,&nprow);
        locc=numroc_(&nrhs,&nb,&mycol,&IZERO,&npcol);
        tmpX=new T[locr*locc]();
        idummy=std::max(locr,1);
        descinit_(desc_tmpX,&htree.rtheta,&nrhs,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
      } else {
        descset_(desc_tmpX,&htree.rtheta,&nrhs,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
      }
      pgemr2d(htree.rtheta,nrhs,X,var_schur,IONE,descX,tmpX,IONE,IONE,desc_tmpX,htree.nodes[htree.dim_tree-1]->allctxt);
      if(myrow>=0 && mycol>=0) {
        pgemm('N','N',htree.rpsi,nrhs,htree.cpsi,NONE,htree.psi,IONE,IONE,htree.desc_psi,tmpX,IONE,IONE,desc_tmpX,ONE,htree.yr,IONE,IONE,htree.desc_yr);
        delete[] tmpX;
      }
      htree.HSS_par_Solve(X,descX,nrhs,B,descB,&ginfo,2);

      /* Flops */
      // TODO
    }
  } else {
    if(np==1) {
      /* Solution expansion with LAPACK */
      if(msg) std::cout << "Solution expansion with LAPACK..." << std::endl;

      lacpy('N',var_schur-1,nrhs,B,n,X,n);
      if(var_schur>1) {
        gemm('N','N',var_schur-1,nrhs,n-var_schur+1,NONE,&LU[(var_schur-1)*n],n,&X[var_schur-1],n,ONE,X,n);
        trsm('L','U','N','N',var_schur-1,nrhs,ONE,LU,n,X,n);
      }

      /* Flops */
      // TODO

    } else if(nprow>0){
      /* Solution expansion with ScaLAPACK */
      if(msg) std::cout << "Solution expansion with ScaLAPACK..." << std::endl;

      placpy('N',var_schur-1,nrhs,B,IONE,IONE,descB,X,IONE,IONE,descX);
      if(var_schur>1) {
        pgemm('N','N',var_schur-1,nrhs,n-var_schur+1,NONE,LU,IONE,var_schur,descLU,X,var_schur,IONE,descX,ONE,X,IONE,IONE,descX);
        ptrsm('L','U','N','N',var_schur-1,nrhs,ONE,LU,IONE,IONE,descLU,X,IONE,IONE,descX);
      }

      /* Flops */
      // TODO
    }
  }

  tend=MPI_Wtime();
#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops2,&mflops);
  flpops=static_cast<double>(flpops2-flpops1);
  MPI_Allreduce((void *) &flpops,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
  MPI_Allreduce((void *) &flpops,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
  MPI_Allreduce((void *) &flpops,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
  av/=static_cast<double>(np);
  if(msg) std::cout << "PAPI GFlop count: min= " << mi/1.0e9 << "; avg= " << av/1.0e9 << "; max= " << ma/1.0e9 << std::endl;
#endif
  if(msg) std::cout << "Done in " << tend-tstart << "s" << std::endl << std::endl;
  time_solve+=tend-tstart;
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::schur_product(char Trans, T alpha, T *X, int *descX, T beta, T *B, int *descB) {
  /* B = alpha * Schur * X + beta * B */
  int ctxt;
  int n;
  int myid;
  int nrhs;
  int myrow, mycol, nprow, npcol;
  int locr, locc;
  int nb;
  int ierr, idummy;
  int nfs, ns;
  int desc_Xn[BLACSCTXTSIZE], desc_Bn[BLACSCTXTSIZE];
  int desc_tmpX[BLACSCTXTSIZE], desc_tmpB[BLACSCTXTSIZE];
  bool msg;
  double tstart, tend;
#if defined(WITH_PAPI)
  float rtime, ptime, mflops;
  long long flpops1, flpops2;
  double flpops, mi, ma, av;
#endif
  T *tmpX, *tmpB;
  T *Xn, *Bn;

  tstart=MPI_Wtime();

  MPI_Comm_rank(comm,&myid);
  msg=verbose && myid==0;

#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops1,&mflops);
#endif

  if(!check_descriptor("X",descX,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }
  if(!check_descriptor("B",descB,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }

  if(use_HSS) {
    if(!compressed) {
      if(myid==0) std::cout << "Matrix was not previously compressed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!partially_factored_HSS) {
      if(myid==0) std::cout << "Matrix was not previously partially ULV-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!schur_HSS) {
      if(myid==0) std::cout << "HSS Schur complement was not previously computed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    /* Get info from the tree */
    n=htree.a_dim;
  } else {
    /* Get info from descLU */
    n=descLU[BLACSn];
    ctxt=descLU[BLACSctxt];
    blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
    if(nprow==-1) nprow=0;
    if(!partially_factored_LU) {
      if(myid==0) std::cout << "Matrix was not previously partially LU-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!schur_LU) {
      if(myid==0) std::cout << "Exact Schur complement was not previously computed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
  }
  nrhs=descB[BLACSn];
  nb=descB[BLACSnb];

  if(n-var_schur+1!=descX[BLACSm]) {
    std::cout << "X must have the same row count as the Schur complement!" << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }
  if(n-var_schur+1!=descB[BLACSm]) {
    std::cout << "B must have the same row count as the Schur complement!" << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }

  if(!use_HSS) {
    if(descB[BLACSm]!=descX[BLACSm]) {
      std::cout << "X, B have different row counts! B:" << descB[BLACSm] << "; X:" << descX[BLACSm] << std::endl;
      std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(descLU[BLACSmb]!=descX[BLACSmb]||descLU[BLACSmb]!=descB[BLACSmb]||descB[BLACSmb]!=descX[BLACSmb]) {
      std::cout << "A, X, B have different block sizes! A:" << descLU[BLACSmb] << "; B" << descB[BLACSmb] << "; X:" << descX[BLACSmb] << std::endl;
      std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
  }
  if(descB[BLACSn]!=descX[BLACSn]) {
    std::cout << "X, B have different column counts! B:" << descB[BLACSn] << "; X:" << descX[BLACSn] << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }

  nfs=var_schur-1;
  ns=n-nfs;

  if(use_HSS) {
    if(np==1) {
      /* Sequential HSS Schur complement product */
      if(msg) std::cout << "Sequential Schur complement product..." << std::endl;

      /* The input vectors are ns-sized but the product routine expects n-sized
       * vectors so we make a copy. This should be improved later.
       */
      tmpB=new T[n*nrhs]();
      tmpX=new T[n*nrhs]();
      lacpy('N',ns,nrhs,B,ns,&tmpB[nfs],n);
      lacpy('N',ns,nrhs,X,ns,&tmpX[nfs],n);
      /* B' = beta B + alpha op(F22) * X */
      htree.HSS_seq_Product(Trans,alpha,tmpX,nrhs,beta,tmpB,1);
      lacpy('N',ns,nrhs,&tmpB[nfs],n,B,ns);
      delete[] tmpB;
      delete[] tmpX;

      if(htree.rpsi) {
        if(Trans=='N') {
          /* B" = B' - alpha theta Vtilda_k^* psi X = B + alpha (F22 - theta Vtilda_k^* psi) X = B + alpha Schur X.
           * Compute the product with psi first.
           */
          tmpX=new T[htree.rpsi*nrhs]();
          gemm('N','N',htree.rpsi,nrhs,htree.cpsi,ONE,htree.psi,htree.rpsi,X,ns,ZERO,tmpX,htree.rpsi);
          tmpB=new T[htree.ctheta*nrhs]();
          gemm('C','N',htree.ctheta,nrhs,htree.rpsi,ONE,htree.nodes[htree.knode]->Vt,htree.rpsi,tmpX,htree.rpsi,ZERO,tmpB,htree.ctheta);
          delete[] tmpX;
          gemm('N','N',ns,nrhs,htree.ctheta,-alpha,htree.theta,htree.rtheta,tmpB,htree.ctheta,ONE,B,ns);
          delete[] tmpB;
        } else {
          /* B" = B' - alpha psi^* Vtilda_k theta^* X = B + alpha (F22 - theta Vtilda_k^* psi) X = B + alpha Schur^* X.
           * Compute the product with psi first.
           */
          tmpX=new T[htree.ctheta*nrhs]();
          gemm('C','N',htree.ctheta,nrhs,htree.rtheta,ONE,htree.theta,htree.rtheta,X,ns,ZERO,tmpX,htree.ctheta);
          tmpB=new T[htree.rpsi*nrhs]();
          gemm('N','N',htree.rpsi,nrhs,htree.ctheta,ONE,htree.nodes[htree.knode]->Vt,htree.rpsi,tmpX,htree.ctheta,ZERO,tmpB,htree.rpsi);
          delete[] tmpX;
          gemm('C','N',ns,nrhs,htree.rpsi,-alpha,htree.psi,htree.rpsi,tmpB,htree.rpsi,ONE,B,ns);
          delete[] tmpB;
        }
      }

      /* Flops */
      // TODO

    } else {
      /* Parallel HSS Schur complement product */
      if(msg) std::cout << "Parallel HSS Schur complement product..." << std::endl;

      ctxt=htree.nodes[htree.dim_tree-1]->context;
      blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);

      /* The input vectors are ns-sized but the product routine expects n-sized
       * vectors so we make a copy. This should be improved later.
       */
      if(myrow>=0 && mycol>=0) {
        locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
        locc=numroc_(&nrhs,&nb,&mycol,&IZERO,&npcol);
        Bn=new T[locr*locc]();
        Xn=new T[locr*locc]();
        idummy=std::max(locr,1);
        descinit_(desc_Bn,&n,&nrhs,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
        descinit_(desc_Xn,&n,&nrhs,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
      } else {
        Bn=NULL;
        Xn=NULL;
        descset_(desc_Bn,&n,&nrhs,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
        descset_(desc_Xn,&n,&nrhs,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
      }

      pgemr2d(ns,nrhs,B,IONE,IONE,descB,Bn,nfs+1,IONE,desc_Bn,htree.nodes[htree.dim_tree-1]->allctxt);
      pgemr2d(ns,nrhs,X,IONE,IONE,descX,Xn,nfs+1,IONE,desc_Xn,htree.nodes[htree.dim_tree-1]->allctxt);

      /* B' = beta B + alpha F22 * X */
      htree.HSS_par_Product(Trans,alpha,Xn,desc_Xn,nrhs,beta,Bn,desc_Bn,&ginfo,1);

      if(Trans=='N') {
        /* B" = B' - alpha theta Vtilda_k^* psi X = B + alpha (F22 - theta Vtilda_k^* psi) X = B + alpha Schur X.
         * Compute the product with psi first.
         */
        if(myrow>=0 && mycol>=0) {
          locr=numroc_(&htree.rpsi,&nb,&myrow,&IZERO,&nprow);
          locc=numroc_(&nrhs,&nb,&mycol,&IZERO,&npcol);
          tmpX=new T[locr*locc]();
          idummy=std::max(locr,1);
          descinit_(desc_tmpX,&htree.rpsi,&nrhs,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
          pgemm('N','N',htree.rpsi,nrhs,htree.cpsi,ONE,htree.psi,IONE,IONE,htree.desc_psi,Xn,nfs+1,IONE,desc_Xn,ZERO,tmpX,IONE,IONE,desc_tmpX);

          locr=numroc_(&htree.ctheta,&nb,&myrow,&IZERO,&nprow);
          idummy=std::max(locr,1);
          tmpB=new T[locr*locc]();
          descinit_(desc_tmpB,&htree.ctheta,&nrhs,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
          pgemm('C','N',htree.ctheta,nrhs,htree.rpsi,ONE,htree.nodes[htree.knode]->Vt,IONE,IONE,htree.nodes[htree.knode]->descVt,tmpX,IONE,IONE,desc_tmpX,ZERO,tmpB,IONE,IONE,desc_tmpB);
          delete[] tmpX;

          pgemm('N','N',ns,nrhs,htree.ctheta,-alpha,htree.theta,IONE,IONE,htree.desc_theta,tmpB,IONE,IONE,desc_tmpB,ONE,Bn,nfs+1,IONE,desc_Bn);
          delete[] tmpB;
        }
      } else {
        /* B" = B' - alpha psi^* Vtilda_k theta^* X = B + alpha (F22 - theta Vtilda_k^* psi) X = B + alpha Schur^* X.
         * Compute the product with psi first.
         */
        if(myrow>=0 && mycol>=0) {
          locr=numroc_(&htree.ctheta,&nb,&myrow,&IZERO,&nprow);
          locc=numroc_(&nrhs,&nb,&mycol,&IZERO,&npcol);
          tmpX=new T[locr*locc]();
          idummy=std::max(locr,1);
          descinit_(desc_tmpX,&htree.ctheta,&nrhs,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
          pgemm('C','N',htree.ctheta,nrhs,htree.rtheta,ONE,htree.theta,IONE,IONE,htree.desc_theta,Xn,nfs+1,IONE,desc_Xn,ZERO,tmpX,IONE,IONE,desc_tmpX);

          locr=numroc_(&htree.rpsi,&nb,&myrow,&IZERO,&nprow);
          idummy=std::max(locr,1);
          tmpB=new T[locr*locc]();
          descinit_(desc_tmpB,&htree.rpsi,&nrhs,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
          pgemm('N','N',htree.rpsi,nrhs,htree.ctheta,ONE,htree.nodes[htree.knode]->Vt,IONE,IONE,htree.nodes[htree.knode]->descVt,tmpX,IONE,IONE,desc_tmpX,ZERO,tmpB,IONE,IONE,desc_tmpB);
          delete[] tmpX;

          pgemm('C','N',ns,nrhs,htree.rpsi,-alpha,htree.psi,IONE,IONE,htree.desc_psi,tmpB,IONE,IONE,desc_tmpB,ONE,Bn,nfs+1,IONE,desc_Bn);
          delete[] tmpB;
        }
      }

      pgemr2d(ns,nrhs,Bn,nfs+1,IONE,desc_Bn,B,IONE,IONE,descB,htree.nodes[htree.dim_tree-1]->allctxt);
      delete[] Bn;
      delete[] Xn;

      /* Flops */
      // TODO

    }
  } else {
    if(np==1) {
      /* Schur complement product with LAPACK */
      if(msg) std::cout << "Schur complement product with LAPACK..." << std::endl;

      gemm('N','N',ns,nrhs,ns,alpha,&LU[nfs*n+nfs],n,X,ns,beta,B,ns);

      /* Flops */
      flops_prod+=2*static_cast<double>(n-var_schur+1)*static_cast<double>(n-var_schur+1)*static_cast<double>(nrhs);

    } else if(nprow>0){
      /* Schur complement product with ScaLAPACK */
      if(msg) std::cout << "Schur complement product with ScaLAPACK..." << std::endl;

      pgemm('N','N',ns,nrhs,ns,alpha,LU,var_schur,var_schur,descLU,X,IONE,IONE,descX,beta,B,IONE,IONE,descB);

      /* Flops */
      flops_prod+=2*static_cast<double>(ns)*static_cast<double>(ns)*static_cast<double>(nrhs)/static_cast<double>(nprow)/static_cast<double>(npcol);
    }
  }

  tend=MPI_Wtime();
#if defined(WITH_PAPI)
  PAPI_flops(&rtime,&ptime,&flpops2,&mflops);
  flpops=static_cast<double>(flpops2-flpops1);
  MPI_Allreduce((void *) &flpops,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
  MPI_Allreduce((void *) &flpops,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
  MPI_Allreduce((void *) &flpops,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
  av/=static_cast<double>(np);
  if(msg) std::cout << "PAPI GFlop count: min= " << mi/1.0e9 << "; avg= " << av/1.0e9 << "; max= " << ma/1.0e9 << std::endl;
#endif
  if(msg) std::cout << "Done in " << tend-tstart << "s" << std::endl << std::endl;
  time_solve+=tend-tstart;
}

template<typename T, typename S> S StrumpackDensePackage<T,S>::refine(T *A, int *descA, T *X, int *descX, T *B, int *descB) {
  /* Iterative refinement */
  int ctxt;
  int locr, locc;
  int n;
  int nb;
  int myid;
  int nrhs;
  int myrow, mycol, nprow, npcol;
  int go;
  bool msg, old_verbose, old_compress;
  S err;
  T *R, *DX;
  int steps;
  double tstart, tend;

  tstart=MPI_Wtime();

  MPI_Comm_rank(comm,&myid);
  msg=verbose && myid==0;

  if(!check_descriptor("X",descX,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return -1;
  }
  if(!check_descriptor("B",descB,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return -1;
  }
  if(descA[BLACSm]!=descX[BLACSm]||descA[BLACSm]!=descB[BLACSm]||descB[BLACSm]!=descX[BLACSm]) {
    std::cout << "A, X, B have different row counts! A:" << descA[BLACSm] << "; B" << descB[BLACSm] << "; X:" << descX[BLACSm] << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return -1;
  }
  if(descA[BLACSmb]!=descX[BLACSmb]||descA[BLACSmb]!=descB[BLACSmb]||descB[BLACSmb]!=descX[BLACSmb]) {
    std::cout << "A, X, B have different block sizes! A:" << descA[BLACSmb] << "; B" << descB[BLACSmb] << "; X:" << descX[BLACSmb] << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return -1;
  }
  if(descB[BLACSn]!=descX[BLACSn]) {
    std::cout << "X, B have different column counts! B:" << descB[BLACSn] << "; X:" << descX[BLACSn] << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return -1;
  }

  n=descA[BLACSn];
  nb=descB[BLACSnb];
  ctxt=descA[BLACSctxt];
  nrhs=descB[BLACSn];
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);

  if(msg) std::cout << "Iterative refinement of the solution..." << std::endl;

  old_verbose=verbose;
  old_compress=use_HSS;
  verbose=false; /* For routines called inside the refinement */
  err=1; /* For processes out the grid */
  err=check_solution(A,descA,X,descX,B,descB);
  steps=0;
  go=err<tol_IR?0:1;
  MPI_Bcast((void *) &go,IONE,MPI_INTEGER,IZERO,comm);
  if(msg) std::cout << "Step " << steps << ": err = " << err << std::endl;
  if(nprow>0) {
    locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
    locc=numroc_(&nrhs,&nb,&mycol,&IZERO,&npcol);
    R=new T[locr*locc]();
    DX=new T[locr*locc]();
  } else if(np==1) {
    R=new T[n*nrhs]();
    DX=new T[n*nrhs]();
  }
  while(steps<steps_IR && go) {
    if(nprow>0)
      lacpy('N',locr,locc,B,locr,R,locr);
    else if(np==1)
      lacpy('N',n,nrhs,B,n,R,n);
    use_HSS=fast_IR;
    product('N',NONE,A,descA,X,descX,ONE,R,descB);
    use_HSS=old_compress;
    solve(DX,descX,R,descB);
    if(nprow>0)
      pgeadd('N',n,nrhs,ONE,DX,IONE,IONE,descX,ONE,X,IONE,IONE,descX);
    else if(np==1)
      for(int i=0;i<n*nrhs;i++)
        X[i]+=DX[i];
    err=check_solution(A,descA,X,descX,B,descB);
    steps++;
    if(msg) std::cout << "Step " << steps << ": err = " << err << std::endl;
    /* Processes out of the grid dont know the value of err so we need to stop them */
    if(err<tol_IR)
      go=0;
    MPI_Bcast((void *) &go,IONE,MPI_INTEGER,IZERO,comm);
  }
  if(nprow>0 || np==1) {
    delete[] R;
    delete[] DX;
  }

  tend=MPI_Wtime();
  if(msg) std::cout << "Done in " << tend-tstart << "s" << std::endl << std::endl;

  verbose=old_verbose;
  use_HSS=old_compress;
  return err;
}

template<typename T, typename S> S StrumpackDensePackage<T,S>::check_solution(T *A, int *descA, T *X, int *descX, T *B, int *descB) {
  /* Compute a relative residual */
  int ctxt;
  int locr, locc;
  int n;
  int myid;
  int nb;
  int nrhs;
  int myrow, mycol, nprow, npcol;
  bool msg;
  S err;
  S *work;
  T *R;
  double tstart, tend;

  tstart=MPI_Wtime();

  MPI_Comm_rank(comm,&myid);
  msg=verbose && myid==0;

  if(!check_descriptor("X",descX,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return -1;
  }
  if(!check_descriptor("B",descB,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return -1;
  }
  if(descA[BLACSm]!=descX[BLACSm]||descA[BLACSm]!=descB[BLACSm]||descB[BLACSm]!=descX[BLACSm]) {
    std::cout << "A, X, B have different row counts! A:" << descA[BLACSm] << "; B" << descB[BLACSm] << "; X:" << descX[BLACSm] << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return -1;
  }
  if(descA[BLACSmb]!=descX[BLACSmb]||descA[BLACSmb]!=descB[BLACSmb]||descB[BLACSmb]!=descX[BLACSmb]) {
    std::cout << "A, X, B have different block sizes! A:" << descA[BLACSmb] << "; B" << descB[BLACSmb] << "; X:" << descX[BLACSmb] << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return -1;
  }
  if(descB[BLACSn]!=descX[BLACSn]) {
    std::cout << "X, B have different column counts! B:" << descB[BLACSn] << "; X:" << descX[BLACSn] << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return -1;
  }

  n=descA[BLACSn];
  nb=descB[BLACSnb];
  ctxt=descA[BLACSctxt];
  nrhs=descB[BLACSn];
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);

  if(msg) std::cout << "Checking solution..." << std::endl;

  if(np==1) {
    /* Check with BLAS/LAPACK */
    work=new S[n];
    R=new T[n*nrhs];
    lacpy('N',n,nrhs,B,n,R,n);
    err=lange('F',n,nrhs,R,n,work);
    gemm('N','N',n,nrhs,n,NONE,A,n,X,n,ONE,R,n);
    err=lange('F',n,nrhs,R,n,work)/err;
    delete[] work;
    delete[] R;
    flops_prod+=2.0*static_cast<double>(n)*static_cast<double>(n)*static_cast<double>(nrhs);
  } else if(nprow>0){
    /* Check with PBLAS/ScaLAPACK */
    locr=numroc_(&n,&nb,&myrow,&IZERO,&nprow);
    locc=numroc_(&nrhs,&nb,&mycol,&IZERO,&npcol);
    R=new T[locr*locc];
    pgeadd('N',n,nrhs,ONE,B,IONE,IONE,descB,ZERO,R,IONE,IONE,descB);
    work=new S[locr];
    err=plange('F',n,nrhs,R,IONE,IONE,descB,work);
    pgemm('N','N',n,nrhs,n,NONE,A,IONE,IONE,descA,X,IONE,IONE,descX,ONE,R,IONE,IONE,descB);
    err=plange('F',n,nrhs,R,IONE,IONE,descB,work)/err;
    delete[] work;
    delete[] R;
    flops_prod+=2.0*static_cast<double>(n)*static_cast<double>(n)*static_cast<double>(nrhs)/static_cast<double>(nprow)/static_cast<double>(npcol);
  }
  if(msg) std::cout << "Relative residual: ||A*x-b||_fro/||b||_fro=" << err << std::endl;

  tend=MPI_Wtime();
  time_prod+=tend-tstart;
  if(msg) std::cout << "Done in " << tend-tstart << "s" << std::endl << std::endl;
  return err;
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::product(char Trans, T alpha, T *A, int *descA, T *B, int *descB, T beta, T *C, int *descC) {
  /* Matrix-vector product */
  int ctxt;
  int n;
  int myid;
  int nrhs;
  int myrow, mycol, nprow, npcol;
  bool msg;
  double tstart, tend;

  tstart=MPI_Wtime();

  MPI_Comm_rank(comm,&myid);
  msg=verbose && myid==0;

  if(!check_descriptor("B",descB,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }
  if(!check_descriptor("C",descC,false)) {
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }
  if(descA[BLACSm]!=descC[BLACSm]||descA[BLACSm]!=descB[BLACSm]||descB[BLACSm]!=descC[BLACSm]) {
    std::cout << "A, B, C have different row counts! A:" << descA[BLACSm] << "; B" << descB[BLACSm] << "; C:" << descC[BLACSm] << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }
  if(descA[BLACSmb]!=descC[BLACSmb]||descA[BLACSmb]!=descB[BLACSmb]||descB[BLACSmb]!=descC[BLACSmb]) {
    std::cout << "A, B, C have different block sizes! A:" << descA[BLACSmb] << "; B" << descB[BLACSmb] << "; C:" << descC[BLACSmb] << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }
  if(descB[BLACSn]!=descC[BLACSn]) {
    std::cout << "B, C have different column counts! B:" << descB[BLACSn] << "; C:" << descC[BLACSn] << std::endl;
    std::cout << "Aborting..." << std::endl << std::endl;
    return;
  }

  n=descA[BLACSn];
  ctxt=descA[BLACSctxt];
  nrhs=descB[BLACSn];
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
  if(nprow==-1) nprow=0;
  msg=verbose && myrow==0 && mycol==0;

  if(msg) std::cout << "Matrix-vector product..." << std::endl;
  if(use_HSS) {
    if(!compressed) {
      if(!myid) std::cout << "Matrix was not previously compressed." << std::endl;
      if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(np==1) {
      htree.HSS_seq_Product(Trans,alpha,B,nrhs,beta,C,0);
      /* Flops */
      flops_prod+=htree.getFlopsProduct(nrhs);
    } else {
      htree.HSS_par_Product(Trans,alpha,B,descB,nrhs,beta,C,descC,&ginfo,0);
      /* Flops */
      flops_prod+=htree.getFlopsProduct(nrhs);
    }
  } else {
    if(np==1) {
      gemm(Trans,'N',n,nrhs,n,alpha,A,n,B,n,beta,C,n);
      flops_prod+=2.0*static_cast<double>(n)*static_cast<double>(n)*static_cast<double>(nrhs);
    } else if(nprow>0) {
      pgemm(Trans,'N',n,nrhs,n,alpha,A,IONE,IONE,descA,B,IONE,IONE,descB,beta,C,IONE,IONE,descC);
      flops_prod+=2.0*static_cast<double>(n)*static_cast<double>(n)*static_cast<double>(nrhs)/static_cast<double>(nprow)/static_cast<double>(npcol);
    }
  }

  tend=MPI_Wtime();
  if(msg) std::cout << "Done in " << tend-tstart << "s" << std::endl << std::endl;
  time_prod+=tend-tstart;
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::extract(T* A, int *descA, T *B, int *descB, int *I, int nI, int *J, int nJ) {
  int myid;
  int ctxt;
  int nprow, npcol;
  int myrow, mycol;
  bool msg;
  int i,j;

  MPI_Comm_rank(comm,&myid);

  ctxt=descA[BLACSctxt];
  blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
  if(nprow==-1) nprow=0;
  msg=verbose && myid==0;

  if(msg) std::cout << "Element extraction..." << std::endl;
  if(use_HSS) {
    if(!compressed) {
      if(!myid) std::cout << "Matrix was not previously compressed." << std::endl;
      if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(np==1) {
      htree.HSS_seq_Extraction(B,I,nI,J,nJ);
      //TODO: flops
    } else {
      htree.HSS_par_Extraction(B,descB,I,nI,J,nJ,&ginfo);
      //TODO: flops
    }
  } else {
    if(np==1) {
      for(i=0;i<nI;i++)
        for(j=0;j<nJ;j++)
          B[i+j*nI]=A[I[i]-1+(J[j]-1)*n];
    } else if(nprow>0) {
      /* Parallel case: nothing smart here. We communicate element
       * by element so communications most likely don't overlap
       * and latency is probably terrible. This case is mostly
       * for test purposes.
       */
      for(i=0;i<nI;i++)
        for(j=0;j<nJ;j++)
          pgeadd('N',IONE,IONE,ONE,A,I[i],J[j],descA,ZERO,B,i+1,j+1,descB);
    }
  }
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::print_statistics() {
  int myid;
  double mi, av, ma;
#if defined(MEMTRACK)
  double tmp;
#endif

  myid=0;
  MPI_Comm_rank(comm,&myid);
  double total_comp_time_diff=0;

  if(!myid) std::cout << "Statistics: " << std::endl << std::endl;
  /* Compression */
  if(compressed) {
    /* Time */
    if(!myid) {
      printf("Time spent in compression (s):                         ");
      print_call(time_compr);
      printf("\n");
      printf("Time spent in initial distrib (s):                     ");
      print_call(htree.time_Dist);
      printf("\n");
    }
    total_comp_time_diff=time_compr-htree.time_Dist;

    MPI_Allreduce((void *) &htree.time_Rand,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &htree.time_Rand,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &htree.time_Rand,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    av/=static_cast<double>(np);
    if(!myid) {
      printf("Random number generation (s):      ");
      print_min_ave_max(mi,av,ma);
    }
    total_comp_time_diff-=av;


    MPI_Allreduce((void *) &htree.time_GEMM,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &htree.time_GEMM,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &htree.time_GEMM,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    av/=static_cast<double>(np);
    if(!myid) {
      printf("Global sample GEMMS (s):           ");
      print_min_ave_max(mi,av,ma);
    }
    total_comp_time_diff-=av;

    MPI_Allreduce((void *) &htree.time_Amult,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &htree.time_Amult,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &htree.time_Amult,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    av/=static_cast<double>(np);
    if(!myid) {
      printf("Amult routine (s):                 ");
      print_min_ave_max(mi,av,ma);
    }
    total_comp_time_diff-=av;

    MPI_Allreduce((void *) &htree.time_samples_GEMM,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &htree.time_samples_GEMM,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &htree.time_samples_GEMM,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    av/=static_cast<double>(np);
    if(!myid) {
      printf("Local sample GEMMS (s):            ");
      print_min_ave_max(mi,av,ma);
    }
    total_comp_time_diff-=av;

    MPI_Allreduce((void *) &htree.time_ID,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &htree.time_ID,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &htree.time_ID,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    av/=static_cast<double>(np);
    if(!myid) {
      printf("Interpolative decompositions (s):  ");
      print_min_ave_max(mi,av,ma);
    }
    total_comp_time_diff-=av;

    MPI_Allreduce((void *) &htree.time_Elements,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &htree.time_Elements,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &htree.time_Elements,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    av/=static_cast<double>(np);
    if(!myid) {
      printf("Element extraction (s):            ");
      print_min_ave_max(mi,av,ma);
    }
    total_comp_time_diff-=av;

    MPI_Allreduce((void *) &htree.time_Barrier,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &htree.time_Barrier,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &htree.time_Barrier,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    av/=static_cast<double>(np);
    if(!myid) {
      printf("Barriers (s):                      ");
      print_min_ave_max(mi,av,ma);
    }
    total_comp_time_diff-=av;

    MPI_Allreduce((void *) &htree.time_Ptc,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &htree.time_Ptc,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &htree.time_Ptc,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    av/=static_cast<double>(np);
    if(!myid) {
      printf("participate_to_calls (s):          ");
      print_min_ave_max(mi,av,ma);
    }
    total_comp_time_diff-=av;

    MPI_Allreduce((void *) &htree.time_Recv,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &htree.time_Recv,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &htree.time_Recv,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    av/=static_cast<double>(np);
    if(!myid) {
      printf("MPI_Recv (s):                      ");
      print_min_ave_max(mi,av,ma);
    }
    total_comp_time_diff-=av;

    MPI_Allreduce((void *) &htree.time_MPI_Allgather,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &htree.time_MPI_Allgather,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &htree.time_MPI_Allgather,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    av/=static_cast<double>(np);
    if(!myid) {
      printf("MPI_Allgather (s):                 ");
      print_min_ave_max(mi,av,ma);
    }
    total_comp_time_diff-=av;

    if(!myid) {
      printf("Unaccounted compression time (s):                      ");
      print_call(total_comp_time_diff);
      printf("\n");
    }

    /* Flops */
    MPI_Allreduce((void *) &flops_compr,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &flops_compr,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &flops_compr,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    if(!myid) {
      printf("Flops for compression (GFlops):    \n");
      print_min_ave_max_total(mi/1.0e9,av/1.0e9/static_cast<double>(np),ma/1.0e9,av/1.0e9);
    }
    av/=static_cast<double>(np);

    /* HSS factors */
    MPI_Allreduce((void *) &mem_HSS,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &mem_HSS,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &mem_HSS,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    if(!myid) {
      printf("HSS factors (MB):                  \n");
      print_min_ave_max_total(mi/1024/1024,av/1024/1024/static_cast<double>(np),ma/1024/1024,av/1024/1024);
    }
    av/=static_cast<double>(np);

    /* Memory */
#if defined(MEMTRACK)
    tmp=av; /* HSS factors */
    MPI_Allreduce((void *) &mem_compr,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &mem_compr,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &mem_compr,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    if(!myid) std::cout << "Total memory usage (MB): min= " << mi/1024/1024 << "; avg= " << av/1024/1024/static_cast<double>(np) << "; max= " << ma/1024/1024 << "; total= " << av/1024/1024 << std::endl;
    av/=static_cast<double>(np);
    if(!myid) std::cout << "Random vectors and samples: " << static_cast<double>(n)*eff_rand_HSS*sizeof(T)*4.0/1024.0/1024.0/np << " MB per process." << std::endl;
    if(np>1) {
      if(!myid) std::cout << "Distributed matrix: " << static_cast<double>(n)*static_cast<double>(n)*sizeof(T)/1024.0/1024.0/np << " MB per process." << std::endl;
      if(!myid) std::cout << "Average overhead (total-samples-factors-matrix): " << (av-tmp-static_cast<double>(n)*(n+eff_rand_HSS*4.0)*sizeof(T)/np)/1024.0/1024.0 << " MB per process." << std::endl;
    } else
      if(!myid) std::cout << "Average overhead (total-samples-factors): " << (av-tmp-static_cast<double>(n)*eff_rand_HSS*sizeof(T)*4.0/np)/1024.0/1024.0 << " MB per process." << std::endl;
#endif

    if(!myid) std::cout << std::endl;
  }

  /* Factorization */
  if(factored_HSS || partially_factored_HSS) {
    /* Time */
    if(!myid) std::cout << "Time spent in factorization: " << time_fact << " (s)" << std::endl;

    /* Flops */
    MPI_Allreduce((void *) &flops_fact,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &flops_fact,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &flops_fact,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    if(!myid) {
      printf("Flops for factorization (GFlops):\n");
      print_min_ave_max_total(mi/1.0e9,av/1.0e9/static_cast<double>(np),ma/1.0e9,av/1.0e9);
    }
    av/=static_cast<double>(np);

    /* Factors */
    MPI_Allreduce((void *) &mem_ULV,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &mem_ULV,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &mem_ULV,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    if(!myid) {
      printf("ULV factors (MB):\n");
      print_min_ave_max_total(mi/1024/1024,av/1024/1024/static_cast<double>(np),ma/1024/1024,av/1024/1024);
    }
    av/=static_cast<double>(np);

    /* Memory */
#if defined(MEMTRACK)
    tmp=av; /* ULV factors */
    MPI_Allreduce((void *) &mem_fact,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &mem_fact,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &mem_fact,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    if(!myid) std::cout << "Total memory usage (MB): min= " << mi/1024/1024 << "; avg= " << av/1024/1024/static_cast<double>(np) << "; max= " << ma/1024/1024 << std::endl;
    av/=static_cast<double>(np);
    if(!myid) std::cout << "Average overhead (total-factors): " << (av-tmp)/1024.0/1024.0 << " MB per process." << std::endl;
#endif
  } else if(factored_LU || partially_factored_LU) {
    /* Flops */
    MPI_Allreduce((void *) &flops_fact,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &flops_fact,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &flops_fact,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    if(!myid) {
      printf("Flops for factorization (GFlops):\n");
      print_min_ave_max_total(mi/1.0e9,av/1.0e9/static_cast<double>(np),ma/1.0e9,av/1.0e9);
    }
    av/=static_cast<double>(np);

    /* Factors */
    MPI_Allreduce((void *) &mem_ULV,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
    MPI_Allreduce((void *) &mem_ULV,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
    MPI_Allreduce((void *) &mem_ULV,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
    if(!myid) {
      printf("LU factors (MB):\n");
      print_min_ave_max_total(mi/1024/1024,av/1024/1024/static_cast<double>(np),ma/1024/1024,av/1024/1024);
    }
    av/=static_cast<double>(np);
  }
  if(!myid) std::cout << std::endl;

  /* Solution */
  if(!myid) std::cout << "Cumulated time in solution steps: " << time_solve << " (s)" << std::endl;
  MPI_Allreduce((void *) &flops_solve,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
  MPI_Allreduce((void *) &flops_solve,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
  MPI_Allreduce((void *) &flops_solve,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
  if(!myid) {
    printf("Cumulated flops in solution steps (GFlops):\n");
    print_min_ave_max_total(mi/1.0e9,av/1.0e9/static_cast<double>(np),ma/1.0e9,av/1.0e9);
  }
  av/=static_cast<double>(np);

  /* Product */
  if(!myid) std::cout << "Cumulated time in matrix-vector products: " << time_prod << " (s)" << std::endl;
  MPI_Allreduce((void *) &flops_prod,(void *) &mi,IONE,MPI_DOUBLE,MPI_MIN,comm);
  MPI_Allreduce((void *) &flops_prod,(void *) &ma,IONE,MPI_DOUBLE,MPI_MAX,comm);
  MPI_Allreduce((void *) &flops_prod,(void *) &av,IONE,MPI_DOUBLE,MPI_SUM,comm);
  if(!myid) {
    printf("Cumulated flops in matrix-vector products (GFlops):\n");
    print_min_ave_max_total(mi/1.0e9,av/1.0e9/static_cast<double>(np),ma/1.0e9,av/1.0e9);
  }
  av/=static_cast<double>(np);

#if defined(WRITESTATS)
  htree.writeStats();
#endif
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::print_min_ave_max(double min, double ave, double max) {
  printf("min= ");
  print_call(min);
  printf("; ");

  printf("ave= ");
  print_call(ave);
  printf("; ");

  printf("max= ");
  print_call(max);
  printf("\n");
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::print_min_ave_max_total(double min, double ave, double max, double total) {
  printf("min= ");
  print_call(min);
  printf("; ");

  printf("ave= ");
  print_call(ave);
  printf("; ");

  printf("max= ");
  print_call(max);
  printf("; ");

  printf("total= ");
  print_call(total);
  printf("\n");
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::print_call(double val) {
  if(val <= 0)
    printf("%8g",val);
  else if(val >= 10000)
    printf("%8.2e",val);
  else if(val >= 1000)
    printf("%8.3f",val);
  else if(val >= 100)
    printf("%8.4f",val);
  else if(val >= 10)
    printf("%8.5f",val);
  else if(val >= 1)
    printf("%8.6f",val);
  else if(val >= 0.1)
    printf("%8.6f",val);
  else if(val >= 0.01)
    printf("%8.6f",val);
  else if(val >= 0.001)
    printf("%8.6f",val);
  else
    printf("%8.2e",val);
}



template<typename T, typename S> bool StrumpackDensePackage<T,S>::check_descriptor(std::string str, int *desc, bool square) {
  /* Checking BLACS descriptor desc */
  bool check=true;

  if(desc[BLACSdtype]!=1) {
    std::cout << "desc" << str << " has a wrong descriptor type; " << desc[BLACSdtype] << "instead of 1" << std::endl;
    check=false;
  }

  if(desc[BLACSm]!=desc[BLACSn] && square) {
    std::cout << "desc" << str << " describes a non-square matrix: " << desc[BLACSm] << "x" << desc[BLACSn] << "; matrix must be square." << std::endl;
    check=false;
  }

  if(desc[BLACSmb]!=desc[BLACSnb]) {
    std::cout << "desc" << str << " describes a grid with non-square blocks: mb=" << desc[BLACSmb] << " != nb=" << desc[BLACSnb] << "; blocks must be square (because we use pxgetrf internally)." << std::endl;
    check=false;
  }

  if(desc[BLACSlld]<1) {
    std::cout << "desc" << str << " describes a wrong leading dimension: lld=" << desc[BLACSlld] << "; must be >=1." << std::endl;
    check=false;
  }

  return check;
}

template<typename T, typename S> void access (void * obj, int *I, int *J, T *B, int *descB) {
  /* Element access routine for explicit matrices.
   * The input matrix A passed by the user is in StrumpackDensePackage.A_in.
   * The SDP object is in the "obj" pointer.
   * This routine extracts B=A(I,J), with B distributed onto the grid defined
   * by descB. Processes out of the grid get a dummy descriptor descB (that
   * defines sizes and block sizes).
   *
   * The routine works as follows:
   *  - We check wheter I and J are contiguous sets of indices (intervals).
   *  - If I and J are intervals:
   *     - We do the extraction in one shot with pgemr2d.
   *  - Otherwise
   *     - We extract A(I,J) to a temporary matrix Btmp distributed onto
   *       the same grid as A_in (StrumpackDensePackage.descA_in).
   *     - We redistribute Btmp onto the grid described by descB.
   */

  int i, j;
  int ctxtglob;
  int mA, nA;
  int ctxtA;
  int nprowA, npcolA, myrowA, mycolA;
  int rsrcA, csrcA;
  int mbA, nbA;
  int mB, nB;
  int idummy, ierr;
  int *descA, desc[BLACSCTXTSIZE];
  bool contiguous;
  gridinfo ginfo;
  T *A, *Btmp;
  StrumpackDensePackage<T,S> *sdp;

  /* Information for the object handle */
  sdp=static_cast<StrumpackDensePackage<T,S>*>(obj);
  A=sdp->A_in;
  descA=sdp->descA_in;
  ctxtglob=sdp->ctxtglob;

  ctxtA=descA[1];
  mA=descA[2];
  nA=descA[3];
  mbA=descA[4];
  nbA=descA[5];
  rsrcA=descA[6];
  csrcA=descA[7];
  blacs_gridinfo_(&ctxtA,&nprowA,&npcolA,&myrowA,&mycolA);

  mB=descB[2];
  nB=descB[3];

  /* Check whether both I and J are intervals */
  contiguous=true;
  for(i=1;i<mB;i++) {
    if(I[i]!=I[i-1]+1) {
      contiguous=false;
      break;
    }
  }
  if(contiguous) {
    for(j=1;j<nB;j++) {
      if(J[j]!=J[j-1]+1) {
        contiguous=false;
        break;
      }
    }
  }

  if(contiguous) {
    /* Redistribution in one shot with pgemr2d */
    pgemr2d(mB,nB,A,I[0],J[0],descA,B,IONE,IONE,descB,ctxtglob);
  } else {
    /* 1/ Create an mB x nB matrix on the grid defined by descA */
    if(myrowA>=0 && mycolA>=0) {
      idummy=std::max(1,numroc_(&mB,&mbA,&myrowA,&rsrcA,&nprowA));
      descinit_(desc,&mB,&nB,&mbA,&nbA,&rsrcA,&csrcA,&ctxtA,&idummy,&ierr);
    } else {
      Btmp=NULL;
      descset_(desc,&mB,&nB,&mbA,&nbA,&rsrcA,&csrcA,&INONE,&IONE);
    }
    /* 2/ Extract A(I,J) into Btmp */
    if(myrowA>=0 && mycolA>=0) {
      ginfo.ctxt=ctxtA;
      ginfo.nprow=nprowA;
      ginfo.npcol=npcolA;
      ginfo.myrow=myrowA;
      ginfo.mycol=mycolA;
      ginfo.mb=mbA;
      ginfo.nb=nbA;
      parExtractSubmatrix<T,S>(A,mA,nA,descA,I,mB,J,nB,&Btmp,&ginfo);
    }
    /* 3/ Redistribute Btmp onto the grid defined by descB */
    pgemr2d(mB,nB,Btmp,IONE,IONE,desc,B,IONE,IONE,descB,ctxtglob);
    delete[] Btmp;
  }
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::extract_schur(T *B, int *descB, int *I, int nI, int *J, int nJ) {
  int myid;
  int ctxt;
  int nprow, npcol;
  int myrow, mycol;
  int locr, locc;
  int nb;
  int idummy, ierr;
  bool msg;
  int i,j;
  int desc[BLACSCTXTSIZE], desc_tmpB[BLACSCTXTSIZE], desc_tmp_theta[BLACSCTXTSIZE], desc_tmp_psi[BLACSCTXTSIZE];
  int *tmpI, *tmpJ;
  gridinfo tmpinfo;
  T *tmp_theta, *tmp_psi, *tmp, *tmpB;

  MPI_Comm_rank(comm,&myid);

  msg=verbose && myid==0;

  if(msg) std::cout << "Schur complement element extraction..." << std::endl;
  if(use_HSS) {
    if(!compressed) {
      if(!myid) std::cout << "Matrix was not previously compressed." << std::endl;
      if(!myid) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(factored_HSS) {
      if(myid==0) std::cout << "Matrix fully factored; this cannot be used." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!partially_factored_HSS) {
      if(myid==0) std::cout << "Matrix was not previously partially ULV-factored." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }
    if(!schur_HSS) {
      if(myid==0) std::cout << "HSS Schur complement was not previously computed." << std::endl;
      if(myid==0) std::cout << "Aborting..." << std::endl << std::endl;
      return;
    }

    if(np==1) {
      /* Extract elements from HSS(F22): indices need to be shifted */
      tmpI=new int[nI];
      for(i=0;i<nI;i++)
        tmpI[i]=I[i]+var_schur-1;
      tmpJ=new int[nJ];
      for(i=0;i<nJ;i++)
        tmpJ[i]=J[i]+var_schur-1;
      htree.HSS_seq_Extraction(B,tmpI,nI,tmpJ,nJ);
      delete[] tmpI;
      delete[] tmpJ;

      /* Add - theta(I,:) Vt_k^* psi(:,J) */
      extractSubmatrix(htree.theta,htree.rtheta,htree.ctheta,htree.rtheta,I,nI,NULL,INONE,&tmp_theta);
      extractSubmatrix(htree.psi,htree.rpsi,htree.cpsi,htree.rpsi,NULL,INONE,J,nJ,&tmp_psi);

      if(htree.ctheta) {
        tmp=new T[htree.ctheta*nJ]();
        gemm('C','N',htree.ctheta,nJ,htree.rpsi,ONE,htree.nodes[htree.knode]->Vt,htree.rpsi,tmp_psi,htree.rpsi,ZERO,tmp,htree.ctheta);
        gemm('N','N',nI,nJ,htree.ctheta,NONE,tmp_theta,nI,tmp,htree.ctheta,ONE,B,nI);
        delete[] tmp;
      }

      delete[] tmp_theta;
      delete[] tmp_psi;

      //TODO: flops

    } else {
      /* Compute the result on the root node grid then move to the B-grid later */
      ctxt=htree.nodes[htree.dim_tree-1]->context;
      nprow=htree.nodes[htree.dim_tree-1]->nprow;
      npcol=htree.nodes[htree.dim_tree-1]->npcol;
      myrow=htree.nodes[htree.dim_tree-1]->myrow;
      mycol=htree.nodes[htree.dim_tree-1]->mycol;
      nb=ginfo.nb;
      if(myrow>=0 && mycol>=0) {
        locr=numroc_(&nI,&nb,&myrow,&IZERO,&nprow);
        locc=numroc_(&nJ,&nb,&mycol,&IZERO,&npcol);
        tmpB=new T[locr*locc]();
        idummy=std::max(locr,1);
        descinit_(desc_tmpB,&nI,&nJ,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
      } else {
        descset_(desc_tmpB,&nI,&nJ,&nb,&nb,&IZERO,&IZERO,&INONE,&IONE);
      }

      /* Extract elements from HSS(F22): indices need to be shifted */
      tmpI=new int[nI];
      for(i=0;i<nI;i++)
        tmpI[i]=I[i]+var_schur-1;
      tmpJ=new int[nJ];
      for(i=0;i<nJ;i++)
        tmpJ[i]=J[i]+var_schur-1;

      htree.HSS_par_Extraction(tmpB,desc_tmpB,tmpI,nI,tmpJ,nJ,&ginfo);
      delete[] tmpI;
      delete[] tmpJ;

      tmpinfo.ctxt=htree.nodes[htree.dim_tree-1]->context;
      tmpinfo.nprow=htree.nodes[htree.dim_tree-1]->nprow;
      tmpinfo.npcol=htree.nodes[htree.dim_tree-1]->npcol;
      tmpinfo.myrow=htree.nodes[htree.dim_tree-1]->myrow;
      tmpinfo.mycol=htree.nodes[htree.dim_tree-1]->mycol;
      tmpinfo.mb=ginfo.mb;
      tmpinfo.nb=ginfo.nb;

      /* Add - theta(I,:) Vt_k^* psi(:,J) */
      if(tmpinfo.myrow>=0 && tmpinfo.mycol>=0) {
        parExtractSubmatrix<T,S>(htree.theta,htree.rtheta,htree.ctheta,htree.desc_theta,I,nI,NULL,-1,&tmp_theta,&tmpinfo);
        parExtractSubmatrix<T,S>(htree.psi,htree.rpsi,htree.cpsi,htree.desc_psi,NULL,-1,J,nJ,&tmp_psi,&tmpinfo);

        idummy=std::max(numroc_(&nI,&nb,&myrow,&IZERO,&nprow),1);
        descinit_(desc_tmp_theta,&nI,&htree.ctheta,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
        idummy=std::max(numroc_(&htree.rpsi,&nb,&myrow,&IZERO,&nprow),1);
        descinit_(desc_tmp_psi,&htree.rpsi,&nJ,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);

        locr=numroc_(&htree.ctheta,&nb,&myrow,&IZERO,&nprow);
        locc=numroc_(&nJ,&nb,&mycol,&IZERO,&npcol);
        tmp=new T[locr*locc]();
        idummy=std::max(locr,1);
        descinit_(desc,&htree.ctheta,&nJ,&nb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);

        pgemm('C','N',htree.ctheta,nJ,htree.rpsi,ONE,htree.nodes[htree.knode]->Vt,IONE,IONE,htree.nodes[htree.knode]->descVt,tmp_psi,IONE,IONE,desc_tmp_psi,ZERO,tmp,IONE,IONE,desc);
        pgemm('N','N',nI,nJ,htree.ctheta,NONE,tmp_theta,IONE,IONE,desc_tmp_theta,tmp,IONE,IONE,desc,ONE,tmpB,IONE,IONE,desc_tmpB);
        delete[] tmp;
        delete[] tmp_theta;
        delete[] tmp_psi;
      }

      /* Move to the B-grid */
      pgemr2d(nI,nJ,tmpB,IONE,IONE,desc_tmpB,B,IONE,IONE,descB,htree.nodes[htree.dim_tree-1]->allctxt);

      if(tmpinfo.myrow>=0 && tmpinfo.mycol>=0)
        delete[] tmpB;

      //TODO: flops
    }
  } else {
    ctxt=descLU[BLACSctxt];
    n=descLU[BLACSm];
    blacs_gridinfo_(&ctxt,&nprow,&npcol,&myrow,&mycol);
    if(nprow==-1) nprow=0;
    if(np==1) {
      for(i=0;i<nI;i++)
        for(j=0;j<nJ;j++)
          B[i+j*nI]=LU[I[i]+var_schur-1-1+(J[j]+var_schur-1-1)*n];
    } else if(nprow>0) {
      /* Parallel case: nothing smart here. We communicate element
       * by element so communications most likely don't overlap
       * and latency is probably terrible. This case is mostly
       * for test purposes.
       */
      for(i=0;i<nI;i++)
        for(j=0;j<nJ;j++)
          pgeadd('N',IONE,IONE,ONE,LU,I[i]+var_schur-1,J[j]+var_schur-1,descLU,ZERO,B,i+1,j+1,descB);
    }
  }
}

template<typename T, typename S> void StrumpackDensePackage<T,S>::destroy_A22() {
  /* Release memory for the rightmost subtree.
   * This works only for the HSS case (modifying the full-rank LU layout would
   * complicate reduce_RHS and expand_solution), and we mostly need this
   * for the HSS case.
   */
  int i;

  // TODO: add a destroyed_A22 state to prevent calls to schur_product, schur_extract...

  if(compressed) {
    for(i=htree.knode+1;i<htree.dim_tree-1;i++) {
      delete[] htree.nodes[i]->D; htree.nodes[i]->D=NULL;

      delete[] htree.nodes[i]->U_E; htree.nodes[i]->U_E=NULL;
      delete[] htree.nodes[i]->U_perm; htree.nodes[i]->U_perm=NULL;
      delete[] htree.nodes[i]->descU; htree.nodes[i]->descU=NULL;
      delete[] htree.nodes[i]->Ir; htree.nodes[i]->Ir=NULL;
      delete[] htree.nodes[i]->Jr; htree.nodes[i]->Jr=NULL;

      delete[] htree.nodes[i]->V_E; htree.nodes[i]->V_E=NULL;
      delete[] htree.nodes[i]->V_perm; htree.nodes[i]->V_perm=NULL;
      delete[] htree.nodes[i]->descV; htree.nodes[i]->descV=NULL;
      delete[] htree.nodes[i]->Ic; htree.nodes[i]->Ic=NULL;
      delete[] htree.nodes[i]->Jc; htree.nodes[i]->Jc=NULL;

      delete[] htree.nodes[i]->B12; htree.nodes[i]->B12=NULL;
      delete[] htree.nodes[i]->B21; htree.nodes[i]->B21=NULL;
    }
  }

}

template<typename T, typename S> StrumpackDensePackage<T,S>::~StrumpackDensePackage() {
  if(compressed)
    htree.cleanup();
  if(factored_LU) {
    delete[] LU;
    delete[] piv;
  }
}

#endif // STRUMPACKDENSEPACKAGE_CPP_HPP
