/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef PAR_MGSINTERPOL_HPP
#define PAR_MGSINTERPOL_HPP

template<typename T, typename S> void par_MGSInterpol(int m, int n, T *A, int* descA, S tol, int *r, T **X, int **J, int **P, gridinfo *ginfo) {
  /* Parallel Interpolative Decomposition of A.
   * Decompose the m x n matrix A into A=A(:,J) [I X]*P^T
   * with J a set of columns of size r, and X rk x (n-rk).
   * Step 1: RRQR(A) => A P = Q R = Q [R1 R2]
   * Step 2: TRSM: A P = (Q R1) (R1^-1 R) => A = A(:,J)*[I X]*P^T
   *
   * Matrices are distributed and assumed to be stored in column-major format.
   * Matrix A is NOT modified.
   *
   * Matrix Q is initially a copy of A, then it contains the Q factor and is eventually discarded.
   * Matrix X has initial size n x n but is eventually reshaped into r x (n-r) (0-rows and identity columns are discarded).
   *
   * In this version, we perform rank-nbupd updates. The default size is nb; it can
   * be modified with -DMGSBLOCK=x
   *
   * Norms are updated in a fast way instead of being constantly recomputed; with
   * -DMGSNORMS, all norms are computed at each step instead.
   */

  int ctxt, nprow, npcol, myrow, mycol, mb, nb;
  int i, j, k, minmn;
  int dummy, dummy2, ierr;
  int locrA, loccA, locrR, loccR;
  int descQ[BLACSCTXTSIZE], descR[BLACSCTXTSIZE], descnorms[BLACSCTXTSIZE];
  T *Q, *R, *tmp;
  S tau, firstnorm;
	  S *norms;
  int imax;
  int nbupd;
  bool rank_revealed;

  T NONE=-1;
  T ZERO= 0;
  T  ONE= 1;

#if defined(checkID)
  T *B, *C;
  S *work;
  S err;
#endif

#if defined(checkID)||defined(debug)
  int myid;
  MPI_Comm_rank(MPI_COMM_WORLD,&myid);
#endif

  ctxt=ginfo->ctxt;
  nprow=ginfo->nprow;
  npcol=ginfo->npcol;
  myrow=ginfo->myrow;
  mycol=ginfo->mycol;
  mb=ginfo->mb;
  nb=ginfo->nb;

  locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
  loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);

#if defined(MGSBLOCK)
  nbupd=MGSBLOCK;
#else
  if(std::min(m,n)<128)
    /* For small problems, use the BLAS1 version. This helps with accuracy. */
    nbupd=1;
  else
    nbupd=nb;
#endif

  /* Q: m x n, copy of A */
  Q=NULL;
  Q=new T[locrA*loccA];
  for(i=0;i<locrA*loccA;i++)
    Q[i]=A[i];
  dummy=std::max(1,locrA);
  descinit_(descQ,&m,&n,&mb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);

  /* R: n x n */
  locrR=numroc_(&n,&mb,&myrow,&IZERO,&nprow);
  loccR=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
  R=NULL;
  R=new T[locrR*loccR];
  for(i=0;i<locrR*loccR;i++)
    R[i]=ZERO;
  dummy=std::max(1,locrR);
  descinit_(descR,&n,&n,&mb,&nb,&IZERO,&IZERO,&ctxt,&dummy,&ierr);

  /* J: indices of permuted colums */
  *J=new int[n];
  for(i=0;i<n;i++)
    (*J)[i]=i+1;

  /* P: permutation of columns */
  *P=new int[n];
  for(i=0;i<n;i++)
    (*P)[i]=i+1;

  /* Norms: distributed 1 x n array containing (the square of) norms */
  norms=new S[loccA];
  descinit_(descnorms,&IONE,&n,&mb,&nb,&IZERO,&IZERO,&ctxt,&IONE,&ierr);
  for(i=0;i<loccA;i++)
    norms[i]=std::pow(nrm2<T,S>(locrA,&Q[i*locrA],IONE),2);
  gsum2d(ctxt,'C',' ',IONE,loccA,norms,IONE,INONE,INONE);

  /* Used to grab a single row of R */
  tmp=new T[loccA];

  /* Step 1: RRQR */
  minmn=std::min(m,n);
  *r=minmn;
  rank_revealed=false;
  i=1;
  j=1;
  while (i<=minmn && j<=minmn && !rank_revealed) {
    for(i=j;i<=std::min(minmn,j+nbupd-1);i++) {
      /* a/ Find the column of A with maximum norm */
      if(i==n) {
        imax=n;
      } else {
        pamax(n-i+1,&tau,&imax,norms,IONE,i,descnorms,IONE);
        if (myrow==0)
          gebs2d<int>(ctxt,'C',' ',IONE,IONE,&imax,IONE);
        else
          gebr2d<int>(ctxt,'C',' ',IONE,IONE,&imax,IONE,IZERO,mycol);
      }

      /* b/ Interchange column imax (max norm) and i (current) for both Q and R */
      if(imax>i) {
        pswap(m,Q,IONE,i,descQ,IONE,Q,IONE,imax,descQ,IONE);
        pswap(IONE,norms,IONE,i,descnorms,IONE,norms,IONE,imax,descnorms,IONE);
        if(i>1)
          pswap(i,R,IONE,i,descR,IONE,R,IONE,imax,descR,IONE);
        dummy=(*J)[i-1];(*J)[i-1]=(*J)[imax-1];(*J)[imax-1]=dummy;
      }
      (*P)[i-1]=imax;

      /* b'/ Update the pivot column with what we have until now in this block, cfr geru */
      pgemv('N',m,i-j,NONE,Q,IONE,j,descQ,R,j,i,descR,IONE,ONE,Q,IONE,i,descQ,IONE);

      /* c/ Compute R(i,i) = ||A(:,i)||
       * Recompute the column norm, this is more accurate than the updated norm in norms[i]
       */
      tau=plange('F',m,IONE,Q,IONE,i,descQ,(S*)(NULL));
      if(i==1) {
        if(tau<tol) {
          *r=0;
          rank_revealed=true;
          break;
        }
        firstnorm = tau;
      }
      if (tau/firstnorm<tol) {
        *r=i-1;
        rank_revealed=true;
        break;
      }

      if(myrow==indxg2p_(&i,&mb,&myrow,&IZERO,&nprow) && mycol==indxg2p_(&i,&nb,&mycol,&IZERO,&npcol))
        /* I own the diagonal element */
        R[(indxg2l_(&i,&nb,&mycol,&IZERO,&npcol)-1)*locrR+indxg2l_(&i,&mb,&myrow,&IZERO,&nprow)-1]=static_cast<T>(tau);

      /* d/ Compute Q(:,i) = A(:,i)/R(i,i) */
      pscal(m,ONE/static_cast<T>(tau),Q,IONE,i,descQ,IONE);

      /* e/ Compute R(i,i+1:n) = Q(:,i)'*A(:,i+1:n); computed via transposition: A(:,i+1)'*Q(:,i) */
      pgemv('C',m,n-i,ONE,Q,IONE,i+1,descQ,Q,IONE,i,descQ,IONE,ZERO,R,i,i+1,descR,n);
      placgv(n-i,R,i,i+1,descR,n);

      /* e'/ Update norms */
      pgemr2d(1,n-i,R,i,i+1,descR,tmp,1,i+1,descnorms,ctxt);
      for(k=0;k<loccA;k++)
        norms[k]-=std::pow(std::abs(tmp[k]),2);
    }

    /* f/ Update trailing matrix: columns j:i-1 have been factored.
     * We update the trailing columns rrrrra rank-[(i-1)-(j)+1=i-j] update.
     * Q(:,i:n) -= Q(:,j:i)*R(j:i,i:n).
     * (instead of Q(:,i)*R(i,i:n) in the rank-1 update version).
     */
    if(i<=n)
      pgemm('N','N',m,n-i+1,i-j,NONE,Q,IONE,j,descQ,R,j,i,descR,ONE,Q,IONE,i,descQ);

#if defined(MGSNORMS)
    /* In some situations, using exact norms might be more stable. */
    for(k=1;k<=loccA;k++)
      if(indxl2g_(&k,&nb,&mycol,&IZERO,&npcol)>=i)
        norms[k-1]=std::pow(nrm2<T,S>(locrA,&Q[(k-1)*locrA],IONE),2);
    gsum2d(ctxt,'C',' ',IONE,loccA,norms,IONE,INONE,INONE);
#endif

    /* Next block */
    j+=nbupd;
  }

  delete[] norms;

#if defined(debug)
  if(!myid) std::cout << "Rank = " << *r << std::endl << std::endl;
  printMatrix("Q: "+SSTR(myid),Q,locrA,loccA,locrA);
  printMatrix("R: "+SSTR(myid),R,locrR,loccR,locrR);
#endif

#if defined(checkID)
  /* Copy A into B. Permute B into AP, compute B=AP-QR and norm(B)/norm(A). */
  B=NULL;
  B=new T[locrA*loccA];
  for(i=0;i<locrA*loccA;i++)
    B[i]=A[i];

  for(i=1;i<=std::min(*r+1,n);i++)
    pswap(m,B,IONE,(*P)[i-1],descA,IONE,B,IONE,i,descA,IONE);

#if defined(debug)
  printMatrix("B: "+SSTR(myid),B,locrA,loccA,locrA);
#endif

  pgemm('N','N',m,n,*r,NONE,Q,IONE,IONE,descQ,R,IONE,IONE,descR,ONE,B,IONE,IONE,descA);

#if defined(debug)
  printMatrix("AP-QR: "+SSTR(myid),B,locrA,loccA,locrA);
#endif

  work=NULL;
  work=new S[locrA];
  err=plange('I',m,n,B,IONE,IONE,descA,work)/plange('I',m,n,A,IONE,IONE,descA,work);
  if(!myid) std::cout << "Parallel RRQR: ||AP-QR||_inf/||A||_inf=    " << err << std::endl << std::endl;
  delete[] B;
  delete[] work;
#endif

  delete[] Q;

  if(*r==0) {
    delete[] R;
    *X=NULL;
    delete[] *J;
    *J=NULL;
    return;
  }

  /* Step 2: TRSM and permutation: R1^-1 R = [I R1^-1 R2] = [I X] with R = [R1 R2], R1 r x r */
  ptrsm('L','U','N','N',*r,n-*r,ONE,R,IONE,IONE,descR,R,IONE,std::min(*r+1,n),descR);

  dummy=n-*r;
  locrR=numroc_(r,&mb,&myrow,&IZERO,&nprow);
  loccR=numroc_(&dummy,&nb,&mycol,&IZERO,&npcol);
  dummy2=std::max(1,locrR);
  *X=NULL;
  *X=new T[locrR*loccR];
  descinit_(descQ,r,&dummy,&mb,&nb,&IZERO,&IZERO,&ctxt,&dummy2,&ierr);
  pgeadd('N',*r,n-*r,ONE,R,IONE,*r+1,descR,ZERO,*X,IONE,IONE,descQ);
  delete[] R;

#if defined(debug)
  printMatrix("X: "+SSTR(myid),*X,locrR,loccR,locrR);
#endif

#if defined(checkID)
  /* Copy A into B. Compute C=A(:,J)*[I X] by copying A(:,J) into C(:,1:r) then
   * computing C(:,r+1:n)=C(:,1:r)*X. Then permute into C=C P^T.
   * Finally compute B=B-C=A-A(:,J)*[I X]*P^T and norm(B)/norm(A). */
  B=NULL;
  B=new T[locrA*loccA];
  for(i=0;i<locrA*loccA;i++)
    B[i]=A[i];

  C=NULL;
  C=new T[locrA*loccA];
  for(i=1;i<=*r;i++)
    pgeadd('N',m,1,ONE,A,IONE,(*J)[i-1],descA,ZERO,C,IONE,i,descA);
  pgemm('N','N',m,n-*r,*r,ONE,C,IONE,IONE,descA,*X,IONE,IONE,descQ,ZERO,C,IONE,*r+1,descA);
  for(i=n;i>=1;i--)
    if(i!=(*P)[i-1])
      pswap(m,C,IONE,(*P)[i-1],descA,IONE,C,IONE,i,descA,IONE);

  for(i=0;i<locrA*loccA;i++)
    B[i]=B[i]-C[i];

  work=NULL;
  work=new S[locrA];
  err=plange('I',m,n,B,IONE,IONE,descA,work)/plange('I',m,n,A,IONE,IONE,descA,work);
  if(!myid) std::cout << "Parallel RRQR: ||A-A(:,J)X||_inf/||A||_inf=" << err << std::endl << std::endl;
  delete[] B;
  delete[] C;
  delete[] work;
#endif

  int *Jtmp=new int[*r];
  for(int i=0;i<*r;i++)
    Jtmp[i]=(*J)[i];
  delete[] *J;
  *J=Jtmp;

  return;

}

#endif // PAR_MGSINTERPOL_HPP
