/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef FRONTAL_MATRIX_HPP
#define FRONTAL_MATRIX_HPP

#include <iostream>
#include <algorithm>
#include <random>
#include "blas_lapack_wrapper.hpp"
#include "CompressedSparseMatrix.hpp"
#include "MatrixReordering.hpp"
#include "HSSMatrix.hpp"
#include "TaskTimer.hpp"
#include "strumpack_parameters.hpp"
#if defined(_OPENMP)
#include "omp.h"
#endif

namespace strumpack {

  // forward declaration
  template<typename scalar_t,typename real_t> class DistributedMatrix;

  template<typename scalar_t,typename real_t,typename integer_t> class FrontalMatrix {
  public:
    integer_t sep;
    integer_t sep_begin;
    integer_t sep_end;
    integer_t dim_sep;
    integer_t dim_upd;
    integer_t dim_blk; // = dim_sep + dim_upd
    integer_t* upd;
    FrontalMatrix<scalar_t,real_t,integer_t>* lchild;
    FrontalMatrix<scalar_t,real_t,integer_t>* rchild;

    // TODO this pointer should not be stored in every front,
    // instead just pass it around whenever needed?
    CompressedSparseMatrix<scalar_t,real_t,integer_t>* A;
    integer_t p_wmem;

    FrontalMatrix(integer_t separator, CompressedSparseMatrix<scalar_t,real_t,integer_t>* _A,
		  FrontalMatrix<scalar_t,real_t,integer_t>* left_child,
		  FrontalMatrix<scalar_t,real_t,integer_t>* right_child);
    FrontalMatrix(CompressedSparseMatrix<scalar_t,real_t,integer_t>* _A,
		  FrontalMatrix<scalar_t,real_t,integer_t>* _lchild,
		  FrontalMatrix<scalar_t,real_t,integer_t>* _rchild,
		  integer_t _sep, integer_t _sep_begin, integer_t _sep_end,
		  integer_t _dim_upd, integer_t* _upd);
    virtual ~FrontalMatrix();
    void compute_parent_indices(FrontalMatrix<scalar_t,real_t,integer_t>* ch, std::vector<integer_t>& parent_ind,
				integer_t& upd_to_sep);
    template<typename index_type>
    void find_my_upd_indices(integer_t n, integer_t* ind, std::vector<index_type>& my_ind,
			     std::vector<index_type>& orig_ind);
    virtual void release_work_memory() = 0;
    virtual void UUtxR(integer_t nr_vecs, scalar_t* newR, scalar_t* newSrow, scalar_t* newScol,
		       FrontalMatrix<scalar_t,real_t,integer_t>* parent, int task_depth) = 0;
    void skinny_extend_add(integer_t cols, scalar_t* Sch, FrontalMatrix<scalar_t,real_t,integer_t>* pa, scalar_t* Spa, int task_depth);
    void extract_child_rows(integer_t cols, scalar_t* Spa, scalar_t* Sch, FrontalMatrix<scalar_t,real_t,integer_t>* pa, int task_depth);
    virtual void multifrontal_factorization(HSSStrategy<real_t>& strat, int etree_level=0, int task_depth=0) = 0;
    virtual void solve_workspace_query(integer_t& mem_size);
    void extend_add_b(FrontalMatrix<scalar_t,real_t,integer_t>* ch, scalar_t* b, scalar_t* wmem);
    void extract_b(FrontalMatrix<scalar_t,real_t,integer_t>* ch, scalar_t* b, scalar_t* wmem);
    virtual void extend_add_b(FrontalMatrix<scalar_t,real_t,integer_t>* ch,
			      DistributedMatrix<scalar_t,real_t>& b, scalar_t* wmem, int tag) {}
    virtual void extract_b(FrontalMatrix<scalar_t,real_t,integer_t>* ch,
			   DistributedMatrix<scalar_t,real_t>& b, scalar_t* wmem) {}
    void look_left(scalar_t* b, scalar_t* wmem);
    void look_left(DistributedMatrix<scalar_t,real_t>& b_sep, scalar_t* wmem);
    void look_right(scalar_t* y, scalar_t* wmem);
    void look_right(DistributedMatrix<scalar_t,real_t>& y_sep, scalar_t* wmem);
    void multifrontal_solve(scalar_t* b, scalar_t* wmem);
    void multifrontal_solve(scalar_t* b_loc, DistributedMatrix<scalar_t,real_t>* b_dist, scalar_t* wmem);
    virtual void forward_multifrontal_solve(scalar_t* b, scalar_t* wmem, int etree_level=0, int task_depth=0) {};
    virtual void forward_multifrontal_solve(scalar_t* b_loc, DistributedMatrix<scalar_t,real_t>* b_dist,
					    scalar_t* wmem, int etree_level=0, int task_depth=0);
    virtual void backward_multifrontal_solve(scalar_t* y, scalar_t* wmem, int etree_level=0, int task_depth=0) {};
    virtual void backward_multifrontal_solve(scalar_t* y_loc, DistributedMatrix<scalar_t,real_t>* b_dist,
					     scalar_t* wmem, int etree_level=0, int task_depth=0);
    virtual void get_submatrix(integer_t* rows, integer_t nr_rows, integer_t* cols, integer_t nr_cols,
			       scalar_t* values, integer_t ld, int task_depth) = 0;
    void extract(integer_t* rows, integer_t nr_rows, integer_t* cols, integer_t nr_cols,
		 scalar_t* values, integer_t ld, int task_depth);
    virtual integer_t get_maximum_rank(int task_depth=0) const { return 0; }
    virtual long long get_factor_nonzeros(int task_depth=0);
    virtual long long get_dense_factor_nonzeros(int task_depth=0);
    virtual bool isHSS() { return false; }
    virtual bool isMPI() { return false; }
    virtual void print_rank_statistics(std::ostream &out) {}
    virtual std::string type() { return "FrontalMatrix"; }
    virtual void set_HSS_partitioning(HSSPartitionTree& tree, bool is_root) {}
#if defined(dbgHSS)
    virtual FrontalMatrix<scalar_t,real_t,integer_t>* get_df() { return this; }
#endif

  protected:
    FrontalMatrix(const FrontalMatrix&) = delete;
    FrontalMatrix& operator=(FrontalMatrix const&) = delete;
    virtual long long get_node_factor_nonzeros() { return dim_blk*dim_blk-dim_upd*dim_upd; }
    virtual long long get_dense_node_factor_nonzeros() { return get_node_factor_nonzeros(); }
  };

  template<typename scalar_t,typename real_t,typename integer_t>
  FrontalMatrix<scalar_t,real_t,integer_t>::FrontalMatrix
  (integer_t separator, CompressedSparseMatrix<scalar_t,real_t,integer_t>* _A,
   FrontalMatrix<scalar_t,real_t,integer_t>* left_child,
   FrontalMatrix<scalar_t,real_t,integer_t>* right_child) :
    sep(separator), sep_begin(0), sep_end(0), dim_sep(0), dim_upd(0), dim_blk(0), upd(NULL),
    lchild(left_child), rchild(right_child), A(_A), p_wmem(0) {
  }

  template<typename scalar_t,typename real_t,typename integer_t>
  FrontalMatrix<scalar_t,real_t,integer_t>::FrontalMatrix
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* _A,
   FrontalMatrix<scalar_t,real_t,integer_t>* _lchild,
   FrontalMatrix<scalar_t,real_t,integer_t>* _rchild,
   integer_t _sep, integer_t _sep_begin, integer_t _sep_end, integer_t _dim_upd, integer_t* _upd)
    : sep(_sep), sep_begin(_sep_begin), sep_end(_sep_end), dim_sep(_sep_end - _sep_begin),
      dim_upd(_dim_upd), dim_blk(_sep_end - _sep_begin + _dim_upd),
      lchild(_lchild), rchild(_rchild), A(_A), p_wmem(0) {
    upd = new integer_t[dim_upd];
    std::copy(_upd, _upd+dim_upd, upd);
  }

  template<typename scalar_t,typename real_t,typename integer_t>
  FrontalMatrix<scalar_t,real_t,integer_t>::~FrontalMatrix() {
    delete[] upd;
    delete lchild;
    delete rchild;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrix<scalar_t,real_t,integer_t>::compute_parent_indices
  (FrontalMatrix<scalar_t,real_t,integer_t>* ch, std::vector<integer_t>& parent_ind,
   integer_t& upd_to_sep) {
    parent_ind.resize(ch->dim_upd);
    integer_t r = 0, upd_pos = 0;
    for (; r<ch->dim_upd; r++) {
      upd_pos = ch->upd[r];
      if (upd_pos >= sep_end) break;
      parent_ind[r] = upd_pos - sep_begin;
    }
    upd_to_sep = r;
    for (upd_pos=0; r<ch->dim_upd; r++) {
      integer_t t = ch->upd[r];
      while (upd[upd_pos] < t) upd_pos++;
      parent_ind[r] = upd_pos + dim_sep;
    }
  }

  /**
   * ind[0:n-1] contains a list of unsorted global indices. This
   * routine finds the corresponding indices into the upd part of this
   * front and puts those indices in my_ind. Not all indices from ind
   * need to be in my_ind, so orig_ind stores for each element in
   * my_ind, to which element in ind that element corresponds.
   */
  template<typename scalar_t,typename real_t,typename integer_t> template<typename index_type> void
  FrontalMatrix<scalar_t,real_t,integer_t>::find_my_upd_indices
  (integer_t n, integer_t* ind, std::vector<index_type>& my_ind,
   std::vector<index_type>& orig_ind) {
    my_ind.reserve(n);
    orig_ind.reserve(n);
    for (integer_t i=0; i<n; i++) {
      integer_t* l = std::lower_bound(upd, upd+dim_upd, ind[i]);
      if (l != upd+dim_upd && *l == ind[i]) {
	my_ind.push_back((l-upd));
	orig_ind.push_back(i);
      }
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrix<scalar_t,real_t,integer_t>::solve_workspace_query(integer_t& mem_size) {
    if (lchild) lchild->solve_workspace_query(mem_size);
    if (rchild) rchild->solve_workspace_query(mem_size);
    p_wmem = mem_size;
    mem_size += dim_upd;
  }

  template<typename scalar_t,typename real_t,typename integer_t> inline void
  FrontalMatrix<scalar_t,real_t,integer_t>::extend_add_b
  (FrontalMatrix<scalar_t,real_t,integer_t>* ch, scalar_t* b, scalar_t* wmem) {
    integer_t r = 0, upd_pos = 0;
    scalar_t* ch_wmem = wmem + ch->p_wmem;
    scalar_t* pa_wmem = wmem + p_wmem;
    for (; r<ch->dim_upd; r++) { // to parent separator
      upd_pos = ch->upd[r];
      if (upd_pos >= sep_end) break;
      b[upd_pos] += ch_wmem[r];
    }
    upd_pos = std::lower_bound(upd, upd+dim_upd, upd_pos) - upd;
    for (; r<ch->dim_upd; r++) { // to parent update matrix
      integer_t t = ch->upd[r];
      while (upd[upd_pos] < t) upd_pos++;
      pa_wmem[upd_pos] += ch_wmem[r];
    }
    STRUMPACK_FLOPS((is_complex<scalar_t>()?2:1)*static_cast<long long int>(ch->dim_upd));
    STRUMPACK_BYTES(sizeof(scalar_t)*static_cast<long long int>(3*ch->dim_upd)+sizeof(integer_t)*(ch->dim_upd+dim_upd));
  }

  /**
   * Assemble b(I^{upd}) from [b(I^{sep});b(I^{upd})] of the parent.
   * b(I^{upd}) is stored in wmem+p_wmem.
   */
  template<typename scalar_t,typename real_t,typename integer_t> inline void
  FrontalMatrix<scalar_t,real_t,integer_t>::extract_b
  (FrontalMatrix<scalar_t,real_t,integer_t>* ch, scalar_t* b, scalar_t* wmem) {
    integer_t r = 0, upd_pos = 0;
    scalar_t* ch_b_upd = wmem+ch->p_wmem;
    scalar_t* b_upd = wmem+p_wmem;
    for (; r<ch->dim_upd; r++) { // to parent separator
      upd_pos = ch->upd[r];
      if (upd_pos >= sep_end) break;
      ch_b_upd[r] = b[upd_pos];
    }
    upd_pos = std::lower_bound(upd, upd+dim_upd, upd_pos) - upd;
    for (; r<ch->dim_upd; r++) { // to parent update matrix
      integer_t t = ch->upd[r];
      while (upd[upd_pos] < t) upd_pos++;
      ch_b_upd[r] = b_upd[upd_pos];
    }
    STRUMPACK_FLOPS((is_complex<scalar_t>()?2:1)*static_cast<long long int>(ch->dim_upd));
    STRUMPACK_BYTES(sizeof(scalar_t)*static_cast<long long int>(3*ch->dim_upd)+sizeof(integer_t)*(ch->dim_upd+dim_upd));
  }

  template<typename scalar_t,typename real_t,typename integer_t> inline void
  FrontalMatrix<scalar_t,real_t,integer_t>::look_left(scalar_t* b, scalar_t* wmem) {
    scalar_t* tmp = wmem + this->p_wmem;
    std::fill(tmp, tmp+dim_upd, scalar_t(0.));
    STRUMPACK_BYTES(sizeof(scalar_t)*static_cast<long long int>(dim_upd));
    if (lchild) extend_add_b(lchild, b, wmem);
    if (rchild) extend_add_b(rchild, b, wmem);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrix<scalar_t,real_t,integer_t>::look_right
  (scalar_t* y, scalar_t* wmem) {
    if (lchild) extract_b(lchild, y, wmem);
    if (rchild) extract_b(rchild, y, wmem);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrix<scalar_t,real_t,integer_t>::look_left
  (DistributedMatrix<scalar_t,real_t>& b_sep, scalar_t* wmem) {
    TIMER_TIME(LOOK_LEFT, 0, t_look);
    if (lchild) extend_add_b(lchild, b_sep, wmem, 0);
    if (rchild) extend_add_b(rchild, b_sep, wmem, 1);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrix<scalar_t,real_t,integer_t>::look_right
  (DistributedMatrix<scalar_t,real_t>& y_sep, scalar_t* wmem) {
    TIMER_TIME(LOOK_RIGHT, 0, t_look);
    if (lchild) extract_b(lchild, y_sep, wmem);
    if (rchild) extract_b(rchild, y_sep, wmem);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrix<scalar_t,real_t,integer_t>::multifrontal_solve(scalar_t* b, scalar_t* wmem) {
    TIMER_TIME(FORWARD_SOLVE, 0, t_fwd);
    forward_multifrontal_solve(b, wmem);
    TIMER_STOP(t_fwd);
    TIMER_TIME(BACKWARD_SOLVE, 0, t_bwd);
    backward_multifrontal_solve(b, wmem);
    TIMER_STOP(t_bwd);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrix<scalar_t,real_t,integer_t>::multifrontal_solve
  (scalar_t* b_loc, DistributedMatrix<scalar_t,real_t>* b_dist, scalar_t* wmem) {
    TIMER_TIME(FORWARD_SOLVE, 0, t_fwd);
    forward_multifrontal_solve(b_loc, b_dist, wmem);
    TIMER_STOP(t_fwd);
    TIMER_TIME(BACKWARD_SOLVE, 0, t_bwd);
    backward_multifrontal_solve(b_loc, b_dist, wmem);
    TIMER_STOP(t_bwd);
  }
  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrix<scalar_t,real_t,integer_t>::forward_multifrontal_solve
  (scalar_t* b_loc, DistributedMatrix<scalar_t,real_t>* b_dist, scalar_t* wmem, int etree_level, int task_depth) {
    forward_multifrontal_solve(b_loc, wmem, etree_level, task_depth);
  }
  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrix<scalar_t,real_t,integer_t>::backward_multifrontal_solve
  (scalar_t* y_loc, DistributedMatrix<scalar_t,real_t>* b_dist, scalar_t* wmem, int etree_level, int task_depth) {
    backward_multifrontal_solve(y_loc, wmem, etree_level, task_depth);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrix<scalar_t,real_t,integer_t>::extract
  (integer_t* rows, integer_t nr_rows, integer_t* cols, integer_t nr_cols,
   scalar_t* values, integer_t ld, int task_depth) {
    A->extract_separator(sep_end, rows, nr_rows, cols, nr_cols, values, ld, task_depth);
    if (lchild) lchild->get_submatrix(rows, nr_rows, cols, nr_cols, values, ld, task_depth);
    if (rchild) rchild->get_submatrix(rows, nr_rows, cols, nr_cols, values, ld, task_depth);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrix<scalar_t,real_t,integer_t>::skinny_extend_add
  (integer_t cols, scalar_t* Sch, FrontalMatrix<scalar_t,real_t,integer_t>* pa, scalar_t* Spa, int task_depth) {
    // TODO use a taskloop instead (OpenMP 4.5)
    for (int c=0; c<cols; c+=params::random_vector_panel) {
#pragma omp task default(shared) firstprivate(c) if((task_depth < params::task_recursion_cutoff_level) && (cols > params::random_vector_panel)) mergeable
      {
	integer_t r = 0, upd_pos = 0;
	auto n = std::min(params::random_vector_panel, int(cols-c));
	for (; r<dim_upd; r++) { // to parent separator
	  upd_pos = upd[r];
	  if (upd_pos >= pa->sep_end) break;
	  blas::axpy(n, scalar_t(1.), Sch+r+c*dim_upd, dim_upd, Spa+upd_pos-pa->sep_begin+c*pa->dim_blk, pa->dim_blk);
	}
	for (upd_pos=0; r<dim_upd; r++) { // to parent update matrix
	  integer_t t = upd[r];
	  while (pa->upd[upd_pos] < t) upd_pos++;
	  blas::axpy(n, scalar_t(1.), Sch+r+c*dim_upd, dim_upd, Spa+pa->dim_sep+upd_pos+c*pa->dim_blk, pa->dim_blk);
	}
      }
    }
#pragma omp taskwait
    STRUMPACK_FLOPS((is_complex<scalar_t>()?2:1)*static_cast<long long int>(dim_upd)*cols);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  FrontalMatrix<scalar_t,real_t,integer_t>::extract_child_rows
  (integer_t cols, scalar_t* Spa, scalar_t* Sch, FrontalMatrix<scalar_t,real_t,integer_t>* pa, int task_depth) {
    // TODO use a taskloop instead (OpenMP 4.5)
    for (int c=0; c<cols; c+=params::random_vector_panel) {
#pragma omp task firstprivate(c) if((task_depth < params::task_recursion_cutoff_level) && (cols > params::random_vector_panel))
      {
	auto n = std::min(params::random_vector_panel, int(cols-c));
	integer_t r = 0, upd_pos = 0;
	for (; r<dim_upd; r++) { // separator
	  upd_pos = upd[r];
	  if (upd_pos >= pa->sep_end) break;
	  blas::copy(n, Spa+upd_pos-pa->sep_begin+c*pa->dim_blk, pa->dim_blk, Sch+r+c*dim_upd, dim_upd);
	}
	for (upd_pos=0; r<dim_upd; r++) { // update matrix
	  integer_t t = upd[r];
	  while (pa->upd[upd_pos] < t) upd_pos++;
	  blas::copy(n, Spa+pa->dim_sep+upd_pos+c*pa->dim_blk, pa->dim_blk, Sch+r+c*dim_upd, dim_upd);
	}
      }
    }
#pragma omp taskwait
  }

  template<typename scalar_t,typename real_t,typename integer_t> long long
  FrontalMatrix<scalar_t,real_t,integer_t>::get_factor_nonzeros(int task_depth) {
    long long nnz = get_node_factor_nonzeros();
    long long nnzl = 0;
    long long nnzr = 0;
    if (lchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
      nnzl = lchild->get_factor_nonzeros(task_depth+1);
    if (rchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
      nnzr = rchild->get_factor_nonzeros(task_depth+1);
#pragma omp taskwait
    return nnz + nnzl + nnzr;
  }

  template<typename scalar_t,typename real_t,typename integer_t> long long
  FrontalMatrix<scalar_t,real_t,integer_t>::get_dense_factor_nonzeros(int task_depth) {
    long long nnz = get_dense_node_factor_nonzeros();
    long long nnzl = 0;
    long long nnzr = 0;
    if (lchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
      nnzl = lchild->get_dense_factor_nonzeros(task_depth+1);
    if (rchild)
#pragma omp task untied default(shared) final(task_depth >= params::task_recursion_cutoff_level-1) mergeable
      nnzr = rchild->get_dense_factor_nonzeros(task_depth+1);
#pragma omp taskwait
    return nnz + nnzl + nnzr;
  }

} // end namespace strumpack

#endif //FRONTAL_MATRIX_HPP
